/*
;
; phantom 16-bit setup routines & 32-bit core startup
; 4.4.1998sa
;
; This code asks the bios for various machine parameters, and
; puts them in a "safe" place: 0x90000-0x901FF, ie where the
; boot-block used to be. It is then up to the protected mode
; system to read them from there before the area is overwritten
; for buffer-blocks (kernel/head.S).
;
; Stack is situated at 0x9000:0x0000, so it should not damage our
; data we have stored (include/kernel.h).
;
; debugging informations (upper right screen corner) legend:
;	'v' - getting video information
;       'm' - getting extended memory size
;       'k' - setting keyboard repeat rate
;       '2' - seeking for ps/2 mouse device
;       'i' - killing all interrupts
;       'a' - turning a20 gate on
;       'c' - reseting mathematic coprocessor
;       'I' - remapping iqrs
;       'p' - switching to protected mode
;       'P' - we are in protected mode
;
; GFX mode selection terminates by enter, esc and space...
;
*/

#include <asm/kernel.h>

[bits 16]
[org 0x200]

#define printd16(a) mov	byte [gs:158],a

#ifndef MONO
#define printd32(a) mov byte [0xb809e],a
#else
#define printd32(a) mov byte [0xb009e],a
#endif

start:
jmp skip_header			/* 3 bytes */
/*

Setup header, nicely aligned. Used by bootsect and MegaLo.

*/
db KERNEL_HEADER		/* 4 bytes */
version:
	db	0		/* unused... */

setup_blocks:
	dw	0
kernel_blocks:
	dw	0			

dd	0

/* always starts at 0x0210 */
sysinfo:
sys_memory:
	dd	0
sys_ps2mouse:
	dd	0
sys_mono:
	dd	0
sys_root_dev:
	dd	0
sys_ramdisk_block:
	dd	0
sys_ramdisk_blocks:
	dd	0
sys_ramdisk_base:
	dd	0
sys_video_mode:
	dd	0		/* TEXT mode */
sys_kernel_text_pages:
	dd	0
sys_kernel_pages:
	dd	0

#ifdef BOOT_GFX_MODE_SELECT
/* video mode informations -- supported VESA video modes */
video_banner:
    db	"chose mode:", 13, 10
    db	"  0 .. 320x200x24b", 13, 10
    db	"  1 .. 320x200x8b", 13, 10
    db	0

mode_table:
__320x200x24b:
    db	'0'
    dw	0x010F
__320x200x8b:
    db	'1'
    dw	0x0013
table_end:
    db	0xFF			/* end of table */
#endif

skip_header:
#ifdef VGA80x50
  mov ax,0x1202
  mov bl,0x30
  int 0x10     			/* set 400 scan lines */
  mov ax,3
  int 0x10     			/* set Text mode */
  mov ax,0x1112
  mov bl,0
  int 0x10     			/* load 8x8 font to page 0 block */
#endif

/* note: fs and gs should not be used in code except for debug */    
	cld			/* do not trust in a loader */
	sti
	mov	ax, SETUP_SEG
	mov	es, ax
	mov	ds, ax
	mov	ss, ax
	xor	sp, sp

#ifdef DEBUG
#ifndef MONO
        mov     ax,0xb800
        mov     gs,ax
        mov     byte [gs:159],15
#else
	mov	ax,0xb000
	mov	gs,ax
	mov	byte [gs:159],0xf8
#endif
#endif

; obtaining root_dev from bootsect is history 3.3.2002
;	mov	ax,[0x1FC]		/* standard root_dev position */
;	mov	[sys_root_dev],ax

	mov	ax,[kernel_blocks]	/* asuming blocksize 512 bytes */
	add	ax,7			/* page alignment */
	shr	ax,3

/* if this were as kernel_text_pages would not be needed... */
	mov	word [sys_kernel_text_pages], ax

#ifdef DEBUG
        printd16('v')
#endif
	call	get_video

#ifdef BOOT_GFX_MODE_SELECT
	push	ds
	xor	ax, ax
	mov	ds, ax
	mov	si, 0x417
	lodsb
	pop	ds
	
	and	al, 0x03		/* any shift down ? */
	jz	@no_setting

@flush_buffer:
	mov	ah, 0x01
	int	0x16
	jz	@ask_for_mode
	
	xor	ah, ah
	int	0x16
	
	jmp	@flush_buffer

@ask_for_mode:
	mov	si, video_banner
	call	print

@wait_for_key:
	mov	ah, 0x01
	int	0x16
	jz	@wait_for_key	

	xor	ah, ah
	int	0x16

	cmp	al, ' '
	je	@no_setting

	cmp	al, 0x0D
	je	@no_setting
	
	cmp	al, 0x1B
	je	@no_setting
	
	mov	si, mode_table
	
@next_mode:
	cmp	byte [si], 0xFF
	je	@wait_for_key

	cmp	byte [si], al
	je	@found_mode
	
	add	si, 3
	jmp	@next_mode

@found_mode:
	inc	si
	mov	bx, [si]
	mov	[sys_video_mode], bx

	cmp	bx, 0x0100
	jge	@vesa_method

	mov	ax, bx
	int	0x10
	
	jmp	@no_setting

@vesa_method:
	mov	ax, 0x4F02
	int	0x10

@no_setting:
#endif

#ifdef DEBUG
        printd16('m')
#endif

; get memory size (extended mem, kB)
        mov     ah,0x88
	int	0x15
        mov     [sys_memory],ax

#ifdef DEBUG
        printd16('k')
#endif

; set the keyboard repeat rate to the max
        mov     ax,0x0305
        xor     bx,bx
	int	0x16
	
#ifdef DEBUG
        printd16('2')
#endif

/* check for PS/2 pointing device */
	push	es	
        pop     ds
        mov     byte [sys_ps2mouse],0           ; default is no pointing device
        int     0x11            		; int 0x11: equipment determination
        test    al,0x04        			; check if pointing device installed
	jz	no_psmouse
        mov     byte [sys_ps2mouse],0xAA   	; device present
no_psmouse:

;       jumping to protected mode..

#ifdef  DEBUG
        printd16('i')
#endif

        cli                             ;interrupts disable
        mov al,0x80
        out 0x70,al                     ;disabling NMI

#ifdef  DEBUG
        printd16('a')
#endif

; enable A20
        call    empty_8042
        mov     al,0xD1                ; command write
        out     0x64,al
	call	empty_8042
        mov     al,0xDF                ; A20 on
        out     0x60,al
	call	empty_8042

#ifdef  DEBUG
        printd16('c')
#endif

; make sure any possible coprocessor is properly reset..
        xor     ax,ax
        out     0xf0,al
	call	delay
        out     0xf1,al
	call	delay

#ifdef  DEBUG
        printd16('I')
#endif

        mov     al,0x11                ; initialization sequence
        out     0x20,al                ; send it to 8259A-1
	call	delay
        out     0xA0,al                ; and to 8259A-2
	call	delay
        mov     al,0x20                /* start of hardware int's (0x20) */
        out     0x21,al
	call	delay
        mov     al,0x28                /* start of hardware int's (0x28) */
        out     0xA1,al
	call	delay
        mov     al,0x04                ; 8259-1 is master
        out     0x21,al
	call	delay
        mov     al,0x02                ; 8259-2 is slave
        out     0xA1,al
	call	delay
        mov     al,0x01                ; 8086 mode for both
        out     0x21,al
	call	delay
        out     0xA1,al
	call	delay
        mov     al,0xFF                ; mask off all interrupts for now
        out     0xA1,al
	call	delay
        mov     al,0xFB                /* mask all irq's but irq2 which */
        out     0x21,al                ; is cascaded

#ifdef DEBUG
        printd16('p')
#endif

        lidt    [idt_48]          ; load idt with 0,0
        lgdt    [gdt_48]          ; load gdt with whatever appropriate


	push	dword 0
	popfd			  	/* preparing eflags */

        mov     eax,cr0
        or      al,1
        mov     cr0,eax

        jmp     @reset_queue
@reset_queue:
        db      0x66,0xea               ; 32-bit far jump
	dd	0x10000
        dw      KERNEL_CS

;
; routines & other procedural stuff
;

; Routine to print asciiz-string at DS:SI (emulate teletype..)
; should not damage anything (?)

print:
	push	ax
        mov     ah,0x0e
@print_nextchar:
        lodsb
        or      al,al
        jz      @print_end
        int     0x10
        jmp     @print_nextchar
@print_end:
	pop	ax
        ret
	
halt:
	push	cs
	pop	ds
        mov     si,halted
        call    print
@halt:
        jmp     @halt

get_video:
	mov	ax,0x1a00
	int	0x10
	cmp	bl,8
        jne	@is_mono
	ret			
@is_mono:	
	cmp	bl,7
	jne	@not_vga
	mov	byte [sys_mono],0xff
	ret
@not_vga:
	mov	si,no_vga
	call	print
	jmp	halt

/*
;
; error messages
;
*/

no_vga		db 'VGA card not found!',13,10,0
wrong_processor db 'wrong processor',13,10,0
halted          db 'system halted!',0

/*
; This routine checks that the keyboard command queue is empty
; (after emptying the output buffers)
;
; No timeout is used - if this hangs there is something wrong with
; the machine, and we probably couldn't proceed anyway.
*/

empty_8042:
	call	delay
        in      al,0x64        ; 8042 status port
        test    al,1           ; output buffer?
	jz	no_output
	call	delay
        in      al,0x60        ; read it
	jmp	empty_8042
no_output:
        test    al,2           ; is input buffer full?
        jnz     empty_8042      ; yes - loop
	ret
/*
;
; Delay is needed after doing I/O
;
*/
delay:
        dw   0x00eb                  ; jmp $+2
	ret

;
; Descriptor tables
;

gdt:
        dw   0,0,0,0         ; dummy

        dw   0xFFFF          ; 4Gb - (0x100000*0x1000 = 4Gb)
        dw   0x0000          ; base address=0
        dw   0x9A00          ; code read/exec
        dw   0x00CF          ; granularity=4096, 386 (+5th nibble of limit)

        dw   0xFFFF          ; 4Gb - (0x100000*0x1000 = 4Gb)
        dw   0x0000          ; base address=0
        dw   0x9200          ; data read/write
        dw   0x00CF          ; granularity=4096, 386 (+5th nibble of limit)

gdt_end:

idt_48:
        dw   0               /* count of interrupt entries (*8) */
        dw   0,0             /* idt base=0L */

gdt_48:
        dw   gdt_end-gdt     ; gdt limit=2048, 256 GDT entries
        dw   gdt,0x9         ; gdt base = 0X9xxxx

