/*
   fs/buffer.c

   Block buffering routines.
   Default blocksize set to 512 bytes.

   FIXME/WRITEME: 
   * add free memory checking/freeing into bmalloc!
   * add breada
 */

#include <phantom/buffer.h>
#include <sys/types.h>
#include <conio.h>
#include <string.h>
#include <phantom/blk.h>
#include <phantom/major.h>
#include <phantom/sched.h>
#include <phantom/mm.h>
#include <phantom/dfs.h>
#include <asm/system.h>
#include <phantom/lock.h>

/* buffer memory manager */
#define NR_BADDR	508	/* buffer addresses per bpage */
#define LTAIL		(-1)	/* list tail */

#define FREE_LIMIT	8192
#define IDLE_BP_LIMIT	0x08

#define BAD_FREE	0x00
#define BAD_VIRGIN	0x01

#define NR_BMAP		0x01

struct baddr_st
{
  ulg addr;
  short next;
  short prev;
};

/* this must be 4096 bytes or less ! */
struct bpage_descriptor
{
  struct bpage_descriptor *next_free;
  struct bpage_descriptor *prev_free;
  struct bpage_descriptor *next_virgin;
  struct bpage_descriptor *prev_virgin;
  struct bmap_descriptor *bm;
  word nr_free;
  word nr_virgin;
  short first_free;
  short first_virgin;
  struct baddr_st baddr[NR_BADDR];
};

struct bmap_descriptor
{
  int block_size;		/* block size */
  int mask;			/* free mask */
  int mask_size;		/* buffers per page */
  int nr_free;			/* free buffers */
  int nr_virgin;		/* unused entries */
  int nr_pages;			/* pages actually used */
  int idle_bp;			/* existing idle bps */
  struct bpage_descriptor *first_free;	/* free-list */
  struct bpage_descriptor *first_virgin;	/* virgin-list */
};

struct bmap_descriptor bmap[NR_BMAP] = {
  {512, 0xFF, 8, 0, 0, 0, 0, NULL, NULL}
};

struct bhead *bhead_list = NULL;

void
bp_fl_unlink (struct bpage_descriptor *bp)
{
  if (bp->prev_free)
    bp->prev_free->next_free = bp->next_free;
  else
    bp->bm->first_free = bp->next_free;
  if (bp->next_free)
    bp->next_free->prev_free = bp->prev_free;
}

void
bp_vl_unlink (struct bpage_descriptor *bp)
{
  if (bp->prev_virgin)
    bp->prev_virgin->next_virgin = bp->next_virgin;
  else
    bp->bm->first_virgin = bp->next_virgin;
  if (bp->next_virgin)
    bp->next_virgin->prev_virgin = bp->prev_virgin;
}

void
baddr_unlink (struct bpage_descriptor *bp, int baddr_nr, int fl_virgin)
{
  struct baddr_st *baddr = bp->baddr + baddr_nr;

  if (baddr->prev != LTAIL)
    bp->baddr[baddr->prev].next = baddr->next;
  else
    {
      if (fl_virgin)
	bp->first_virgin = baddr->next;
      else
	bp->first_free = baddr->next;
    }
  if (baddr->next != LTAIL)
    bp->baddr[baddr->next].prev = baddr->prev;
}

void
baddr_link (struct bpage_descriptor *bp, int baddr_nr, int fl_virgin)
{
  struct baddr_st *baddr = bp->baddr + baddr_nr;

  baddr->prev = LTAIL;
  baddr->next = fl_virgin ? bp->first_virgin : bp->first_free;
  if (fl_virgin)
    bp->first_virgin = baddr_nr;
  else
    bp->first_free = baddr_nr;
}

/* first memory back-gainer, kills everything
   so every buffer should be free (fl_dirty=0) */
void
bp_killall (struct bpage_descriptor *bp)
{
  int i, mask;
  struct baddr_st *baddr;

#ifdef BUFFER_DEBUG
  printk ("bp_killall: bp=0x%p\n", bp);
#endif

  mask = bp->bm->mask;
  baddr = bp->baddr;
  for (i = 0; i < NR_BADDR; i++)
    if (baddr[i].addr)
      {
	if ((baddr[i].addr & mask) != mask)
	  printk ("bp_killall: warning: buffer %d in use\n", i);
	free_pages (baddr[i].addr & ADDR_MASK, 1);
      }

  bp_fl_unlink (bp);
  if (bp->nr_virgin)
    bp_vl_unlink (bp);

  free_pages (TOPAGE (bp), 1);
}

/* this gains memory of obsolete buffers 
   it should spawn virgins from frees and/or
   destroy some buffer_pages 

   this routine could be called by get_free_pages
   and get_free_page (mm/page.c)
   */
void
bcleanup (struct bpage_descriptor *bp, int nr_pages)
{
/*  int i, mask = bp->bm->mask;
  struct baddr_st *baddr;

#ifdef BUFFER_DEBUG
  printk ("bp_killfree: nr_pages=%d\n", nr_pages);
#endif

  if (!nr_pages)
    return;

  baddr = bp->baddr;

  for (i = 0; i < NR_BADDR; i++)
    {
      if ((baddr[i].addr & mask) == mask)
	{			*//* whole free, spawning new virgin */
/*	  free_pages (baddr[i].addr & ADDR_MASK, 1);
	  baddr[i].addr = 0;

	  baddr_unlink (bp, i, BAD_FREE);
	  baddr_link (bp, i, BAD_VIRGIN);

	  nr_pages--;
	  if (!nr_pages)
	    break;
	}
    }

#ifdef BUFFER_DEBUG
  printk ("%d pages left unfreed\n", nr_pages);
#endif*/
}

int
bmalloc (struct bhead *bh)
{
  CODE_LOCK(a);
  struct baddr_st *baddr;
  int i, baddr_nr;
  struct bpage_descriptor *bp;
  struct bmap_descriptor *bm;

  bm = &bmap[0];		/* 512 bytes buffers only... */

  SINGLE_CODE(a);
  if (bm->nr_free)
    {
      bp = bm->first_free;
      if (!bp)
	{
	  MULTI_CODE(a);	
	  printk ("bmalloc: free-list corrupt, bm->nr_free=%d\n",
		  bm->nr_free);
	  return -1;
	}

      baddr_nr = bp->first_free;
      baddr = bp->baddr + baddr_nr;
#ifdef BUFFER_DEBUG
      printk ("FC: bh 0x%p  bp 0x%p  baddr_nr %d  baddr->addr 0x%p\n",
	      bh, bp, baddr_nr, baddr->addr);
#endif

      for (i = 0; i < bm->mask_size; i++)
	if (baddr->addr & (1 << i))
	  break;

      if (i == bm->mask_size)
	{
	  MULTI_CODE(a);
	  printk ("bmalloc: code corrupt\n");
	  return -1;
	}
	
      baddr->addr &= ~(1 << i);
      bh->data = (char *) ((baddr->addr & ADDR_MASK) + i * bm->block_size);

#ifdef BUFFER_DEBUG
      printk (" i %d  data 0x%p  bp  0x%p\n", i, bh->data, bp);
#endif

      bzero ((void *) bh->data, bm->block_size);
      bh->bp = (ulg) bp | bp->first_free;
      bm->nr_free--;
      bp->nr_free--;

#ifdef BUFFER_DEBUG
      printk ("bp->nr_free %d (left)\n", bp->nr_free);
#endif

      if (!(baddr->addr & bm->mask))
	{
#ifdef BUFFER_DEBUG
	  printk ("unlinking baddr\n");
#endif
	  baddr_unlink (bp, baddr_nr, BAD_FREE);
	}

      if (!bp->nr_free)
	{
#ifdef BUFFER_DEBUG
	  printk ("unlinking whole bp from free-list\n");
#endif
	  bp_fl_unlink (bp);
	}

      MULTI_CODE(a);
      return 0;
    }

  if (bm->nr_virgin)
    {
      bp = bm->first_virgin;
      if (!bp)
	{
	  MULTI_CODE(a);
	  printk ("bmalloc: virgin-list corrupt, bm->nr_virgin=%d\n",
		  bm->nr_virgin);
	  return -1;
	}

      baddr_nr = bp->first_virgin;
      baddr = bp->baddr + baddr_nr;

#ifdef BUFFER_DEBUG
      printk ("VC: bh 0x%p  bp 0x%p  baddr_nr %d  baddr->addr 0x%p\n",
	      bh, bp, baddr_nr, baddr->addr);
#endif

      if (baddr->addr)
	{
	  MULTI_CODE(a);
	  printk ("bmalloc: code corrupt ][\n");
	  return -1;
	}

      bh->data = (char *) page_alloc ();
      bzero ((void *) bh->data, PAGE_SIZE);
      bh->bp = (ulg) bp | bp->first_virgin;

      baddr->addr = (ulg) bh->data | (bm->mask ^ 1);

#ifdef BUFFER_DEBUG
      printk ("  data 0x%p  bp 0x%p\n", bh->data, bp);
#endif
      if (!bm->nr_free)
	{
#ifdef BUFFER_DEBUG
	  printk ("linking bp to free-list\n");
#endif
	  bp->next_free = bm->first_free;
	  if (bm->first_free)
	    bp->prev_free = bm->first_free->prev_free;
	  else
	    bp->prev_free = NULL;
	  bm->first_free = bp;
	}

      bm->nr_virgin--;
      bm->nr_free += bm->mask_size - 1;
      bp->nr_virgin--;
      bp->nr_free += bm->mask_size - 1;

      baddr_unlink (bp, baddr_nr, BAD_VIRGIN);
      baddr_link (bp, baddr_nr, BAD_FREE);

      if (!bp->nr_virgin)
	{
#ifdef BUFFER_DEBUG
	  printk ("unlinking whole bp\n");
#endif
	  bp_vl_unlink (bp);
	}

      MULTI_CODE(a);
      return 0;
    }

  if (bm->first_free || bm->first_virgin)
    {
      MULTI_CODE(a);
      printk ("bmalloc: queue corrupt, neither virgin nor free\n");
      return -1;
    }

#ifdef BUFFER_DEBUG
  printk ("bmalloc: creating brand new bpage\n");
#endif
  bp = (struct bpage_descriptor *) page_alloc ();
  bzero ((void *) bp, PAGE_SIZE);
/* also sets bp->next_* & bp->prev_* */

  bp->bm = bm;
  bp->nr_free = bm->mask_size - 1;
  bp->nr_virgin = NR_BADDR - 1;
  bp->first_free = 0;
  bp->first_virgin = 1;

  bm->nr_free = bm->mask_size - 1;
  bm->nr_virgin = NR_BADDR - 1;
  bm->first_free = bp;
  bm->first_virgin = bp;

  bh->data = (char *) page_alloc ();
  bzero ((void *) bh->data, PAGE_SIZE);
  bh->bp = (ulg) bp;		/* bp | 0 */

  baddr = bp->baddr;
  baddr->addr = (ulg) bh->data | (bm->mask ^ 1);
  baddr->next = LTAIL;
  baddr->prev = LTAIL;

#ifdef BUFFER_DEBUG
  printk ("bh->data=0x%p  baddr->addr=0x%p\n", bh->data, baddr->addr);
#endif

/* link all virgins */
  for (i = 1; i < NR_BADDR; i++)
    {
      baddr[i].next = i + 1;
      baddr[i].prev = i - 1;
    }

  baddr[1].prev = LTAIL;
  baddr[NR_BADDR - 1].next = LTAIL;

  MULTI_CODE(a);

  return 0;
}

/* opposite of bmalloc, free buffer data, clean bpage if needed... */
void
bfree (struct bhead *bh)
{
  struct baddr_st *baddr;
  struct bmap_descriptor *bm;
  struct bpage_descriptor *bp;
  int baddr_nr, block_nr, bmask;

  bp = (struct bpage_descriptor *) (bh->bp & ADDR_MASK);
  bm = bp->bm;
  baddr_nr = bh->bp & ~ADDR_MASK;
  block_nr = ((ulg) bh->data & ~ADDR_MASK) / bm->block_size;

#ifdef BUFFERDEBUG
  printk ("bfree: bp=0x%p  baddr_nr=%d  bm=0x%p  block_nr=%d\n",
	  bp, baddr_nr, bm, block_nr);
#endif

  if (block_nr >= NR_BADDR)
    {
      printk ("bfree: block_nr out of bmask\n");
      return;
    }

  bmask = bp->baddr[baddr_nr].addr;
  if (bmask & (1 << block_nr))
    {
      printk ("bfree: cannot free free buffer\n");
      return;
    }

  baddr = bp->baddr + baddr_nr;
  if (!(baddr->addr & bm->mask))	/* add to free-list for 'first' */
    baddr_link (bp, baddr_nr, BAD_FREE);

  baddr->addr |= (1 << block_nr);	/* mark it free */

  bh->data = NULL;		/* clear bhead */
  bh->bp = 0;

  bp->nr_free++;		/* refresh */
  bm->nr_free++;
}

void
clear_strategy ()
{
/* free whole bp if whole free */
/*  if (bp->nr_free / bm->mask_size + bp->nr_virgin == NR_BADDR)
   {
   printk ("bfree: whole bp free, nr_free=%d  nr_virgin=%d  IDLE_BP_LIMIT=%d\n",
   bm->nr_free, bm->nr_virgin, IDLE_BP_LIMIT);
   if (bm->idle_bp > IDLE_BP_LIMIT)
   bp_killall (bp);
   else
   bm->idle_bp++;
   return;
   } */

/* kill frees if enough and spawn virgins */
/*  if (bp->nr_free && bm->nr_free > FREE_LIMIT)
   bp_killfree (bp, bm->nr_free - FREE_LIMIT); */
}

/* -1 for as much as possible */
int
buffer_clean_up (int pages)
{
  struct bpage_descriptor *bp;
  struct bmap_descriptor *bm;
  printk ("buffer_clean_up: pages=%d, not implemented yet\n", pages);

  bm = &bmap[0];		/* 512 bytes buffers only */
/* stroll bpages to seek completely free one */

  return 0;
}

/* non memory management like part... */
void
wait_on_buffer (struct bhead *b)
{
  u32 flg;

#ifdef BUFFER_DEBUG
  printk ("wait_on_buffer: dev=0x%x  block_nr=%d\n", b->dev, b->block_nr);
#endif

  get_flags (flg);

  while (b->fl_lock)
    schedule ();

  set_flags (flg);
}

void
bload (struct bhead *b)
{
  do_request (b->dev, b->block_nr, 1, b->data, b, READ);
  wait_on_buffer(b);
}

void
bsave (struct bhead *b)
{
  do_request (b->dev, b->block_nr, 1, b->data, b, WRITE);
  wait_on_buffer (b);
}

/* remove buffer head from memory 
   note: do not swap with bfree! */
void
bdel (struct bhead *b)
{
#ifdef BUFFER_DEBUG
  printf ("bdel: b=0x%p\n", b);
#endif
  if (b->next)
    b->next->prev = b->prev;
  if (b->prev)
    b->prev->next = b->next;
  else
    /* it's a head buffer, update tree */
    {
      if (b->prev_dev)
	b->prev_dev->next_dev = b->next_dev;
      else
	bhead_list = b->next_dev;
    }
  bfree (b);
  kfree (b);
}

struct bhead *
bfind (int dev, int block_nr)
{
  struct bhead *b;

#ifdef BUFFER_DEBUG
  printk ("bfind: dev 0x%x  block_nr %d\n", dev, block_nr);
#endif

  b = bhead_list;
  while (b && b->dev != dev)
    b = b->next_dev;

  while (b)
    {
      if (b->block_nr == block_nr)
	{
#ifdef BUFFER_DEBUG
	  printk ("bfind: found\n");
#endif
	  return b;
	}
      b = b->next;
    }

#ifdef BUFFER_DEBUG
  printk ("bfind: not found\n");
#endif
  return NULL;
}

/* called only when buffer not found by bfind */
struct bhead *
badd (int dev, int block_nr)
{
  struct bhead *b, *b0, *b1;

#ifdef BUFFER_DEBUG
  printk ("badd: adding on dev 0x%x  block_nr %d\n", dev, block_nr);
#endif

  b = (struct bhead *) kmalloc (sizeof (struct bhead), PG_KERNEL);
  if (!b)
    return NULL;

  b->fl_lock = 1;
  b->dev = dev;
  b->block_nr = block_nr;

  if (bmalloc (b))
    {
      kfree (b);
      return NULL;
    }

#ifdef BUFFER_DEBUG
  printk ("buffer added, now linking\n");
#endif

  if (bhead_list == NULL)
    {
      bhead_list = b;
      return b;
    }

  b0 = bhead_list;
  while (b0 && b0->dev != dev)
    b0 = b0->next_dev;

  if (b0 == NULL)
    {
#ifdef BUFFER_DEBUG
      printk ("badd: device queue does not exist\n");
#endif
      b->next_dev = bhead_list;
      bhead_list->prev_dev = b;
      bhead_list = b;
      return b;
    }

/* device queue found, now sort the block in */
  while (b0->next && b0->block_nr < block_nr)
    b0 = b0->next;

  if (b0->next == NULL)		/* duplicate: first and only one or the last */
    {
      if (b0->block_nr < block_nr)	/* b0, b */
	{
	  b->prev = b0;		/* b->next = NULL by kmalloc */
	  b0->next = b;
	  return b;
	}
    }

/* b, b0 */
  if (b0->prev != NULL)
    {
      b0->prev->next = b;
      b->prev = b0->prev;
    }
  else
    {
/* update device queue -- device queue handling should be optimised */
      b->prev = NULL;
      b->prev_dev = b0->prev_dev;
      b->next_dev = b0->next_dev;
      b0->prev_dev = NULL;
      b0->next_dev = NULL;
    }

  b->next = b0;
  b0->prev = b;

  return b;
}

/* get buffer without reading it */
struct bhead *
bempty (int dev, int block_nr)
{
  struct bhead *b;

  if (b = bfind (dev, block_nr))
    {
#ifdef BUFFER_DEBUG
      if (b->used)
	printk ("bempty: warning, buffer in use!\n");
#endif

      if (b->fl_lock)
	wait_on_buffer (b);

      b->used++;
      return b;
    }

  b = badd (dev, block_nr);
  if (!b)
    return NULL;

  b->fl_lock = 0;		/* ? */
  b->used++;
  return b;
}

/* release a buffer */
void
brelse (struct bhead *b)
{
  if (b->used)
    b->used--;
  else
    printk ("brelse: unused buffer given\n");
}

int
sync_dev (int dev)
{
  struct bhead *b;

#ifdef BUFFER_DEBUG
  printk ("sync_dev: %02X:%02X: ", MAJOR (dev), MINOR (dev));
#endif

  b = bhead_list;
  while (b && b->dev != dev)
    b = b->next_dev;

  if (!b)
    {
      printk ("sync_dev: nothing to be synced, dev=0x%x\n", dev);
      return 0;
    }

  while (b)
    {
      if (b->fl_lock)
	wait_on_buffer (b);

      if (b->fl_dirty)
	bsave (b);
#ifdef BUFFER_DEBUG
      printk ("b%d ", b->block_nr);
#endif
      b->fl_uptodate = 0;
      b = b->next;
    }
#ifdef BUFFER_DEBUG
  printk ("\n");
#endif
  return 0;
}

struct bhead *
bread (int dev, int block_nr)
{
  struct bhead *b;

#ifdef BUFFER_DEBUG
  printk ("bread: dev 0x%x  block_nr %d\n", dev, block_nr);
#endif

/* already present ? */
  if (b = bfind (dev, block_nr))
    {
      b->used++;
      if (b->fl_lock)
	wait_on_buffer (b);
      if (!b->fl_uptodate)
	{
#ifdef BUFFER_DEBUG
	  printk ("bread: not uptodate\n");
#endif
	  bload (b);
	  if (!b->fl_uptodate)
	    {
	      bdel (b);
	      printk
		("bread: %02X:%02X, blk=%d  not uptodate before and after bload\n",
		 MAJOR (dev), MINOR (dev), block_nr);
	      return NULL;
	    }
	}
      return b;
    }

  if (!(b = badd (dev, block_nr)))
    {
      printk
	("bread: %02X:%02X, blk=%d  badd failed\n",
	 MAJOR (dev), MINOR (dev), block_nr);

      return NULL;
    }

  b->used = 1;
  b->fl_lock = 1;
  bload (b);
  if (!b->fl_uptodate)
    {
      bdel (b);
      printk ("bread: %02X:%02X, blk=%d  not uptodate after bload\n",
	      MAJOR (dev), MINOR (dev), block_nr);

      return NULL;
    }
  b->fl_lock = 0;

/* this should be placed before bload so we won't lose
   the buffer... routine as touch_floppy could be written 
   or could be placed inside floppy.c */
  if (MAJOR (dev) == FLOPPY_MAJOR && fd_changed (dev))
    {
/* write-caching of floppies should not be permited?
   compile-time flag could be used... */
      printk ("bread: floppy change indicated, syncing\n");
      sync_dev (dev);
    }
  return b;
}

void
blist ()
{
  int i;
  char *dname;
  struct bhead *bdev, *btmp;
  struct bmap_descriptor *bm;

  for (i = 0; i < NR_BMAP; i++)
    {
      bm = &bmap[i];
      printf ("block_size=%d\n", bm->block_size);
      printf (" mask=0x%x\n", bm->mask);
      printf (" mask_size=%d\n", bm->mask_size);
      printf (" nr_free=%d\n", bm->nr_free);
      printf (" nr_virgin=%d\n", bm->nr_virgin);
      printf (" nr_pages=%d\n", bm->nr_pages);
      printf (" idle_bp=%d\n", bm->idle_bp);
      printf (" first_free=0x%p\n", bm->first_free);
      printf (" first_virgin=0x%p\n", bm->first_virgin);

      printf ("any key");
      getch ();
      putchar ('\n');
    }

  printf ("\nsizeof (struct bpage_descriptor) = %d\n",
	  sizeof (struct bpage_descriptor));

  bdev = bhead_list;
  i = 0;
  while (bdev)
    {
      btmp = bdev;
      dname = dfs_getdevname (btmp->dev);
      if (!dname)
	printf ("device %02X:%02X:\n", MAJOR (btmp->dev), MINOR (btmp->dev));
      else
	printf ("%s:\n", dname);
      printf ("%-4s %-8s %-4s %-10s %-10s %s\n", "nr", "block", "used",
	      "data", "block_page", "uptodate");
      while (btmp)
	{
	  printf ("%-4d %-8d %-4d 0x%08x 0x%08x %-2d\n",
		  i, btmp->block_nr, btmp->used,
		  btmp->data, btmp->bp, btmp->fl_uptodate);
	  btmp = btmp->next;
	  i++;
	  if (!(i % 22))
	    {
	      printf ("scroll?");
	      getch ();
	      putchar ('\n');
	    }
	}
      bdev = bdev->next_dev;
      printf ("\n");
    }
  printf ("%d buffers total\n", i);
}

int
block (struct bhead *b)
{
  if (b->fl_lock)
    wait_on_buffer (b);
  b->fl_lock = 1;
}

int
bunlock (struct bhead *b)
{
  b->fl_lock = 0;
}

/* sync all devices */
int
sys_sync ()
{
  struct bhead *bdev, *btmp;

  bdev = bhead_list;
  while (bdev)
    {
      btmp = bdev;
      while (btmp)
	{
	  if (btmp->fl_dirty)
	    bsave (btmp);
	  btmp = btmp->next;
	}
      bdev = bdev->next_dev;
    }
  return 0;
}

void
buffer_done ()
{
  struct bhead *bb, *bd, *bb2, *bd2;

  printf ("buffer machinery: killing all buffers\n");

  bd = bhead_list;
  bb = bd;
  while (bd)
    {
      bd2 = bd->next_dev;
      while (bb)
	{
	  if (bb->fl_dirty)
	    {
	      printf ("s");
	      bsave (bb);
	    }
	  else
	    printf (".");
	  if (bb->used)
	    printf ("unfreed buffer: dev=0x%x  block=%d\n",
		    bb->dev, bb->block_nr);
	  bb2 = bb->next;
	  bdel (bb);
	  bb = bb2;
	}
      bd = bd2;
    }
  bhead_list = NULL;

/* 
   well, following should have been clear, but we'll reset it as well... 
   debug purpose
 */
  if (bmap[0].first_free)
    printf ("warning, bmap[0].first_free=0x%x\n", bmap[0].first_free);
  if (bmap[0].first_virgin)
    printf ("warning, bmap[0].first_virgin=0x%x\n", bmap[0].first_virgin);
  if (bmap[0].nr_free)
    printf ("warning, bmap[0].nr_free=%d\n", bmap[0].nr_free);
  if (bmap[0].nr_virgin)
    printf ("warning, bmap[0].nr_virgin=%d\n", bmap[0].nr_virgin);
  if (bmap[0].nr_pages)
    printf ("warning, bmap[0].nr_pages=%d\n", bmap[0].nr_pages);
  if (bmap[0].idle_bp)
    printf ("warning, bmap[0].idle_bp=%d\n", bmap[0].idle_bp);

  bmap[0].first_free = NULL;
  bmap[0].first_virgin = NULL;
  bmap[0].nr_free = 0;
  bmap[0].nr_virgin = 0;
  bmap[0].nr_pages = 0;
  bmap[0].idle_bp = 0;

  printf ("  everything ok\n");
}
