/*
   dev/hd.c
   Phantom IDE HD driver, 29.8.1998           
 */

#include <asm/regdump.h>
#include <asm/system.h>
#include <asm/io.h>
#include <asm/irq.h>
#include <string.h>
#include <phantom/major.h>
#include <phantom/hdregs.h>
#include <phantom/fs.h>
#include <errno.h>
#include <phantom/blk.h>
#include <conio.h>
#include <phantom/buffer.h>
#include <phantom/dfs.h>
#include <phantom/ide.h>

#define STAT_MASK (BUSY_STAT | READY_STAT | WRERR_STAT | SEEK_STAT | ERR_STAT)
#define STAT_OK (READY_STAT | SEEK_STAT)

static void hd_write_intr (IDE *, IDECTRL *, IDECMD *);
static void hd_read_intr (IDE *, IDECTRL *, IDECMD *);

void
hd_dumpstatus (int st)
{
  int err;

  if (st & BUSY_STAT)
    printk ("BUSY ");
  if (st & READY_STAT)
    printk ("READY ");
  if (st & WRERR_STAT)
    printk ("WRERR ");
  if (st & SEEK_STAT)
    printk ("SEEK ");
  if (st & DRQ_STAT)
    printk ("DRQ ");
  if (st & ECC_STAT)
    printk ("ECC ");
  if (st & INDEX_STAT)
    printk ("INDEX ");
  if (st & ERR_STAT)
    {
      printk ("ERR: ");

      err = inb (HD_ERROR);
      if (err & BBD_ERR)
	printk ("bad block,");
      if (err & ECC_ERR)
	printk ("uncorrectable ecc error,");
      if (err & ID_ERR)
	printk ("ID not found,");
      if (err & ABRT_ERR)
	printk ("command aborted,");
      if (err & TRK0_ERR)
	printk ("could not find track 0,");
      if (err & MARK_ERR)
	printk ("bad address mark,");
    }
  else
    putchar ('\n');
}

int
hd_busy (int port)		/* wait for busy controller */
{
  int i, ret = 1000000;

  while (ret--)
    {
      i = inb_p (port + HD_STATUS);
      if (!(i & BUSY_STAT))
	return 0;
    }

  return (i);
}

int
hd_sync (int port, int select_mask)
{
  int ret = 100;

  if (hd_busy (port) & BUSY_STAT)
    {
      printk ("hd_sync: drive still busy\n");
      return -1;
    }

  outb (select_mask, port + HD_CURRENT);

  while (ret--)
    {
      if (inb (port + HD_STATUS) & READY_STAT)
	return 0;
    }

  printk ("hd_sync: timeout\n");

  hd_dumpstatus (inb (port + HD_STATUS));
  return -1;
}

int
hd_access (IDE * ide, IDECTRL * ctrl, int block, int nsec, int cmd,
	   void (*intr) (IDE *, IDECTRL *, IDECMD *))
{
  int cyl, hd, sec, trk;
  int port = ctrl->port;
  int sel_mask = ide->select_mask;

#ifdef HDDEBUG
  printk ("hd_access: %02X:00  blk=%d  nsec=%d  cmd=%d\n",
	  MAJOR(ide->dev), block, nsec, cmd);
#endif

  IDE_HOOK_IRQ (ctrl, intr);
  hd_sync (port, sel_mask);

  if (ide->fl_lba)
    {
      port += HD_NSECTOR;
      outb (nsec, port++);	/* HD_NSECTOR */
      outb (block, port++);	/* HD_SECTOR */
      outb (block >> 8, port++);	/* HD_LCYL */
      outb (block >> 16, port++);	/* HD_HCYL */
      outb (sel_mask | (block >> 24) | HD_LBA, port++);	/* HD_CURRENT */
      outb (cmd, port);		/* HD_COMMAND */
    }
  else
    {
      sec = block % ide->u.hd.sec + 1;
      trk = block / ide->u.hd.sec;
      hd = trk % ide->u.hd.hd;
      cyl = trk / ide->u.hd.hd;

#ifdef HDDEBUG
      printf ("  CHS=%d/%d/%d\n", cyl, hd, sec);
#endif

      port += HD_NSECTOR;
      outb (nsec, port++);	/* HD_NSECTOR */
      outb (sec, port++);	/* HD_SECTOR */
      outb (cyl, port++);	/* HD_LCYL */
      outb (cyl >> 8, port++);	/* HD_HCYL */
      outb (sel_mask | hd, port++);	/* HD_CURRENT */
      outb (cmd, port);		/* HD_COMMAND */
    }
}

void
hd_dorequest (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  int i, disk, ret;
  REQUEST *req = cmd->req;

  switch (req->cmd)
    {
    case READ:
      hd_access (ide, ctrl, req->sec, 1, WIN_READ, hd_read_intr);
      break;
    case WRITE:
      hd_access (ide, ctrl, req->sec, 1, WIN_WRITE, hd_write_intr);
      break;
    default:
      printk ("hd_dorequest: unknown cmd %d\n", req->cmd);
      ide_end_request (cmd, 1);
      break;
    }
}

void
hd_read_intr (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  REQUEST *req = cmd->req;
  int i, ret = 100000;
  int port = ctrl->port;

  IDE_UNHOOK_IRQ (ctrl);

#ifdef HDDEBUG
  printk ("hd_read_intr: start\n");
#endif

  do
    {
      i = inb (port + HD_STATUS);
      if (i & BUSY_STAT)
	continue;
      if ((i & STAT_MASK) != STAT_OK)
	break;
      if (i & DRQ_STAT)		/* data request, buffer is waiting ... */
	goto read_ready;
    }
  while (ret--);

  printk ("hd_read_intr: read panic!\n");
  ide_end_request (cmd, 1);
  return;

read_ready:
#ifndef IDE32
  insw (port + HD_DATA, req->data, 256);
#else
  insl (port + HD_DATA, req->data, 128);
#endif

  req->data += 512;
  req->sec++;
  req->nsec--;

  if (!req->nsec)
    {
      ide_end_request (cmd, 0);
      return;
    }

  hd_dorequest (ide, ctrl, cmd);	/* finish undone cmd... */
}

static void
hd_write_intr (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  REQUEST *req = cmd->req;
  int i, ret = 100000;
  int port = ctrl->port;

  IDE_UNHOOK_IRQ (ctrl);

#ifdef HDDEBUG
  printk ("hd_read_intr: start\n");
#endif

  do
    {
      i = inb (port + HD_STATUS);
      if ((i & STAT_MASK) != STAT_OK)
	break;
      if (i & DRQ_STAT)
	goto write_ready;
    }
  while (ret--);

  printk ("hd_write_intr: write panic!\n");
  ide_end_request (cmd, 1);
  return;

write_ready:

#ifndef IDE32
  outsw (port + HD_DATA, req->data, 256);
#else
  outsl (port + HD_DATA, req->data, 128);
#endif

  req->data += 512;
  req->sec++;
  req->nsec--;

  if (!req->nsec)
    {
      ide_end_request (cmd, 0);
      return;
    }

  hd_dorequest (ide, ctrl, cmd);
}

struct file_operations hd_fops = {
  NULL,				/* lseek */
  block_read,			/* read */
  block_write,			/* write */
  NULL,				/* readdir */
  NULL,				/* ioctl */
  block_open,			/* open */
  NULL,				/* close */
};

/* Attach hd drive which was detected. This routine is used by ide_probe.c */
int
hd_attach (IDE * ide, struct drive_info *id)
{
  int ret;

  if(!id->cur_cyls || !id->cur_heads || !id->cur_sectors)
    {
    printk("hd_attach: invalid media\n");
    return -EIO;
    }

  ide->u.hd.cyl = id->cur_cyls;
  ide->u.hd.hd = id->cur_heads;
  ide->u.hd.sec = id->cur_sectors;
  ide->u.hd.bpi = (id->multsect_valid & 0x01) ? id->multsect : 1;
/* type set by ide_probe.. */

  return 0;
}
