#include <phantom/blk.h>
#include <phantom/ide.h>
#include <phantom/lock.h>
#include <asm/io.h>

/* ATAPI packet command phases. */
#define PHASE_CMDOUT    (ARS_DRQ | ARI_CMD)
#define PHASE_DATAIN    (ARS_DRQ | ARI_IN)
#define PHASE_DATAOUT   ARS_DRQ
#define PHASE_COMPLETED (ARI_IN | ARI_CMD)
#define PHASE_ABORTED   0	/* nonstandard - for NEC 260 */

/*
 * Wait uninterruptibly until controller is not busy and certain
 * status bits are set.
 * The wait is usually short unless it is for the controller to process
 * an entire critical command.
 * Return 1 for (possibly stale) controller errors, -1 for timeout errors,
 * or 0 for no errors.
 */
int
ide_wait (int port, u8 bits_wanted)
{
  int cnt;
  u8 s;

  /* Wait 5 sec for BUSY deassert. */
  for (cnt = 500000; cnt > 0; --cnt)
    {
      s = inb (port + AR_STATUS);
      if (!(s & ARS_BSY))
	break;
      sys_sleep (10);
    }
  if (cnt <= 0)
    return (-1);
  if (!bits_wanted)
    return (s & ARS_CHECK);

  /* Wait 50 msec for bits wanted. */
  for (cnt = 5000; cnt > 0; --cnt)
    {
      s = inb (port + AR_STATUS);
      if ((s & bits_wanted) == bits_wanted)
	return (s & ARS_CHECK);
      sys_sleep (10);
    }
  return (-1);
}

void
ide_trf_intr (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  if (ide == NULL)
    printk ("ide_trf_intr: ide NULL\n");

  if (ctrl == NULL)
    printk ("ide_trf_intr: ctrl NULL\n");

  if (cmd == NULL)
    printk ("ide_trf_intr: cmd NULL\n");

  if (ide_io (ide, ctrl, cmd))	/* transfer continues */
    return;

/* transfer done */
  IDE_UNHOOK_IRQ (ctrl);

  if (cmd->callback != NULL)
    cmd->callback (cmd);
  else
    {
      printk ("ide_trf_intr: callback NULL\n");
      ide_end_request (cmd, 1);
    }
}

void
ide_execcmd (IDECMD * cmd)
{				/* packet command */
  IDECTRL *ctrl = cmd->ctrl;

  IDE_HOOK_IRQ (ctrl, ide_trf_intr);

  if (ide_issue (cmd->ide, ctrl, cmd))
    {
      IDE_UNHOOK_IRQ (ctrl);
#ifdef IDEDEBUG
      printk ("ide_execcmd: cmd 0x%x ends with error\n");
#endif
      ide_end_request (cmd, 1);
      return;
    }

#ifdef IDEDEBUG
  printk ("ide_execcmd: issue ok, 0x%x waiting for interrupt..\n", cmd);
#endif
}

/* this will not run concurently */
int
ide_docmd (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  if (cmd == NULL)
    {
      printk ("ide_docmd: NULL cmd given...\n");
      ctrl->busy = 0;
      return;
    }

  if (cmd->retries < 0)
    {
      printk ("ide_docmd: negative retries, assuming 0\n");
      cmd->retries = 0;
    }
  else if (cmd->retries > MAXRETRIES)
    {
      printk ("ide_docmd: retries too high (MAXRETRIES=%d)\n", MAXRETRIES);
      cmd->retries = MAXRETRIES;
    }

/* retry purposes */
  cmd->orig_count = cmd->count;
  cmd->orig_buff = cmd->buff;

  if (cmd->req != NULL)
    {
      switch (ide->dev_type)
	{
	case IDE_HD:
	  hd_dorequest (ide, ctrl, cmd);
	  break;

	case IDE_FLOPPY:
	  zip_dorequest (ide, ctrl, cmd);
	  break;

	default:
	  printk ("ide_docmd: cannot handle such requests yet...\n");
	  ide_end_request (cmd, 1);
	  break;
	}
    }
  else
    ide_execcmd (cmd);

  return 0;
}

void
dump_queue (char *str, IDECTRL * ctrl)
{
  IDECMD *tmp;

  printk ("(%s) %s queue:", str, ctrl->name);

  tmp = ctrl->head;
  while (tmp != NULL)
    {
      printk (" 0x%x", tmp);
      tmp = tmp->next;
    }
  printk ("\n");
}

void
ide_end_request (IDECMD * cmd, int fl_error)
{
  REQUEST *req = cmd->req;
  IDECTRL *ctrl = cmd->ctrl;

  if (cmd == NULL)
    {
      printk ("ide_end_request: NULL cmd given\n");
      return;
    }

  if (fl_error)
    {
#ifdef IDEDEBUG
      printk ("ide_end_request: %s  error reported, retries=%d\n",
	      ctrl->name, cmd->retries);
#endif
      if (cmd->retries > 0)
	{
	  cmd->retries--;
	  cmd->count = cmd->orig_count;
	  cmd->buff = cmd->orig_buff;

	  ide_docmd (cmd->ide, ctrl, cmd);

	  return;
	}
    }

  dump_queue ("ide_end_request", ctrl);

  if (cmd != ctrl->head)
    {
      printk ("ide_end_request: panic, broken cmd queue..\n"
	      "(ctrl=%s, cmd=0x%x, ctrl->head=0x%x)\n", ctrl->name, cmd,
	      ctrl->head);
    }
  else
    {
      if (cmd->next == NULL)
	{
	  ctrl->head = NULL;
	  ctrl->tail = NULL;
	  ctrl->busy = 0;	/* set ctrl free */
	}
      else
	ctrl->head = cmd->next;
    }

  kfree (cmd);

  if (req != NULL)
    end_request (req->dev, fl_error);

  if (cmd = ctrl->head)		/* continue with next cmd */
    ide_docmd (cmd->ide, ctrl, cmd);
}

void
ide_enqueue_cmd (IDECMD * cmd, IDECTRL * ctrl)
{
  printk ("ide_enqueue_cmd: %s +0x%x\n", ctrl->name, cmd);

  cmd->next = NULL;

  single_code ();

  if (ctrl->head == NULL)
    {
      if (ctrl->tail != NULL)
	printk ("ide_enqueue_cmd: NULL head and non-NULL tail\n");

      ctrl->head = cmd;
      ctrl->tail = cmd;
    }
  else
    {
      if (ctrl->tail == NULL)
	{
	  multi_code ();
	  printk
	    ("ide_enqueue_cmd: broken queue on %s (head=0x%x, tail=NULL)\n",
	     ctrl->name, ctrl->head);
	  return;
	}

      ctrl->tail->next = cmd;
      ctrl->tail = cmd;
    }

  dump_queue ("ide_enqueue_cmd", ctrl);

  multi_code ();
}

void
ide_start_processing (IDECTRL * ctrl)
{
  IDECMD *cmd = ctrl->head;

  single_code ();
  if (cmd == NULL)
    {
      ctrl->busy = 0;
      multi_code ();
      printk ("ide_start_processing: NULL cmd\n");
      return;
    }

  if (!ctrl->busy)		/* idle ? */
    {
      ctrl->busy = 1;
      multi_code ();
      ide_docmd (cmd->ide, ctrl, cmd);
      return;
    }

  multi_code ();
}

void
ide_dorequest (REQUEST * req)
{
  int dev = req->dev;
  IDE *ide = ide_get_ide (dev);
  IDECTRL *ctrl = ide->ctrl;
  IDECMD *cmd;

  if ((cmd = ide_newcmd (ide, NULL)) == NULL)
    {
      end_request (dev, 1);
      return;
    }

  cmd->req = req;

#ifdef IDEDEBUG
  printk ("ide_dorequest: req=0x%x, cmd=0x%x, starting over..\n", req, cmd);
  printk (" %02X:%02X  blk=%d  nblk=%d\n", MAJOR (req->dev), MINOR (req->dev),
	  req->sec, req->nsec);
#endif

  ide_enqueue_cmd (cmd, ctrl);
  ide_start_processing (ctrl);
}

void
ide_runcmd (IDECMD * cmd)
{
  ide_enqueue_cmd (cmd, cmd->ctrl);
  ide_start_processing (cmd->ctrl);
}

int
ide_start_cmd (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  int port = ctrl->port;

  cmd->error = 0;
  cmd->status = 0;

  outb (ide->select_mask, port + AR_DRIVE);

  if (ide_wait (port, 0) < 0)
    {
      printk ("%s: controller not ready for cmd\n", ide->name);
      cmd->code = RES_NOTRDY;
      return -1;
    }

  /* Set up the controller registers. */
  outb (0, port + AR_FEATURES);
  outb (0, port + AR_IREASON);
  outb (0, port + AR_TAG);
  outb (cmd->count & 0xff, port + AR_CNTLO);
  outb (cmd->count >> 8, port + AR_CNTHI);
  outb (ATAPIC_PACKET, port + AR_COMMAND);

#ifdef IDEDEBUG
  printk ("%s: start\n", ide->name);
#endif

  return 0;
}

/* wait for DRQ */
int
ide_wait_cmd (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  /* Wait for DRQ from 100 usec to 3 msec for slow devices */
  int cnt = ide->fl_intrcmd ? 10000 : ide->fl_slow ? 3000 : 100;
  int ireason = 0, phase = 0, port = ctrl->port;

  /* Wait for command phase. */
  for (; cnt > 0; cnt -= 10)
    {
      ireason = inb (port + AR_IREASON);
      cmd->status = inb (port + AR_STATUS);
      phase = (ireason & (ARI_CMD | ARI_IN)) |
	(cmd->status & (ARS_DRQ | ARS_BUSY));

      if (phase == PHASE_CMDOUT)
	break;

      sys_sleep (10);
    }

  if (phase != PHASE_CMDOUT)
    {
      cmd->code = RES_NODRQ;
      cmd->error = inb (port + AR_ERROR);
      printk
	("%s: invalid command phase, ireason=0x%x, status=0x%x, error=0x%x\n",
	 ide->name, ireason, cmd->status, cmd->error);
      return -1;
    }

  return 0;
}

/*
 * Process the i/o phase, transferring the command/data to/from the device.
 * Return 1 if op continues, and we are waiting for new interrupt.
 * Return 0 when idle.
 */
int
ide_io (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  int port;
  u16 len, i;
  u8 ireason;

  port = ctrl->port;

  if (ide_wait (port, 0) < 0)
    {
      cmd->status = inb (port + AR_STATUS);
      cmd->error = inb (port + AR_ERROR);
      cmd->code = RES_NOTRDY;
      printk ("%s: controller not ready, status=0x%x, error=0x%x\n",
	      ide->name, cmd->status, cmd->error);
      return 0;
    }

  cmd->status = inb (port + AR_STATUS);
  cmd->error = inb (port + AR_ERROR);
  len = inb (port + AR_CNTLO);
  len |= inb (port + AR_CNTHI) << 8;
  ireason = inb (port + AR_IREASON);

#ifdef IDEDEBUG
  printk ("ide_io: dev=%s, ireason=0x%x, len=%d, status=0x%x, error=0x%x\n",
	  ide->name, ireason, len, cmd->status, cmd->error);
#endif
  switch ((ireason & (ARI_CMD | ARI_IN)) | (cmd->status & ARS_DRQ))
    {
    default:
      printk ("ide_io: unknown phase\n");
      cmd->code = RES_ERR;
      break;

    case PHASE_CMDOUT:
#ifdef IDEDEBUG
      printk ("ide_io: PHASE_CMDOUT\n");
#endif
      /* Send packet command. */
      if (!(cmd->status & ARS_DRQ))
	{
	  printk ("%s: no cmd drq\n", ide->name);
	  cmd->code = RES_NODRQ;
	  break;
	}

      /* Send packet command. */
      outsw (port + AR_DATA, cmd->pkt_data, 6);

      return 1;

    case PHASE_DATAOUT:
#ifdef IDEDEBUG
      printk ("ide_io: PHASE_DATAOUT\n");
#endif

      /* Write data */
      if (cmd->count > 0)
	{
	  printk ("%s: invalid data direction\n", ide->name);
	  cmd->code = RES_INVDIR;
	  break;
	}
      if (-cmd->count < len)
	{
	  printk ("%s: send data underrun, %d bytes left\n",
		  ide->name, -cmd->count);
	  cmd->code = RES_UNDERRUN;
	  outsw (port + AR_DATA, cmd->buff, -cmd->count >> 1);
	  for (i = -cmd->count; i < len; i += 2)
	    outw (0, port + AR_DATA);
	}
      else
	outsw (port + AR_DATA, cmd->buff, len >> 1);

      cmd->buff += len;
      cmd->count += len;
      return 1;

    case PHASE_DATAIN:
#ifdef IDEDEBUG
      printk ("ide_io: PHASE_DATAIN\n");
#endif

      /* Read data */
      if (cmd->count < 0)
	{
	  printk ("%s: invalid data direction\n", ide->name);
	  cmd->code = RES_INVDIR;
	  break;
	}
      if (cmd->count < len)
	{
	  printk ("%s: recv data overrun, %d bytes left\n",
		  ide->name, cmd->count);
	  cmd->code = RES_OVERRUN;
	  insw (port + AR_DATA, cmd->buff, cmd->count >> 1);
	  for (i = cmd->count; i < len; i += 2)
	    inw (port + AR_DATA);
	}
      else
	insw (port + AR_DATA, cmd->buff, len >> 1);

      cmd->buff += len;
      cmd->count -= len;
      return 1;

    case PHASE_ABORTED:
#ifdef IDEDEBUG
      printk ("ide_io: PHASE_ABORTED\n");
#endif
    case PHASE_COMPLETED:
#ifdef IDEDEBUG
      printk ("ide_io: PHASE_COMPLETED\n");
#endif
      if (cmd->status & (ARS_CHECK | ARS_DF))
	{
#ifdef IDEDEBUG
	  printk ("ide_io: setting RES_ERR\n");
#endif
	  cmd->code = RES_ERR;
	}
      else if (cmd->count < 0)
	{
	  printk ("%s: send data overrun, %d bytes left\n",
		  ide->name, -cmd->count);
	  cmd->code = RES_OVERRUN;
	}
      else if (cmd->count > 0)
	{
	  printk ("%s: recv data underrun, %d bytes left\n",
		  ide->name, cmd->count);
	  cmd->code = RES_UNDERRUN;
	  bzero (cmd->buff, cmd->count);
	}
      else
	{
	  cmd->code = RES_OK;
#ifdef IDEDEBUG
	  printk ("%s: setting RES_OK\n", ide->name);
#endif
	}
      break;
    }

  return 0;
}

void
ide_issue_immediate (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  int cnt, port = ctrl->port;

  /* Start packet command, wait for DRQ. */
  if (ide_start_cmd (ide, ctrl, cmd))
    return;

  if (ide_wait_cmd (ide, ctrl, cmd))
    return;

  /* Send packet command. */
  outsw (port + AR_DATA, cmd->pkt_data, 6);

  /* Wait for data i/o phase. */
  for (cnt = 20000; cnt > 0; cnt--)
    if (((inb (port + AR_IREASON) & (ARI_CMD | ARI_IN)) |
	 (inb (port + AR_STATUS) & ARS_DRQ)) != PHASE_CMDOUT)
      break;

  /* Do all needed i/o. */
  while (ide_io (ide, ctrl, cmd))
    {				/* Wait for DRQ deassert. */
      for (cnt = 2000; cnt > 0; --cnt)
	{
	  if (!(inb (port + AR_STATUS) & ARS_DRQ))
	    break;
	  sys_sleep (10);
	}
    }
}

/* RETURN VALUE: 0 .. command started, interrupt expected */
int
ide_issue (IDE * ide, IDECTRL * ctrl, IDECMD * cmd)
{
  int ret;

#ifdef IDEDEBUG
  printk ("ide_issue: starting command\n");
#endif

  if (ret = ide_start_cmd (ide, ctrl, cmd))
    return ret;

  if (ide->fl_intrcmd)		/* Wait for interrupt before sending packet command */
    return 0;

  if (ret = ide_wait_cmd (ide, ctrl, cmd))
    return ret;

#ifdef IDEDEBUG
  printk ("ide_issue: sending packet data\n");
#endif

  /* Send packet command. */
  outsw (ctrl->port + AR_DATA, cmd->pkt_data, 6);
  return 0;
}

IDECMD *
ide_newcmd (IDE * ide, void (*callback) (IDECMD *))
{
  IDECMD *tmp;

  tmp = (IDECMD *) kmalloc (sizeof (IDECMD), 0);

  if (tmp == NULL)
    {
      printk ("ide_newcmd: insufficient memory\n");
      return NULL;
    }

  bzero (tmp, sizeof (IDECMD));
  tmp->ide = ide;
  tmp->ctrl = ide->ctrl;
  tmp->callback = callback;

  return tmp;
}
