/*
   dev/mixer.c

   Phantom sound mixer.
 */

#include <phantom/fs.h>
#include <phantom/major.h>
#include <phantom/audio.h>
#include <phantom/audio_io.h>
#include <asm/timer.h>
#include <errno.h>
#include <conio.h>

#define MAJOR_NR	SOUND_MAJOR
#define NR_VOICE	32	/* number of voices to mix */

struct voice_st voice[NR_VOICE];
unsigned char *mixer_buffer;
int mixer_buffer_pos = 0;
int dev_buffer_pos = 0;
int buffer_len = 0;
int mixer_step = 0;		/* step bytes */
int mixer_freq = 0;

extern struct audio_driver sb_driver;
struct audio_driver *audiodriver = &sb_driver;

int
mixer_write (struct file *file, char *buff, size_t count)
{
  int voice_nr;

  voice_nr = MINOR (file->f_dev);
  if (voice_nr > NR_VOICE)
    return (-EIO);

  voice[voice_nr].data = buff;
  voice[voice_nr].size = count;
  voice[voice_nr].pos = 0;
  voice[voice_nr].div = (voice[voice_nr].freq > mixer_freq) ? voice[voice_nr].freq : mixer_freq;

  return (count);
}

int
mixer_ioctl (struct file *file, unsigned int cmd, unsigned int arg)
{
  int voice_nr;

  voice_nr = MINOR (file->f_dev);
  if (voice_nr > NR_VOICE)
    return (-EIO);

  switch (cmd)
    {
    case MIOC_SETFREQ:		/* set frequency of the voice */
      voice[voice_nr].freq = arg;
      voice[voice_nr].div = (arg > mixer_freq) ? arg : mixer_freq;
      break;
    case MIOC_SETMODE:		/* set mode, stereo, 8/16 bit, etc. */
      voice[voice_nr].mode = arg;
      break;
    case MIOC_SETLOOP:
      voice[voice_nr].loop = arg;
      break;
    case MIOC_PLAYING:
      if (!voice[voice_nr].loop && voice[voice_nr].pos == voice[voice_nr].size)
	return (0);
      return (1);
    }
  return (-EINVAL);
}

int
mixer_open (struct file *file)
{
  int voice_nr;

  for (voice_nr = 0; voice_nr < NR_VOICE; voice_nr++)
    if (voice[voice_nr].used == 0)
      break;

#ifdef MIXERDEBUG
  printf ("mixer.c: open: %d voice\n", voice_nr);
#endif

  if (voice_nr == NR_VOICE)
    return (-EIO);

  voice[voice_nr].freq = mixer_freq;
  voice[voice_nr].used = 1;
  file->f_dev = MKDEV (MAJOR_NR, voice_nr);

  return (0);
}

int
mixer_close (struct file *file)
{
  int voice_nr;

  voice_nr = MINOR (file->f_dev);
  if (voice_nr > NR_VOICE)
    return (-EINVAL);

  bzero ((void *) &voice[voice_nr], sizeof (struct voice_st));

  return (0);
}

struct file_operations mix_fops =
{
  NULL,				/* lseek */
  NULL,				/* read */
  mixer_write,
  NULL,				/* readdir */
  mixer_ioctl,
  mixer_open,
  mixer_close
};

void
mixer_timer ()
{
  int i, j, k, voices, todo, tmp;
  short mix_cache[MAXMIXSTEP];

  dev_buffer_pos = audiodriver->getpos ();

  bzero ((void *) mix_cache, sizeof (mix_cache));
  voices = 0;

  for (i = 0; i < NR_VOICE; i++)
    {
      if (voice[i].pos < voice[i].size)
	{
	  voices++;
	  if (voice[i].size - voice[i].pos >= mixer_step)
	    todo = mixer_step;
	  else
	    todo = voice[i].size - voice[i].pos;

	  k = voice[i].pos;
	  if (voice[i].freq > mixer_freq)
	    {
	      j = 0;
	      while (j < todo)
		{
		  voice[i].div -= mixer_freq;
		  if (voice[i].div <= 0)
		    {
		      voice[i].div += voice[i].freq;
		      mix_cache[j++] += voice[i].data[k++];
		    }
		  else
		    k++;
		  if (k >= voice[i].size)
		    j = todo;
		}
	    }
	  else if (voice[i].freq < mixer_freq)
	    {
	      for (j = 0; j < todo; j++)
		{
		  voice[i].div -= voice[i].freq;
		  if (voice[i].div <= 0)
		    {
		      voice[i].div += mixer_freq;
		      mix_cache[j] += voice[i].data[k];
		    }
		  mix_cache[j] += voice[i].data[k++];
		}
	    }
	  else
	    {
	      for (j = 0; j < todo; j++)
		mix_cache[j] += voice[i].data[k++];
	    }

	  voice[i].pos = k;
	  if (voice[i].loop && voice[i].pos >= voice[i].size)
	    voice[i].pos = 0;	/* fuck it off... */
	}
    }

/* copy the buffer */
  if (voices)
    {
      for (i = 0; i < mixer_step; i++)
	mixer_buffer[mixer_buffer_pos + i] = mix_cache[i] < 256 ? mix_cache[i] : 255;
    }
  else
    /* nothing to copy, silence */
    memset ((void *) (mixer_buffer + mixer_buffer_pos), 0x80, mixer_step);

  mixer_buffer_pos += mixer_step;

  if (mixer_buffer_pos >= buffer_len)
    mixer_buffer_pos = 0;
}

void
mixer_setup (int rate, int mode)
{
  mixer_freq = rate;
  mixer_step = rate / 50;

#ifdef MIXERDEBUG
  printf ("mixer_setup(%d, %d);\n", rate, mode, &mixer_buffer, &buffer_len);
  printf ("mixer_step=%d\n", mixer_step);
#endif


  if (audiodriver->setup (rate, mode, &mixer_buffer, &buffer_len))
    {
      printf ("audio device setup failed!\n");
      return;
    }

  add_timer (TIMER_MIXER, &mixer_timer, 20);	/* 50 Hz mixing freq. */

  mixer_buffer_pos = mixer_step * 3;

  bzero ((void *) voice, sizeof (struct voice_st) * NR_VOICE);
  memset ((void *) mixer_buffer, 0x80, 65536);
}

int
mixer_init ()
{
  int i;

  setattr (ATTR (GREEN, BLACK));
  printk ("Sound mixing engine\n");
  setattr (DEFAULT_ATTR);
  printk ("  driver='%s'\n", audiodriver->name);

  if (register_chrdev (MAJOR_NR, "snd", &mix_fops))
    {
      printf ("Cannot register mixer!\n");
      return (-1);
    }

  mixer_setup (11050, 0);
  return (0);
}

int 
mixer_done ()
{
  if (audiodriver->close)
    audiodriver->close ();

  return(0);
}
