#include <phantom/sched.h>
#include <phantom/tty.h>
#include <phantom/fs.h>
#include <phantom/major.h>
#include <errno.h>
#include <conio.h>
#include <termix.h>

#define ESCAPE			0x1b

#define E_NOTHING		0
#define E_MOVECURSOR		1
#define E_ATTR			6
#define E_TEXT			7
#define E_BG			8
#define E_TYPECHAR		9
#define E_SETCHAR		10
#define E_PUTATTR		11	/* excluded */
#define E_CURSOR		12
#define E_BASIC			-1
#define E_ESCAPE		-2

void
_setchar (TTY * t, int c)
{
  ((u16 *) t->screen)[t->cursor_x + t->cursor_y * t->size_x] =
    c | (t->attr << 8);

  if (t == current_tty)
    t->vdriver->tty->draw_char (t, c);
}

void
_updatecursor (TTY * t)
{
  if (t == current_tty && t->cursor_mode != TTY_NOCURSOR)
    t->vdriver->tty->set_cursor (t);
}

void
_shapecursor (TTY * t)
{
  if (t == current_tty)
    t->vdriver->tty->shape_cursor (t);
}

void
_gotoxy (TTY * t, int x, int y)
{
  if (x < 0)
    x = 0;
  else if (x >= t->size_x)
    x = t->size_x - 1;

  if (y < 0)
    y = 0;
  else if (y >= t->size_y)
    y = t->size_y - 1;

  t->cursor_x = x;
  t->cursor_y = y;

  _updatecursor (t);
}

void
_newline (TTY * t)
{
  int i;

  if (t->cursor_y < t->size_y - 1)
    _gotoxy (t, 0, t->cursor_y + 1);
  else
    {				/* scroll screen one line */
      memmove (t->screen, (void *) ((u32) t->screen + (t->size_x << 1)),
	       (t->size_x * (t->size_y - 1)) << 1);

      for (i = 0; i < t->size_x; i++)
	((u16 *) t->screen)[i + t->size_x * (t->size_y - 1)] =
	  (DEFAULT_ATTR << 8) | ' ';

      if (t == current_tty)
	t->vdriver->tty->scroll_screen (t, 1);

      _gotoxy (t, 0, t->size_y - 1);
    }
}

#define SHIFT_RIGHT		1
#define SHIFT_LEFT		(-1)

void
_shiftcursor (TTY * t, int direction)
{
  switch (direction)
    {
    case SHIFT_RIGHT:
      if (t->cursor_x == t->size_x - 1)
	_newline (t);
      else
	t->cursor_x++;
      break;

    case SHIFT_LEFT:
      if (t->cursor_x)
	t->cursor_x--;
      else if (t->cursor_y)
	{
	  t->cursor_y--;
	  t->cursor_x = t->size_x - 1;
	}
      break;
    }
}

void
_writechar (TTY * t, int c)
{
  switch (c)
    {
    case 0x0A:
      _newline (t);
      break;

    case 0x08:
      _shiftcursor (t, SHIFT_LEFT);
      _setchar (t, ' ');
      _updatecursor (t);
      break;

    default:
      _setchar (t, c);
      _shiftcursor (t, SHIFT_RIGHT);
      _updatecursor (t);
      break;
    }
}

void
_typechar (TTY * t, int c)
{
  _setchar (t, c);
  _shiftcursor (t, SHIFT_RIGHT);
  _updatecursor (t);
}

void
_clrscr (TTY * t)
{
  int i;

  for (i = 0; i < t->size_x * t->size_y; i++)
    ((u16 *) t->screen)[i] = (DEFAULT_ATTR << 8) | ' ';

  if (t == current_tty)
    t->vdriver->tty->clear_screen (t);

  _gotoxy (t, 0, 0);
}

/* FIXME: update cursor after whole string was processed */
void
term_send (TTY * t, char *s, int count)
{
  int c;

  while (count--)
    {
      c = *(unsigned char *) s;
      s++;

      switch (t->expect)
	{
	case E_NOTHING:
	  if (c == ESCAPE)
	    t->expect = E_BASIC;
	  else
	    _writechar (t, c);
	  break;

	case E_BASIC:
	  if (c == '[')
	    t->expect = E_ESCAPE;
	  else
	    _writechar (t, c);
	  break;

	case E_ESCAPE:		/* main escape sequence */
	  switch (c)
	    {
	    case 'A':		/* cursor up one character */
	      _gotoxy (t, t->cursor_x, t->cursor_y - 1);
	      t->expect = E_NOTHING;
	      break;
	    case 'B':		/* cursor down one character */
	      _gotoxy (t, t->cursor_x, t->cursor_y + 1);
	      t->expect = E_NOTHING;
	      break;
	    case 'D':		/* cursor left one character */
	      _gotoxy (t, t->cursor_x - 1, t->cursor_y);
	      t->expect = E_NOTHING;
	      break;
	    case 'C':		/* cursor right one character */
	      _gotoxy (t, t->cursor_x + 1, t->cursor_y);
	      t->expect = E_NOTHING;
	      break;
	    case 'c':		/* clear the screen and cursor home */
	      _clrscr (t);
	      t->expect = E_NOTHING;
	      break;
	    case 'm':		/* move cursor to [x,y] */
	      t->expect = E_MOVECURSOR;
	      t->expect2 = 0;
	      break;
	    case 'a':		/* change current attrib */
	      t->expect = E_ATTR;
	      break;
	    case 'b':		/* change background color */
	      t->expect = E_BG;
	      break;
	    case 't':		/* change text color */
	      t->expect = E_TEXT;
	      break;
	    case 'T':		/* type char */
	      t->expect = E_TYPECHAR;
	      break;
	    case 'S':
	      t->expect = E_SETCHAR;
	      break;
	    case 'f':		/* flush keyboard buffer */
	      t->kb_buff_i = 0;
	      break;
	    case 's':		/* show/hide cursor */
	      t->expect = E_CURSOR;
	      break;
	    default:		/* mistake */
	      t->expect = E_NOTHING;
	    }
	  break;

	case E_CURSOR:
	  switch (c)
	    {
	    case TTYCMD_SHOWCURSOR:
	      t->cursor_mode &= ~1;
	      _updatecursor (t);
	      break;
	    case TTYCMD_HIDECURSOR:
	      t->cursor_mode |= 1;
	      _updatecursor (t);
	      break;
	    case TTYCMD_SMALLCURSOR:
	      t->cursor_mode &= ~TTY_BIGCURSOR;
	      _shapecursor (t);
	      break;
	    case TTYCMD_BIGCURSOR:
	      t->cursor_mode |= TTY_BIGCURSOR;
	      _shapecursor (t);
	      break;
	    }
	  t->expect = E_NOTHING;
	  break;

	case E_MOVECURSOR:
	  t->buff[t->expect2] = c;
	  if (t->expect2 == 1)
	    {
	      t->expect = E_NOTHING;
	      _gotoxy (t, t->buff[0], t->buff[1]);
	    }
	  t->expect2++;
	  break;

	case E_ATTR:
	  t->attr = c;
	  t->expect = E_NOTHING;
	  break;
	case E_BG:
	  t->attr = (c << 4) | (t->attr & 0x0F);
	  t->expect = E_NOTHING;
	  break;
	case E_TEXT:
	  t->attr = (t->attr & 0xF0) | c;
	  t->expect = E_NOTHING;
	  break;
	case E_TYPECHAR:
	  _typechar (t, c);
	  t->expect = E_NOTHING;
	  break;
	case E_SETCHAR:
	  _setchar (t, c);
	  t->expect = E_NOTHING;
	  break;
	}
    }
}

/* rg/kernel specials :) */
int
_getch (TTY * t)
{
  int c, j;

  while (!t->kb_buff_i);

  c = t->kb_buff[0];
  t->kb_buff_i--;

  for (j = 1; j < t->kb_buff_i; j++)
    t->kb_buff[j - 1] = t->kb_buff[j];

  return c;
}

int
_kbhit (TTY * t)
{
  return t->kb_buff_i;
}

/* kernel specials */
void
putchar (int c)
{
  term_send (current_tty, (char *) &c, 1);
}

int
getch ()
{
  return _getch (current_tty);
}

int
kbhit ()
{
  return _kbhit (current_tty);
}
