/*
   dev/tty.c
   Terminal handling routines / hw part.
 */

#include <conio.h>
#include <phantom/vga.h>
#include <phantom/vgaregs.h>
#include <asm/io.h>
#include <phantom/tty.h>
#include <sys/ttyio.h>
#include <asm/sysinfo.h>
#include <asm/system.h>
#include <string.h>
#include <phantom/fs.h>
#include <phantom/sched.h>
#include <phantom/mm.h>
#include <phantom/major.h>
#include <phantom/video.h>
#include <phantom/k_unistd.h>	/* STDIN_FILENO, ... */
#include <phantom/wait.h>
#include <phantom/lock.h>

#define MAJOR_NR		TTY_MAJOR

TTY tty_table[NR_TTY] =
{
  {0,}
  ,
};				/* can be used only here... */

TTY *current_tty, *printk_tty, *rg_tty;
int nr_tty = 1;

int
tty_switch (int nr)
{
  TTY *t;
  u32 flg;
  int okb_flags;

  if (nr >= nr_tty)
    return -1;

  t = &tty_table[nr];
  if (t == current_tty)
    return 0;

  okb_flags = current_tty->kb_flags;

  get_flags (flg);
  cli ();

  t->vdriver->tty->switch_screen (current_tty, t);
  current_tty = t;

  set_flags (flg);

/* kb flags as ctrl, alt and shift must be overtaken */
  t->kb_flags &= ~(FL_SHIFT | FL_CTRL | FL_ALT);
  t->kb_flags |= (okb_flags & (FL_SHIFT | FL_CTRL | FL_ALT));

/* restore leds */
  kb_setleds (t->kb_leds);

  return 0;
}

int
tty_write (struct file *file, char *buff, size_t count)
{
  term_send (&tty_table[MINOR (file->f_dev)], buff, count);
  return count;
}

int
tty_read (struct file *file, char *buff, size_t count)
{
  TTY *t;
  int i, j, c;
  u32 flg;
  CODE_LOCK(a);
  WAIT w =
  {ctask, NULL};

  t = &tty_table[MINOR (file->f_dev)];

  i = 0;

  SINGLE_CODE(a);

  while (count)
    {
      if (!t->kb_buff_i)
	sleep_on (&t->key_wait, &w);

      c = buff[i++] = t->kb_buff[0];      

      for (j = 1; j < t->kb_buff_i; j++)
	t->kb_buff[j - 1] = t->kb_buff[j];

      t->kb_buff_i--;
    
      if(c=='\n')
        break;
    
      count--;
    }

  MULTI_CODE(a);

  return i;
}

/* FIXME: vnode and file could be locked for "/dev/console"... */
int
tty_open (struct file *file)
{
  int min = MINOR (file->f_dev);

  if (min == CURRENT_TTY_MINOR)	/* "/dev/console" case */
    {
      file->f_dev = MKDEV (TTY_MAJOR, ctask->tty);
      return 0;
    }

  if (min < 0 && min >= NR_TTY)
    return -EIO;

  return 0;
}

int
_readkey (TTY * t)
{
  int res = 0;
  u32 fl;
  WAIT w =
  {ctask, NULL};

  get_flags (fl);
  cli ();

  if (!t->kb_buff_i)
    sleep_on (&t->key_wait, &w);

  if (t->kb_buff_i)
    {
      res = (t->raw_buff[0] << 8) | t->kb_buff[0];
      memmove (t->kb_buff, t->kb_buff + 1, KB_BUFFSIZE - 1);
      memmove (t->raw_buff, t->raw_buff + 1, KB_BUFFSIZE - 1);
      t->kb_buff_i--;
    }
  set_flags (fl);

  return res;
}

int
_xkbhit (TTY * t)
{
  if (t->kb_buff_i)
    return 1;

  return 0;
}

/* we can only return positive values ... */
int
tty_ioctl (struct file *file, unsigned int cmd, unsigned int arg)
{
  int min;

  if (MAJOR (file->f_dev) != TTY_MAJOR)
    return (-ENOTTY);

  min = MINOR (file->f_dev);
  if (min > NR_TTY)
    return (-ENOTTY);

  switch (cmd)
    {
    case TTY_GETX:
      return (tty_table[min].size_x);

    case TTY_GETY:
      return (tty_table[min].size_y);

    case TTY_READKEY:
      return (_readkey (&tty_table[min]));

    case TTY_GETKEY:
      return (getkey (arg));

    case TTY_GETCURSOR:
      switch (arg)
	{
	case TTY_ACX:
	  return (tty_table[min].cursor_x);

	case TTY_ACY:
	  return (tty_table[min].cursor_y);
	}
      break;

    case TIOC_KBHIT:
      return (tty_table[min].kb_buff_i);
    }

  return (-EINVAL);
}

struct file_operations tty_operations =
{
  NULL,				/* lseek */
  tty_read,
  tty_write,
  NULL,				/* readdir */
  tty_ioctl,
  tty_open,
  NULL,				/* release */
};

/* this sets up first (and the only one virtual console) which is used
   by other inits (such as mm_init...) */
void
tty_init ()
{
  int i;
  TTY *t = &tty_table[0];

#ifdef FORCE_COLOR
  sysinfo.mono = 0;
#elif FORCE_MONO
  sysinfo.mono = 1;
#endif

  t->size_x = 80;
#ifndef VGA80x50
  t->size_y = 25;
#else
  t->size_y = 50;
#endif

  t->attr = DEFAULT_ATTR;
  t->expect = 0;
  t->expect2 = 0;
  t->kb_leds = DEFAULT_KBLEDS;
  t->kb_flags = DEFAULT_KBFLAGS;
  t->kb_buff_i = 0;
  t->cursor_mode = TTY_SMALLCURSOR;

  t->vdriver = vdriver;
  t->screen = vdriver->screen_base;

  t->cursor_x = 0;
  t->cursor_y = 0;

  current_tty = t;
  printk_tty = t;

  clrscr ();

  /* all should be zero... *//* user purposes only... */
  file_table[0].f_dev = MKDEV (MAJOR_NR, 0);
  file_table[0].f_op = &tty_operations;
  file_table[0].f_used = 1;
  file_table[1].f_dev = MKDEV (MAJOR_NR, 0);
  file_table[1].f_op = &tty_operations;
  file_table[1].f_used = 1;
  file_table[2].f_dev = MKDEV (MAJOR_NR, 0);
  file_table[2].f_op = &tty_operations;
  file_table[2].f_used = 1;
  ctask->filp[STDIN_FILENO] = &file_table[0];
  ctask->filp[STDOUT_FILENO] = &file_table[1];
  ctask->filp[STDERR_FILENO] = &file_table[2];

  register_chrdev (TTY_MAJOR, "tty", &tty_operations);

  setattr (ATTR (GREEN, BLACK));
  printf ("TTY initialization stage I (%dx%d)\n", t->size_x, t->size_y);
  setattr (DEFAULT_ATTR);
}

/* memory management should be working at this point... */

void
tty_init2 ()
{
/* register 'old' first console */
  dfs_add ("console", MKDEV (TTY_MAJOR, CURRENT_TTY_MINOR),
	   S_IRWXU | S_IFCHR);
  dfs_add ("tty0", MKDEV (TTY_MAJOR, 0), S_IRWXU | S_IFCHR);
/* oughta be all */
}

/* 
   this will set up virtual consoles -- this must be runned AFTER mm_init 
   but it does not need to be called, here's surely one working
   console   
 */
void
tty_init3 ()
{
  int i;
  char *banner = "Welcome at PHANTOM's, console %d\n";
  char str_tmp[64];
  char *dev_name = "tty%c";
  TTY *t = &tty_table[0];
  TTY *y;

  setattr (ATTR (GREEN, BLACK));
  printf ("TTY initialization stage ]I[\n");
  setattr (DEFAULT_ATTR);

/* initialize first console screen to be able to switch */
/* t->screen = (char *) kmalloc ((t->size_x * t->size_y) << 2, PG_KERNEL); */
#ifdef VGA80x50
  t->vdriver->backup_screen = (char *) pages_alloc (2);
#else
  t->vdriver->backup_screen = (char *) page_alloc ();
#endif

  for (i = 1; i < NR_TTY; i++)
    {
      y = &tty_table[i];

      y->cursor_mode = TTY_SMALLCURSOR;
      y->size_x = t->size_x;
      y->size_y = t->size_y;
      y->expect = 0;
      y->expect2 = 0;
      y->attr = DEFAULT_ATTR;
      y->vdriver = vdriver;	/* global vdriver */

      y->kb_buff_i = 0;		/* empty buffer */
      y->kb_leds = DEFAULT_KBLEDS;
      y->kb_flags = DEFAULT_KBFLAGS;

#ifdef VGA80x50
      y->screen =		/*(char *) kmalloc ((y->size_x * y->size_y) << 2, PG_KERNEL); */
	(char *) pages_alloc (2);
#else
      y->screen =		/*(char *) kmalloc ((y->size_x * y->size_y) << 2, PG_KERNEL); */
	(char *) page_alloc ();
#endif
      sprintf (str_tmp, banner, i);
      _clrscr (y);		/* set the cursor */
      term_send (y, str_tmp, strlen (str_tmp));

      sprintf (str_tmp, dev_name, i + '0');
      dfs_add (str_tmp, MKDEV (TTY_MAJOR, i), S_IRWXU | S_IFCHR);
    }

  printk_tty = &tty_table[DEFAULT_PRINTK_TTY];
  rg_tty = &tty_table[DEFAULT_RG_TTY];

  nr_tty = NR_TTY;
}
