/*
   fs/dfs/vnode.c
 */

#include <errno.h>
#include <phantom/fs.h>
#include <phantom/blk.h>
#include <phantom/mm.h>
#include <phantom/dfs.h>
#include <string.h>

extern struct super_block *task_super;
extern struct vnode_operations task_vnode_operations;
extern struct file_operations task_file_operations;

#define DFS struct dfs_descriptor

extern struct super_block *dfs_super;

DFS
{
  char name[MAX_DFS_NAME_LEN + 1];
  int dev;			/* device number */
  int mode;			/* file mode */
  DFS *next;
};

#define I_TASK	2
#define I_SYS	3

DFS dfs_sys =
{"sys", 0, S_IFDIR, NULL};	/* 3 */
DFS dfs_task =
{"task", 0, S_IFDIR, &dfs_sys};	/* 2 */
DFS dfs_dot =
{".", 0, S_IFDIR, &dfs_task};	/* 1 */
DFS dfs_dotdot =
{"..", 0, S_IFDIR, &dfs_dot};	/* 0 */

DFS *dfs_queue = &dfs_dotdot;
DFS *dfs_tail = &dfs_sys;

int dfs_lookup (struct vnode **, struct vnode *, char *, int);
int dfs_readdir (struct file *, struct dirent *, int);

/* WRITEME: handle rename and erase maybe... (add chmod generally) */
struct vnode_operations dfs_vnode_operations =
{
  NULL,				/* create */
  NULL,				/* erase */
  dfs_lookup,			/* lookup */
  NULL,				/* rename */
  NULL,				/* truncate */
  NULL,				/* mkdir */
  NULL				/* rmdir */
};

struct file_operations dfs_file_operations =
{
  NULL,				/* default seek */
  NULL,				/* read */
  NULL,				/* write */
  dfs_readdir,			/* readdir */
  NULL,				/* ioctl */
  NULL,				/* open */
  NULL				/* close */
};

int
dfs_lookup (struct vnode **dst, struct vnode *src, char *name, int len)
{
  int nr;
  DFS *dfs = dfs_queue;

  if (!len || len > MAX_DFS_NAME_LEN)
    {
#ifdef DFSDEBUG
      printk ("dfs_lookup: invalid len\n");
#endif
      return -ENOENT;
    }

  nr = 0;
  while (dfs && strncmp (dfs->name, name, MAX_DFS_NAME_LEN))
    {
      dfs = dfs->next;
      nr++;
    }

  if (!dfs)
    return -ENOENT;

#ifdef DFSDEBUG
  printk ("dfs_lookup: found \"%s\", device: %02X:%02X  mode: 0x%x\n",
	  name, MAJOR (dfs->dev), MINOR (dfs->dev), dfs->mode);
#endif

  *dst = vnode_get (src->v_sb, nr);

  return 0;
}

DFS * dfs_getenth (int nr)
{
  int i;
  DFS *d = dfs_queue;

  i = 0;
  while (i < nr && d)
    {
      d = d->next;
      i++;
    }

#ifdef DFSDEBUG
  if (!d)
    printk ("dfs_getenth: fatal, entry not found, nr=%d\n", i);
#endif

  return d;
}

int
dfs_readdir (struct file *file, struct dirent *dirent, int count)
{
  int i;
  DFS *d;

#ifdef DFSDEBUG
  printk ("dfs_readdir: vnode=0x%p  file=0x%p  dirent=0x%p  count=%d\n",
	  vnode, file, dirent, count);
  printk ("  file->f_pos=%d\n", file->f_pos);
#endif

  d = dfs_getenth (file->f_pos);
  if (!d)
    return 0;

  dirent->d_reclen = strlen (d->name);
  strcpy (dirent->d_name, d->name);
  file->f_pos++;

  return 1;
}

int
dfs_add (char *name, int dev, int mode)
{
  DFS *dfs;

#ifdef DFSDEBUG
  printk ("dfs_add: name %s  dev %02X:%02X  mode 0x%x\n",
	  name, MAJOR (dev), MINOR (dev), mode);
#endif

  dfs = dfs_tail;
  while (dfs)
    {
      if (!strcmp (dfs->name, name))
	{
	  printf ("dfs_add: duplicate device %s\n", name);
	  return -EBUSY;
	}
      dfs = dfs->next;
    }

  dfs = (struct dfs_descriptor *)
    kmalloc (sizeof (struct dfs_descriptor), PG_KERNEL);

  if (!dfs)
    {
      printf ("dfs_add: failed, oom\n");
      return -ENOMEM;
    }

  strncpy (dfs->name, name, MAX_DFS_NAME_LEN);
  dfs->dev = dev;
  dfs->mode = mode;

  dfs->next = dfs_tail->next;
  dfs_tail->next = dfs;

  return 0;
}

int
dfs_del (int dev)
{
  DFS *dfs;

#ifdef DFSDEBUG
  printk ("dfs_del: removing device %02X:%02X\n", MAJOR (dev), MINOR (dev));
#endif

  dfs = dfs_queue;
  if (dfs_queue->dev == dev)
    {
      dfs_queue = dfs_queue->next;
      kfree (dfs);
      return 0;
    }

  while (dfs->next && dfs->next->dev != dev)
    dfs = dfs->next;

  if (!dfs->next)
    {
      printk ("dfs_del: device %02X:%02X not found\n",
	      MAJOR (dev), MINOR (dev));
      return -ENOENT;
    }

  dfs->next = dfs->next->next;
  kfree (dfs->next);
  return 0;
}

#define DEVMATCH(a,b)	((MAJOR(a) == MAJOR(b)) && MINOR(b))

int
dfs_removechilds (int dev)
{
  int count = 0;
  DFS *dfs, *tmp;

  printk ("dfs_removechilds (%02X:%02X):", MAJOR (dev), MINOR (dev));

  if (DEVMATCH (dev, dfs_queue->dev))
    {
      printk (" %s", dfs_queue->name);
      dfs_queue = dfs_queue->next;
      kfree (dfs);
      count++;
    }

  dfs = dfs_queue;

  while (dfs->next)
    {
      if (DEVMATCH (dev, dfs->next->dev))
	{
	  printk (" %s", dfs->name);
	  tmp = dfs->next;
	  dfs->next = dfs->next->next;
	  kfree (tmp);
	  count++;
	}

      dfs = dfs->next;
    }

  if (!count)
    printk ("nothing found");

  printk ("\n");

  return 0;
}

int
dfs_read_vnode (struct vnode *vnode)
{
  DFS *dfs;

#ifdef DFSDEBUG
  printk ("dfs_read_vnode: vnode=0x%x  v_num=%d\n", vnode, vnode->v_num);
#endif
  vnode->v_atime = 0;
  vnode->v_ctime = 0;
  vnode->v_mtime = 0;

  if (vnode->v_num == DFS_ROOT_VNODE_NUMBER)
    {
      vnode->v_mode =
	S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH | S_IFDIR;
      vnode->f_op = &dfs_file_operations;
      vnode->v_op = &dfs_vnode_operations;

      return 0;
    }

  dfs = dfs_getenth (vnode->v_num);
  if (!dfs)
    {
      printf ("dfs_read_vnode: getenth failed, v_num=%d\n", vnode->v_num);
      return -EIO;
    }

  vnode->v_mode = dfs->mode;
  vnode->v_op = &dfs_vnode_operations;

  if (S_ISBLK (dfs->mode))	/* block device */
    vnode->f_op = blkdev[MAJOR (dfs->dev)].fops;
  else if (S_ISCHR (dfs->mode))
    vnode->f_op = chrdev[MAJOR (dfs->dev)].fops;
  else if (S_ISDIR (dfs->mode))
    {
      switch (vnode->v_num)
	{
	case I_TASK:
/*	  vnode->v_op = &task_vnode_operations;*/
/*	  vnode->f_op = &task_file_operations;*/
	  vnode->v_mount = task_super->s_mount;
	  break;

	case I_SYS:
	  break;
	}
    }
  else
    printk ("dfs_readvnode: unknown mode 0x%x for %s\n",
	    dfs->name);

/* WARNING: dev number override but everything should be still ok */
  vnode->v_dev = dfs->dev;

#ifdef DFSDEBUG
  printk ("  dev=0x%x  mode=0x%x\n", vnode->v_dev, vnode->v_mode);
#endif
  return 0;
}

char *
dfs_getdevname (int dev)
{
  struct dfs_descriptor *dfs = dfs_queue;

  printf ("&dfs_queue = 0x%p  dfs_queue = 0x%p\n", &dfs_queue, dfs_queue);
  while (dfs)
    {
      if (dfs->dev == dev)
	{
	  printf ("dgdn: dfs_queue at 0x%p\n", dfs_queue);
	  printf ("      dfs->name at 0x%p\n", dfs->name);
	  return dfs->name;
	}
      dfs = dfs->next;
    }
  return NULL;
}
