/*
   fs/exec.c
   Process loader.
 */

#include <phantom/mm.h>
#include <phantom/fs.h>
#include <errno.h>
#include <sys/types.h>
#include <phantom/sched.h>

#define PHANTOM_EXE_HEADER	"code"
#define PHANTOM_EXE_HEADER_LEN	4

/* phantom executable file header */
struct p_code
  {
    char id[PHANTOM_EXE_HEADER_LEN];
    u32 header_len;
    u32 text_end;
    u32 end;
  };

int sys_exec (const char *name, char **argv, char **envp);

/* tell whether it is an executable and obtain info */
int
read_header (char *fname, TASK *task)
{
  int ret;
  struct vnode *vnode;
  struct bhead *bh;
  struct file file =
  {0,};
  struct p_code header;

#ifdef TASKDEBUG
  printk ("read_header: exploring %s\n", fname);
#endif

  if (ret = vnode_open (&vnode, fname, 0))
    {
      printk ("read_header: vnode_open failed, ret=%d (%s)!\n", 
      ret, translate_error(ret));
      return ret;
    }

  if (!S_ISREG (vnode->v_mode))
    {
      printk ("read_header: only regular files can be executed\n");
      return -EINVAL;
    }

  if (!vnode->v_op || !vnode->f_op->read)
    {
      printk ("read_header: cannot read exec, no read operation defined\n");
      return -EIO;
    }

  file.f_vnode = vnode;
  file.f_op = vnode->f_op;
  file.f_used = 1;
  file.f_dev = vnode->v_dev;

  if (file.f_op->read (&file, (char *) &header, sizeof (header)) !=
      sizeof (header))
    {
      printk ("get_exec_info: read failed\n");
      return -EIO;
    }

  if (strncmp (header.id, PHANTOM_EXE_HEADER, PHANTOM_EXE_HEADER_LEN))
    {
      printk ("read_header: invalid header\n");
      return -EINVAL;
    }

  task->vnode = vnode;
  task->code_pos = 0x40000000 + header.header_len;
  task->code_end = header.text_end;
  task->data_end = header.end;

  task->code_pages = TOPAGE (task->code_end - 0x40000000 + PAGESIZE - 1);
  task->nr_pages = TOPAGE (task->data_end - task->code_pos + PAGESIZE - 1);

#ifdef CODELOCKDEBUG
  printk ("read_header: code_pages=%d  code_pos=0x%x  code_end=0x%x\n",
	  task->code_pages, task->code_pos, task->code_end);
#endif

#ifdef TASKDEBUG
  printk ("read_header: vnode=0x%p  filesize=%d  pages=%d\n",
	  vnode, vnode->v_size, task->nr_pages);
#endif

  return 0;
}

int
read_exec (struct task *task)
{
  u32 *pt;
  int todo, now;
  int entry, table;
  struct file file =
  {0,};
  struct vnode *vnode = task->vnode;

  file.f_vnode = vnode;
  file.f_op = vnode->f_op;
  file.f_used = 1;
  file.f_dev = vnode->v_dev;

  entry = 0;
  table = 0x100;
  pt = CPT (task, 0x100);
  todo = vnode->v_size;
#ifdef TASKDEBUG
  printk ("read_exec: pt=0x%x  todo=%d\n", pt, todo);
#endif
  while (todo)
    {
      if (!pt || !TOPAGE (pt[entry]))
	{
	  printk ("read_exec: ");
	  if (!pt)
	    printk ("pt = NULL, tbl=0x%x\n", table);
	  else
	    printk ("pt[%d] = NULL, tbl=0x%x\n", entry, table);

	  return -EINVAL;
	}

      now = MIN (PAGE_SIZE, todo);
#ifdef TASKDEBUG
      printk ("  reading %d bytes, addr=0x%x\n", now, pt[entry] & ADDR_MASK);
#endif
      file.f_op->read (&file, (char *) (pt[entry] & ADDR_MASK), now);
      todo -= now;
      entry++;
      if (entry == 1024)
	{
	  entry = 0;
	  table++;
	  pt = CPT (task, table);
	}
    }
  return 0;
}
