/*
   fs/fat/rw.c

   Fat read/write routines.
 */

#include <phantom/fs.h>
#include <phantom/fat.h>
#include <errno.h>
#include <phantom/buffer.h>

int fat_read (struct file *, char *, size_t);
int fat_write (struct file *file, char *buff, size_t count);
int fat_readdir (struct file *file, struct dirent *dirent, int count);
int fat_rawrw (int cmd, struct vnode *vnode, struct file *file, char *buff,
	       size_t count);

struct file_operations fat_file_operations = {
  NULL,				/* default seek */
  fat_read,
  fat_write,
  fat_readdir,
  NULL,				/* ioctl */
  NULL,				/* open */
  NULL				/* close */
};


int
fat_read (struct file *file, char *buff, size_t count)
{
  return (fat_rawrw (FAT_READ, file->f_vnode, file, buff, count));
}

int
fat_write (struct file *file, char *buff, size_t count)
{
  struct vnode *vnode = file->f_vnode;
  struct super_block *sb = vnode->v_sb;
  struct diritem *di;
  int newsize, oldsize, clusters;
  int clustersize, cluster, sec, subsec, newcluster, i;
  struct bhead *buffer;

  if (!count)
    return (0);

  newsize = count + file->f_pos;
  if (newsize > vnode->v_size)	/* enlarging? */
    {
      oldsize = vnode->v_size;
      cluster = vnode->u.fat.cluster;
      newcluster = cluster;
      clustersize = sb->u.fat.clusterblocks * sb->s_blocksize;

      if (!oldsize)		/* we have to reserve one cluster... */
	{
	  for (i = 0; i < sb->u.fat.fatentries; i++)
	    if (!fat_getfatcell (sb, i))
	      break;
	  newcluster = i;
	  cluster = newcluster;
	  fat_setfatcell (sb, i, FAT_EOF);
	}

#ifdef FATDEBUG
      printf ("fat_write(): writing behind the file\n");
      printf ("  oldsize=%d  newsize=%d\n", oldsize, newsize);
      printf ("  clustersize=%d\n", clustersize);
#endif

      clusters = newsize / clustersize - oldsize / clustersize;
      if (clusters)
	{
#ifdef FATDEBUG
	  printf ("  reserving %d clusters\n", clusters);
#endif
	  if (fat_cluster (sb, &cluster, cluster, oldsize / clustersize))
	    return (-EIO);

#ifdef FATDEBUG
	  printf ("  last file cluster=%d\n", cluster);
#endif
	  cluster = fat_reserveclusters (sb, cluster, clusters);
#ifdef FATDEBUG
	  printf ("  %d cluster(s) reserved of %d\n", cluster, clusters);
#endif

/* FIXME: reduce newsize (if lowspace?) */
/* fat table will be written with umount */
	}
#ifdef FATDEBUG
      else
	{
	  printf ("  no cluster reservation needed\n");
	}
#endif
/* changing file attributes */

      sec = VNUM2SEC (vnode->v_num);
      subsec = VNUM2SUBSEC (vnode->v_num);

      buffer = bread (sb->s_dev, sec);
      block (buffer);
      buffer->fl_dirty = 1;

      di = (struct diritem *) buffer->data + subsec;
      vnode->v_size = newsize;
      vnode->u.fat.cluster = newcluster;
      di->cluster = newcluster;
      di->size = newsize;

      bunlock (buffer);
      brelse (buffer);
    }

  return (fat_rawrw (FAT_WRITE, vnode, file, buff, count));
}

int
fat_readdir (struct file *file, struct dirent *dirent, int count)
{
  struct vnode *vnode = file->f_vnode;
  struct diritem di;

#ifdef F32DEBUG
  int i;
#endif

  file->f_pos &= ~ITEMSIZEMASK;

  do
    {
      if (fat_rawrw (FAT_READ | FAT_DIR, vnode, file, (char *) &di, ITEMSIZE)
	  != ITEMSIZE)
	return (0);
      if (!di.name[0])
	{
#ifdef F32DEBUG
	  printk ("!di.name[0]\n");
	  for (i = 0; i < 11; i++)
	    putchar (di.name[i]);
#endif
	  file->f_pos -= ITEMSIZE;	/* seek it back */
	  return (0);
	}
    }
  while (di.name[0] == 0xE5
	 || (!FAT_ISDIR (di.attr) && !FAT_ISREG (di.attr)));

  fat_parse (dirent->d_name, di.name);
  dirent->d_reclen = strlen (di.name);

  return (1);
}

int
fat_rawrw (int cmd, struct vnode *vnode, struct file *file, char *buff,
	   size_t count)
{
  struct super_block *sb = vnode->v_sb;
  int sec, dev, tmp0, tmp1, tcount, cluster, block, blocksize;
  struct bhead *buffer;

  if (file->f_pos + count > vnode->v_size)
    count = vnode->v_size - file->f_pos;

  if (count <= 0)
    return (0);

  blocksize = sb->s_blocksize;
  cluster = vnode->u.fat.cluster;

#ifdef F32DEBUG
  printk ("frw: blocksz=%d  clust=%d  vnode=0x%x\n",
	  blocksize, cluster, vnode);
#endif

  if (!cluster)
    {
      sec = file->f_pos / blocksize;
#ifdef FATDEBUG
      printf ("fat_rawrw(): root operation, sector %d\n", sec);
#endif
    }
  else
    {
/* zero-length files cannot get here... */
      if (fat_blockseek (sb, &cluster, &sec, cluster, 0,
			 file->f_pos / blocksize))
	{
#ifdef F32DEBUG
	  printk ("blockseekfails...\n");
#endif
	  return (-EIO);
	}

#ifdef FATDEBUG
      printf ("fat_rawrw(): sec=%d, cluster=%d\n", sec, cluster);
#endif
    }

  cmd &= FAT_CMDMASK;
  dev = vnode->v_dev;

  tcount = count;

  tmp0 = file->f_pos % blocksize;
  if (tmp0)
    {
#ifdef FATDEBUG
      printf ("fat_rawrw(): head ahead!\n");
#endif

      tmp1 = blocksize - tmp0;
      if (tmp1 > tcount)
	tmp1 = tcount;

      block = fat_toblock (sb, cluster) + sec;
#ifdef F32DEBUG
      printk ("blk=%d\n", block);
#endif
      buffer = bread (dev, block);
      if (buffer == NULL)
	return count - tcount;

      switch (cmd)
	{
	case FAT_READ:
	  memcpy ((void *) buff, (void *) buffer->data + tmp0, tmp1);
	  break;
	case FAT_WRITE:
	  memcpy ((void *) buffer->data + tmp0, (void *) buff, tmp1);
	  buffer->fl_dirty = NORMAL_DIRTYNESS;
	  break;
	}

      brelse (buffer);
      file->f_pos += tmp1;
      tcount -= tmp1;
      if (!tcount)		/* short head */
	return (count);
      buff += tmp1;
      if (fat_blockseek (sb, &cluster, &sec, cluster, sec, 1))
	return (-EIO);
    }

  tmp0 = tcount / blocksize;
  if (tmp0)
    {
#ifdef FATDEBUG
      printf ("fat_rawrw: main io: ");
#endif

      while (tmp0--)
	{
	  block = fat_toblock (sb, cluster) + sec;
#ifdef F32DEBUG
	  printk ("blk=%d\n", block);
#endif

	  switch (cmd)
	    {
	    case FAT_READ:
#ifdef FATDEBUG
	      putchar ('r');
#endif
	      buffer = bread (dev, block);
	      if (buffer == NULL)
		return count - tcount;

	      memcpy (buff, buffer->data, blocksize);
	      break;
	    case FAT_WRITE:
#ifdef FATDEBUG
	      putchar ('w');
#endif
	      buffer = bempty (dev, block);
	      if (buffer == NULL)
		return count - tcount;

	      memcpy (buffer->data, buff, blocksize);
	      buffer->fl_dirty = NORMAL_DIRTYNESS;
	      break;
	    }

	  brelse (buffer);
	  file->f_pos += blocksize;
	  tcount -= blocksize;
	  if (!tcount)
	    return (count);
	  buff += blocksize;
	  if (fat_blockseek (sb, &cluster, &sec, cluster, sec, 1))
	    return (-EIO);
	}
#ifdef FATDEBUG
      putchar ('\n');
#endif
    }

  if (tcount)
    {
#ifdef FATDEBUG
      printf ("fat_rawrw(): tail\n");
#endif
      block = fat_toblock (sb, cluster) + sec;
#ifdef F32DEBUG
      printk ("blk=%d\n", block);
#endif

      buffer = bread (dev, block);
      if (buffer == NULL)
	return count - tcount;

      switch (cmd)
	{
	case FAT_READ:
	  memcpy ((void *) buff, (void *) buffer->data, tcount);
	  break;
	case FAT_WRITE:
	  memcpy ((void *) buffer->data, (void *) buff, tcount);
	  buffer->fl_dirty = NORMAL_DIRTYNESS;
	  break;
	}
      brelse (buffer);
      file->f_pos += tcount;
      tcount = 0;
    }

  return (count - tcount);
}
