/*
   fs/fat/spec.c

   Special but essential fat routines.
   WRITEME: add circular error check
 */

#include <phantom/fs.h>
#include <phantom/fat.h>
#include <errno.h>
#include <string.h>
#include <phantom/buffer.h>

int
fat_toblock (struct super_block *sb, int cluster)
{
  if (cluster > 1)
    return cluster * sb->u.fat.clusterblocks + sb->u.fat.datastart;

  return sb->u.fat.rootentrypos;
}

/* root dir recognized by !cluster */
int
fat_blockseek (struct super_block *sb, int *dstcluster, int *dstsec,
	       int cluster, int sec, int whence)
{
#ifdef FATDEBUG
  printf ("fat_blockseek: sec=%d  cluster=%d  whence=%d\n", sec, cluster,
	  whence);
  printf ("  clusterblocks=%d\n", sb->u.fat.clusterblocks);
#endif

  if (!whence)
    {
      *dstsec = sec;
      *dstcluster = cluster;
      return (0);
    }

  if (whence < 0)
    return (-EIO);

  if (cluster)
    {
      if (sec + whence < sb->u.fat.clusterblocks)
	sec += whence;
      else
	{
	  if (fat_cluster
	      (sb, &cluster, cluster,
	       (sec + whence) / sb->u.fat.clusterblocks))
	    return (-EIO);
	  sec = (sec + whence) % sb->u.fat.clusterblocks;
	}
    }
  else
    {
      if (sec + whence < sb->u.fat.rootentries)
	sec += whence;
      else
	{
#ifdef FATDEBUG
	  printf ("fat_blockseek: tried to seek behind root " \
		  "(sec=%d  whence=%d)\n", sec, whence);
#endif
	  return (-EIO);
	}
    }
#ifdef FATDEBUG
  printf ("  results: cluster=%d  sec=%d\n", cluster, sec);
#endif
  *dstsec = sec;
  *dstcluster = cluster;
  return (0);
}

/* convert common filename to fat 8.3 format */
void
fat_parse (char *dst, const char *src)
{
  int i, j, k;

  j = 0;
  for (i = 0; i < 8; i++)
    {
      if (src[i] != ' ')
	{
	  dst[j] = toupper (src[i]);
	  j++;
	}
    }
  k = j;
  for (i = 8; i < 11; i++)
    {
      if (src[i] != ' ')
	{
	  j++;
	  dst[j] = toupper (src[i]);
	}
    }
  if (j > k)
    {
      dst[k] = '.';
      dst[j + 1] = 0;
    }
  else
    dst[k] = 0;
}

/* convert fat 8.3 to common filename */
void
fat_unparse (char *dst, const char *src, int len)
{
  int i, j;

  i = 0;
  while (src[i] != '.' && len)
    {
      dst[i] = toupper (src[i]);
      i++;
      len--;
    }

  if (!len)
    {
      while (i < 11)
	dst[i++] = ' ';
    }
  else
    {
      for (j = i; j < 8; j++)
	dst[j] = ' ';
      i++;
      len--;
      while (len && j < 11)
	{
	  dst[j++] = toupper (src[i++]);
	  len--;
	}
      while (j < 11)
	dst[j++] = ' ';
    }
  dst[11] = 0;
}

/* get a size of a cluster queue */
int
fat_getsize (struct super_block *sb, int cluster)
{
  int i, j;
  int term = (sb->u.fat.type == 16 ? 0xFFF8 : 0x0FF8);

  i = 1;			/* as itself */
  while ((j = fat_getfatcell (sb, cluster)) < term)
    {
      if (i > 1000000)
	{
	  printf ("fat_getsize: fat table corrupt\n");
	  return -1;
	}
      cluster = j;
      i++;
    }
  return (i);
}

/* only getfatcell and setfatcell can be used for accessing/modifying
   fat table */
u32
fat_getfatcell (struct super_block * sb, int cluster)
{
  u32 res;
  int blk, pos, dev = sb->s_dev;
  struct bhead *b, *bb;

  switch (sb->u.fat.type)
    {
    case 12:			/* obtain 'interresant' fat block */
      blk = sb->u.fat.reserved + (cluster >> 10) * 3;
      /* (cluster / 1024) * 3 */
#ifdef DFATDEBUG
      printf ("getfatcell: 12 bit, cluster %d, blk %d\n", cluster, blk);
#endif
      cluster %= 1024;		/* 1024 clusters per 3 blocks (1536 bytes) */

      switch (cluster)
	{
	case 341:		/* offset 511.5 */
#ifdef DFATDEBUG
	  printf ("  341 case\n");
#endif
	  b = bread (dev, blk);
	  bb = bread (dev, blk + 1);
	  res = b->data[511] | ((bb->data[0] & 0x0F) << 8);
	  brelse (bb);
	  brelse (b);
	  break;

	case 682:		/* offset 1023 */
#ifdef DFATDEBUG
	  printf ("  682 case\n");
#endif
	  blk++;
	  b = bread (dev, blk);
	  bb = bread (dev, blk + 1);
	  res = (b->data[511] >> 4) | (bb->data[0] << 4);
	  brelse (bb);
	  brelse (b);
	  break;

	default:
	  if (cluster > 341)	/* seek the data */
	    {
	      blk++;
	      if (cluster > 682)
		{
		  blk++;
		  pos = ((cluster * 3) >> 1) - 1024;
		}
	      else
		pos = ((cluster * 3) >> 1) - 512;
	    }
	  else
	    pos = (cluster * 3) >> 1;

	  b = bread (dev, blk);

	  if (cluster & 1)
	    res = (b->data[pos] >> 4) | (b->data[pos + 1] << 4);
	  else
	    res = b->data[pos] | ((b->data[pos + 1] & 0x0F) << 8);

	  brelse (b);
	}
      break;

    case 16:
      blk = (cluster >> 8) + sb->u.fat.reserved;	/* 256 clusters per block */
#ifdef DFATDEBUG
      printf ("getfatcell: 16 bit, cluster %d, blk %d\n", cluster, blk);
#endif
      b = bread (dev, blk);
      res = ((u16 *) b->data)[cluster % 256];
      brelse (b);
      break;

    case 32:
      blk = (cluster >> 7) + sb->u.fat.reserved;	/* 128 cluster per block */
#ifdef DFATDEBUG
      printf ("getfatcell: 32 bit, cluster %d, blk %d\n", cluster, blk);
#endif
      b = bread (dev, blk);
      res = ((u32 *) b->data)[cluster % 128];
      brelse (b);
      break;

    default:
      printf ("fat_getfatcell: invalid fat type %d\n", sb->u.fat.type);
      return -1;
    }

#ifdef DFATDEBUG
  printf ("getfatcell: result=0x%x\n", res);
#endif

  return res;
}

int
fat_setfatcell (struct super_block *sb, int cluster, u32 next)
{
  int blk, pos, dev = sb->s_dev;
  struct bhead *b, *bb;

  switch (sb->u.fat.type)
    {
    case 12:			/* (cluster / 1024) * 3 */
      blk = sb->u.fat.reserved + (cluster >> 10) * 3;
#ifdef DFATDEBUG
      printf ("setfatcell: cluster %d, blk %d\n", cluster, blk);
#endif

      cluster %= 1024;		/* 1024 clusters per 3 blocks (1536 bytes) */

      switch (cluster)		/* 'critical clusters' handling */
	{
	case 341:		/* offset 511.5 */
#ifdef DFATDEBUG
	  printf ("  case 341\n");
#endif
	  b = bread (dev, blk);
	  bb = bread (dev, blk + 1);

	  block (b);
	  block (bb);

	  b->fl_dirty = 1;
	  bb->fl_dirty = 1;

	  b->data[511] = next & 0xFF;
	  bb->data[0] = (bb->data[0] & 0xF0) | ((next >> 8) & 0x0F);

	  bunlock (b);
	  bunlock (bb);

	  brelse (bb);
	  brelse (b);

	  sb->u.fat.table_dirt[blk] = 1;
	  sb->u.fat.table_dirt[blk + 1] = 1;
	  break;

	case 682:		/* offset 1023 */
#ifdef DFATDEBUG
	  printf ("  case 682\n");
#endif

	  blk++;
	  b = bread (dev, blk);
	  bb = bread (dev, blk + 1);

	  block (b);
	  block (bb);

	  b->fl_dirty = 1;
	  bb->fl_dirty = 1;

	  b->data[511] = (b->data[511] & 0x0F) | ((next & 0x0F) << 4);
	  bb->data[0] = (next >> 4) & 0xFF;

	  bunlock (b);
	  bunlock (bb);

	  brelse (bb);
	  brelse (b);

	  sb->u.fat.table_dirt[blk] = 1;
	  sb->u.fat.table_dirt[blk + 1] = 1;
	  break;

	default:
#ifdef DFATDEBUG
	  printf ("setfatcell: common case\n");
#endif

	  if (cluster > 341)	/* seek the data */
	    {
	      blk++;
	      if (cluster > 682)
		{
		  blk++;
		  pos = ((cluster * 3) >> 1) - 1024;
		}
	      else
		pos = ((cluster * 3) >> 1) - 512;
	    }
	  else
	    pos = (cluster * 3) >> 1;

	  b = bread (dev, blk);

	  block (b);
	  b->fl_dirty = 1;

	  if (cluster & 1)
	    {
	      b->data[pos] = (b->data[pos] & 0x0F) | ((next & 0x0F) << 4);
	      b->data[pos + 1] = (next >> 4) & 0xFF;
	    }
	  else
	    {
	      b->data[pos] = next & 0xFF;
	      b->data[pos + 1] =
		(b->data[pos + 1] & 0xF0) | ((next >> 8) & 0x0F);
	    }

	  bunlock (b);
	  brelse (b);

	  sb->u.fat.table_dirt[blk] = 1;
	}
      break;

    case 16:
      blk = (cluster >> 8) + sb->u.fat.reserved;	/* 256 clusters per block */
#ifdef DFATDEBUG
      printf ("setfatcell: 16 bits fat, cluster %d, blk %d\n", cluster, blk);
#endif

      b = bread (dev, blk);

      block (b);
      b->fl_dirty = 1;
      ((u16 *) b->data)[cluster % 256] = next;
      bunlock (b);

      brelse (b);

      sb->u.fat.table_dirt[blk] = 1;
      break;

    case 32:
      blk = (cluster >> 7) + sb->u.fat.reserved;	/* 128 clusters per block */
#ifdef DFATDEBUG
      printf ("setfatcell: 32 bits fat, cluster %d, blk %d\n", cluster, blk);
#endif

      b = bread (dev, blk);
      block (b);
      b->fl_dirty = 1;

      ((u32 *) b->data)[cluster % 128] = next;

      bunlock (b);
      brelse (b);

      sb->u.fat.table_dirt[blk] = 1;
      break;

    default:
      printf ("setfatcell: invalid fat type %d\n", sb->u.fat.type);
      return -1;
    }

  return 0;
}

/* 
   get eNth cluster, take CLUSTER as begin
   negative means 'no next' block/cluster
 */
int
fat_cluster (struct super_block *sb, int *dst, int cluster, int n)
{
/*
   #ifdef FATDEBUG
   printf ("fat_cluster, searching %d(th) cluster of %d\n", n, cluster);
   #endif
 */
  while (n--)
    {
      cluster = fat_getfatcell (sb, cluster);
      switch (sb->u.fat.type)
	{
	case 12:
	  if (cluster >= 0x0FF8)
	    return (-1);
	  break;
	case 16:
	  if (cluster >= 0xFFF8)
	    return (-1);
	  break;
	case 32:
	  if ((cluster >> 16 == 0xFFFF) && (cluster & 0xFFFF) >= 0xFFF8)
	    return -1;
	  break;
	default:
	  printk ("fat_cluster: unknown fat type %d!\n", sb->u.fat.type);
	}
    }
  *dst = cluster;
/*
   #ifdef FATDEBUG
   printf("it is %d\n", cluster);
   #endif  
 */
  return (0);
}

int
fat_reserveclusters (struct super_block *sb, int cluster, int count)
{
  int res, i, tmp;

#ifdef FATDEBUG
  printf ("fat_reserveclusters: reserving %d clusters from %d\n", count,
	  cluster);
#endif

  res = 0;
  tmp = fat_getsize (sb, cluster);

#ifdef FATDEBUG
  printf ("fat_reserveclusters: getsize check from cluster %d is %d\n",
	  cluster, tmp);
#endif

  if (tmp - 1 > 0)
    {
      printf ("fat_reserveclusters: reserving not from the end?\n");
      res = tmp - 1;
      printf ("                       %d prereserved clusters\n", res);
/* FIXME: get to the end of the chain */
      if (res >= count)
	{
	  printf
	    ("                   wow, no need to reserve... (already reserved)\n");
	  return (count);
	}
    }

#ifdef FATDEBUG
  printf ("fat_reserveclusters: fat queue=\"%d", cluster);
#endif

  i = cluster;
  while (res < count && i < sb->u.fat.fatentries)
    {
      if (fat_getfatcell (sb, i) == 0)	/* oh, it's free */
	{
#ifdef FATDEBUG
	  printf ("->%d", i);
#endif

	  fat_setfatcell (sb, cluster, i);
	  cluster = i;
	  res++;
	}
      i++;
    }

#ifdef FATDEBUG
  printf ("\"\nqueue nr. 2=\"");
#endif

  if (res == count)
    {
      fat_setfatcell (sb, cluster, FAT_EOF);	/* value will be cutted */
      return (res);
    }

/* we still did not find enough */
  i = 2;			/* skip this... (or not?) */
  while (res < count && i < cluster)
    {
      if (fat_getfatcell (sb, i) == 0)
	{
#ifdef FATDEBUG
	  printf ("->%d", i);
#endif
	  fat_setfatcell (sb, cluster, i);
	  cluster = i;
	  res++;
	}
      i++;
    }

#ifdef FATDEBUG
  putchar ("\"\n");
#endif

  fat_setfatcell (sb, cluster, 0xFFF8);	/* value will be cutted */

#ifdef FATDEBUG
  printf ("fat_reserveclusters: %d clusters of %d reserved\n", res, count);
#endif
  return (res);
}
