/*
   fs/fat/vnode.c

   Fat vnode related functions.
 */

#include <phantom/fs.h>
#include <phantom/fat.h>
#include <errno.h>
#include <phantom/buffer.h>

extern struct file_operations fat_file_operations;	/* in rw.c */

int fat_erase (struct vnode *, char *, int);
int fat_rmdir (struct vnode *, char *, int);
int fat_create (struct vnode **dst, struct vnode *dir, char *name,
		int namelen, int);
int fat_lookup (struct vnode **, struct vnode *, char *, int);
int fat_rename (struct vnode *, char *);
int fat_truncate (struct vnode *vnode, int newsize);
int fat_mkdir (struct vnode *, char *, int, int);

static struct vnode_operations fat_file_vnode_operations = {
  fat_create,			/* create */
  fat_erase,			/* erase */
  fat_lookup,			/* lookup */
  fat_rename,			/* rename */
  fat_truncate,			/* truncate */
  fat_mkdir,			/* mkdir */
  fat_rmdir			/* rmdir */
};

int
fat_translate (struct diritem *di)
{
  int ret;

  ret = 0;

/*  if (FAT_ISVOL (di->attr))
   return (-1); */

  if (FAT_ISREG (di->attr))
    ret |= S_IFREG;

  if (FAT_ISDIR (di->attr))
    {
      if (ret)
	{
	  printf ("fat.c: invalid attribute\n");
	  return (-1);
	}
      ret |= S_IFDIR;
    }

  ret |= S_IRWXU;
  return (ret);
}

#define FATDOT		".          "
#define FATDOTS		"..         "

int
fat_lookup (struct vnode **dst, struct vnode *dir, char *name, int len)
{
  struct super_block *sb = dir->v_sb;
  struct bhead *buffer, *bb;
  int sec, subsec, cluster;
  int blocks = dir->v_size / sb->s_blocksize;
  int searched_cluster = 0;	/* 0 .. do not search */
  char fatname[12];
  struct diritem *di;
  struct vnode *vtmp;

#ifdef FATDEBUG
  printk ("fat_lookup: seeking %s\n", name);
#endif

/* root directory has number zero, no need to test it with FAT_ISROOT */
  cluster = dir->u.fat.cluster;

  if (len == 2 && name[0] == '.' && name[1] == '.')
    {				/* dots case */
#ifdef FATDOTSDEBUG
      printk ("fat_lookup: seeking dots\n");
#endif
      buffer = bread (sb->s_dev, fat_toblock (sb, cluster));
/* it should be on first (second) position */
      di = (struct diritem *) buffer->data + 1;

      if (strncmp (di->name, FATDOTS, 11))
	{			/* it's root dir or invalid dir ==> returning root... */
#ifdef FATDOTSDEBUG
	  printk ("fat_lookup: cannot find dots (root dir?)\n");
#endif
	  *dst = vnode_get (sb, FAT_ROOT_VNODE_NUMBER);
	  if (*dst)
	    return 0;

	  return -EIO;
	}

      bb = bread (sb->s_dev,
		  fat_toblock (sb, (di->cluster_hi << 16) | di->cluster));
      brelse (buffer);
      di = (struct diritem *) bb->data + 1;

      if (strncmp (di->name, FATDOTS, 11))
	{
#ifdef FATDOTSDEBUG
	  printk ("fat_lookup: dots seem to be root...\n");
#endif
	  vtmp = vnode_get (sb, FAT_ROOT_VNODE_NUMBER);
	  if (!vtmp)
	    {
	      printk
		("fat_lookup: panic, cannot get root (dev=0x%x)\n",
		 sb->s_dev);
	      return -EIO;
	    }
	  *dst = vtmp;
	  return 0;
	}

      cluster = (di->cluster_hi << 16) | di->cluster;
      if (!cluster)
	cluster = sb->u.fat.rootdircluster;

      di--;			/* switch to first entry */
      searched_cluster = (di->cluster_hi << 16) | di->cluster;

      if (strncmp (di->name, FATDOT, 11))
	{
	  printk ("fat_lookup: invalid directory, dot does not exist...\n");
	  return -EIO;
	}

      blocks = fat_getsize (sb, cluster) * sb->u.fat.clusterblocks;

      brelse (bb);
#ifdef FATDOTSDEBUG
      printk ("fat_lookup: blocks=%d  searched_cluster=%d  cluster=%d\n",
	      blocks, searched_cluster, cluster);
#endif
    }

  if (!searched_cluster)
    {
      fat_unparse (fatname, name, len);
#ifdef FATDEBUG
      printf
	("  directory blocks: %d  searched fatname: \"%-12s\"\n",
	 blocks, fatname);
#endif
    }

  sec = 0;

  while (blocks--)
    {
#ifdef FATDEBUG
      printf ("cluster=%d  sec=%d\n", cluster, sec);
#endif
      buffer = bread (sb->s_dev, fat_toblock (sb, cluster) + sec);
      for (subsec = 0; subsec < ITEMCOUNT; subsec++)
	{
	  di = (struct diritem *) buffer->data + subsec;
	  if (searched_cluster && FAT_ISDIR (di->attr) &&
	      ((di->cluster_hi << 16) | di->cluster) == searched_cluster)
	    {
#ifdef FATDOTSDEBUG
	      printk ("fat_lookup: found ..'s entry\n");
#endif
	      goto fat_lookup_found;
	    }
	  else if (!strncmp (fatname, di->name, 11))
	    goto fat_lookup_found;
	}
      brelse (buffer);
      if (fat_blockseek (sb, &cluster, &sec, cluster, sec, 1))
	return -EIO;
    }

fat_lookup_found:
  if (subsec >= ITEMCOUNT || (!FAT_ISDIR (di->attr) && !FAT_ISREG (di->attr)))
    {
      brelse (buffer);
      return -ENOENT;
    }

  vtmp = vnode_get (sb, SS2VNUM (fat_toblock (sb, cluster) + sec, subsec));
  *dst = vtmp;
  vtmp->v_mode = fat_translate (di);
  vtmp->u.fat.cluster = (di->cluster_hi << 16) | di->cluster;

  if (S_ISREG (vtmp->v_mode))
    vtmp->v_size = di->size;
  else
    vtmp->v_size = fat_getsize (sb, di->cluster)
      * sb->u.fat.clusterblocks * BLOCKSIZE;

  brelse (buffer);

#ifdef FATDEBUG
  printk ("  found, cluster=%d; size=%d\n", di->cluster, vtmp->v_size);
  printk ("  vmode=0x%x\n", vtmp->v_mode);
#endif
  return 0;
}

int
fat_createitem (struct vnode **dst, struct vnode *dir,
		char *name, int namelen, int mode, char fl_dir)
{
  char fatname[12];
  struct bhead *bh;
  struct diritem *di;
  struct vnode *vnode;
  struct super_block *sb = dir->v_sb;
  int blocks = TOBLOCK (dir->v_size);
  int i, sec, freesec, freesubsec, cluster;

  sec = 0;
  freesec = -1;
  cluster = dir->u.fat.cluster;
  fat_unparse (fatname, name, namelen);

#ifdef FATDEBUG
  printk
    ("fat_create: directory: sec=%d  blocks=%d  filename=\"%-12s\"\n",
     sec, blocks, fatname);
#endif

  while (blocks--)
    {
      bh = bread (sb->s_dev, fat_toblock (sb, cluster) + sec);
      if (!bh)
	return -EIO;
      for (i = 0; i < ITEMCOUNT; i++)
	{
	  di = (struct diritem *) bh->data + i;
/* use of free entry is better than use of deleted, undelete purposes, etc. */
/* this still should be checked out, should not rewrite data of 'deleted' file  */
	  if ((freesec == -1 && di->name[0] == DELETED) || !di->name[0])
	    {
	      freesec = fat_toblock (sb, cluster) + sec;
	      freesubsec = i;
	      if (!di->name[0])
		goto createitem_found;
	    }

	  if (!strncmp (fatname, di->name, 11))
	    {
#ifdef FATDEBUG
	      printk ("fat_create: file already exists\n");
#endif
	      if (mode & O_TRUNC)
		{
#ifdef FATDEBUG
		  printk ("  and will be truncated (WRITEME!)\n");
#endif
		  /* WRITEME: open, fat_truncate, ... */
		  brelse (bh);
		  return -EIO;
		}
	      else
		{
		  brelse (bh);
		  return -EEXIST;
		}
	    }
	}
      if (blocks && fat_blockseek (sb, &cluster, &sec, cluster, sec, 1))
	return -EIO;

      brelse (bh);
    }

/* if there exists at least DELETED entry in the directory which
is 'too' full the directory is not extended but first DELETED
is used (see above) */

/* directory 'too' full and no DELETED entry found */
  if (freesec == -1)
    {
#ifdef FATDEBUG
      printf ("fat_create: enlarging parent directory by one cluster\n");
#endif
      if (fat_reserveclusters (sb, cluster, 1))
	return (-ENOSPC);
      if (fat_blockseek (sb, &cluster, &sec, cluster, sec, 1))
	return (-EIO);
      freesec = fat_toblock (sb, cluster);	/* sec==0 */
      freesubsec = 0;
      bh = bempty (sb->s_dev, freesec);
      if (!bh)
	return -EIO;
    }

createitem_found:

#ifdef FATDEBUG
  printf ("fat_create: entry found, filling, writing\n");
#endif

/* reserve one cluster for a directory but NOT FOR A FILE! */
  cluster = 0;
  if (fl_dir)
    {
      while (cluster < sb->u.fat.fatentries)
	if (!fat_getfatcell (sb, cluster++))
	  break;

      if (cluster >= sb->u.fat.fatentries)
	{
#ifdef FATDEBUG
	  printf ("fat_create: no space\n");
#endif
	  return -ENOSPC;
	}

      fat_setfatcell (sb, cluster, FAT_EOF);
    }

  block (bh);
  bh->fl_dirty = 1;

  bzero (di, ITEMSIZE);
  strncpy (di->name, fatname, 11);
  di->cluster = cluster;
  di->attr = fl_dir ? A_DIR : 0;
  di->size = 0;

  bunlock (bh);
  brelse (bh);

  vnode = vnode_get (sb, SS2VNUM (freesec, freesubsec));
  vnode->u.fat.cluster = cluster;

#ifdef FATDEBUG
  printk ("fat_create: freesec=%d  freesubsec=%d\n", freesec, freesubsec);
#endif

  *dst = vnode;

  if (fl_dir)
    {				/* whole cluster should be freed... */
      for (i = 0; i < sb->u.fat.clusterblocks - 1; i++)
	{
	  bh = bempty (sb->s_dev, fat_toblock (sb, cluster) + i + 1);
	  if (!bh)
	    return -EIO;
	  brelse (bh);
	}

      bh = bempty (sb->s_dev, fat_toblock (sb, cluster));
      if (!bh)
	return -EIO;
      block (bh);
      bh->fl_dirty = 1;

      di = (struct diritem *) bh->data;
      memcpy (di->name, FATDOT, 11);
      di->attr = A_DIR;
      di->cluster = cluster & 0xFFFF;
      di->cluster_hi = cluster >> 16;

      di++;
      memcpy (di->name, FATDOTS, 11);
      di->attr = A_DIR;
      di->cluster = dir->u.fat.cluster & 0xFFFF;
      di->cluster_hi = dir->u.fat.cluster >> 16;

      bunlock (bh);
      brelse (bh);
    }

#ifdef FATDEBUG
  printf ("fat_create: %s successfully created\n",
	  fl_dir ? "directory" : "file");
#endif

  return 0;
}

int
fat_create (struct vnode **dst, struct vnode *dir,
	    char *name, int len, int mode)
{
  return (fat_createitem (dst, dir, name, len, mode, 0));
}

int
fat_mkdir (struct vnode *dir, char *name, int len, int mode)
{
  int ret;
  struct vnode *dumb;
  ret = fat_createitem (&dumb, dir, name, len, mode, 1);
  vnode_release (dumb);
  return (ret);
}

int
fat_eraseitem (struct vnode *dir, char *name, int len, char fl_dir)
{
  struct super_block *sb = dir->v_sb;
  int clustersize, cluster, fileclusters;
  int i, j, sec, subsec;
  struct bhead *buffer;
  struct diritem *di;
  struct vnode *vnode;
  if (len == 1 && name[0] == '.')
    return (-ENOENT);
  if (len == 2 && name[0] == '.' && name[1] == '.')
    return (-ENOENT);
  if (lookup (&vnode, dir, name, len))	/* not fat_lookup! */
    {
#ifdef FATDEBUG
      printf ("fat_erase: object not found\n");
#endif
      return (-ENOENT);
    }

  if (FAT_ISROOT (vnode))
    {
#ifdef FATDEBUG
      printf ("fat_erase(): cannot erase root :)\n");
#endif
      return (-EACCES);
    }

  cluster = vnode->u.fat.cluster;
  if (fl_dir)
    {
      if (!S_ISDIR (vnode->v_mode))
	{
#ifdef FATDEBUG
	  printf ("fat_erase(): not a regular directory...\n");
#endif
	  return (-EACCES);
	}
      else
	{
	  if (!S_ISREG (vnode->v_mode))
	    {
#ifdef FATDEBUG
	      printf ("fat_erase(): not a regular file...\n");
#endif
	      return (-EACCES);
	    }
	}
    }

  clustersize = sb->u.fat.clusterblocks * sb->s_blocksize;
  fileclusters = vnode->v_size / clustersize;
  if (vnode->v_size % clustersize)
    fileclusters++;
  if (!fileclusters)		/* used for file fat chain check */
    fileclusters++;
  if (fl_dir)			/* the dir must be free in order to delete it */
    {
      sec = 0;
      do
	{
	  buffer = bread (sb->s_dev, fat_toblock (sb, cluster) + sec);
	  for (i = 0; i < ITEMCOUNT; i++)
	    {
	      di = (struct diritem *) (buffer->data + i + ITEMSIZE);
	      if (di->name[0] && di->name[0] != DELETED)
		{
#ifdef FATDEBUG
		  printf
		    ("fat_eraseitem: tried to erase non-empty directory\n");
#endif
		  return (-ENOTEMPTY);
		}
	    }
	  brelse (buffer);
	}
      while (!fat_blockseek (sb, &cluster, &sec, cluster, sec, 1));
      cluster = vnode->u.fat.cluster;	/* restore */
    }

/* free the fat and count file clusters */
  i = 1;
  while (!fat_cluster (sb, &j, cluster, 1))
    {
      fat_setfatcell (sb, cluster, 0);
      cluster = j;
      i++;
    }
  fat_setfatcell (sb, cluster, 0);
  if (!fl_dir && i != fileclusters)
    {
      if (i > fileclusters)
	printf ("fat_erase: more clusters owned by deleted file\n");
      else if (i < fileclusters)
	printf ("fat_erase: less clusters owned by deleted file\n");
      printf ("       FIXED!\n");
    }

  sec = VNUM2SEC (vnode->v_num);
  subsec = VNUM2SUBSEC (vnode->v_num);
  buffer = bread (sb->s_dev, sec);
  di = ((struct diritem *) buffer->data) + subsec;
#ifdef FATDEBUG
  printf ("fat_erase(): file %s erased\n", di->name);
#endif
  di->name[0] = DELETED;
  brelse (buffer);
  vnode_release (vnode);
  return (0);
}

int
fat_erase (struct vnode *dir, char *name, int len)
{
  return (fat_eraseitem (dir, name, len, 0));
}

int
fat_rmdir (struct vnode *dir, char *name, int len)
{
  return (fat_eraseitem (dir, name, len, 1));
}

int
fat_read_vnode (struct vnode *vnode)
{
  vnode->v_op = &fat_file_vnode_operations;	/* oughta be all */
  vnode->f_op = &fat_file_operations;
/* WRITEME: modify/ctreate time should be read, add entry to name cache */
/* fill fat specific structure(s) (-> what should be in fat_lookup?) */
  return (0);
}

int
fat_truncate (struct vnode *vnode, int newsize)
{
  struct super_block *sb = vnode->v_sb;
  struct diritem *di;
  int newclusters, oldclusters;
  int clustersize, tmp0, tmp1, sec, subsec;
  struct bhead *buffer;
  int i;

  if (newsize >= vnode->v_size)
    return (0);			/* ok, do nothing?, WRITEME... */

  clustersize = sb->u.fat.clusterblocks * sb->s_blocksize;
  newclusters = (newsize + clustersize - 1) / clustersize;
  oldclusters = (vnode->v_size + clustersize - 1) / clustersize;

  tmp0 = vnode->v_num;
  if (fat_cluster (sb, &tmp0, tmp0, newclusters))
    return (-EIO);

/* free fat */
  tmp1 = fat_getfatcell (sb, tmp0);
  fat_setfatcell (sb, tmp0, FAT_EOF);

  for (i = 0; i < newclusters - oldclusters; i++)
    {
      tmp0 = fat_getfatcell (sb, tmp1);
      fat_setfatcell (sb, tmp1, 0);
      tmp1 = tmp0;
    }

/* modify header */
  sec = VNUM2SEC (vnode->v_num);
  subsec = VNUM2SUBSEC (vnode->v_num);

  buffer = bread (sb->s_dev, sec);
  di = ((struct diritem *) buffer->data) + subsec;
  di->size = newsize;
  vnode->v_size = newsize;
  brelse (buffer);

  return (0);
}

int
fat_rename (struct vnode *vnode, char *name)
{
  char fatname[12];
  struct bhead *buffer;
  struct diritem *di;
  int sec, subsec;
  if (!vnode->v_num)
    {
#ifdef FATDEBUG
      printf ("fat_rename(): Cannot rename root!\n");
#endif
      return (-EINVAL);
    }

  fat_unparse (fatname, name, strlen (name));
#ifdef FATDEBUG
  printf ("fat_rename(): rename: %s -> %s\n", name, fatname);
#endif
  sec = VNUM2SEC (vnode->v_num);
  subsec = VNUM2SUBSEC (vnode->v_num);
  buffer = bread (vnode->v_dev, sec);
  di = ((struct diritem *) buffer->data) + subsec;
  strcpy (di->name, fatname);
  brelse (buffer);
  return (0);
}
