/*
   fs/open.c
 */

#include <phantom/fs.h>
#include <phantom/sched.h>
#include <errno.h>
#include <string.h>
#include <phantom/major.h>
#include <phantom/tty.h>
#include <phantom/lock.h>

extern struct file_operations *get_chrfops (unsigned int, unsigned int);
extern struct file_operations *get_blkfops (unsigned int);

struct file file_table[NR_FILE] = { {0,}, };

/* this one returns one free entry of file_table -- if exists */
struct file *
get_free_filp ()
{
  struct file *file;

  single_code ();
  for (file = file_table; file < file_table + NR_FILE; file++)
    {
      if (!file->f_used)
	{
	  bzero (file, sizeof (struct file));
	  file->f_used = 1;
	  multi_code ();
	  return (file);
	}
    }

  multi_code ();
  return (NULL);
}

int
get_fd ()
{
  int fd;
  struct file *file;

  fd = 0;
  while (fd < NR_OPEN && ctask->filp[fd])
    fd++;

  if (fd == NR_OPEN)
    {
      printk ("get_fd: too many open files\n");
      return -1;
    }

  file = get_free_filp ();
  if (file == NULL)
    {
      printk ("get_fd: cannot get free filp!\n");
      return -1;
    }

  ctask->filp[fd] = file;

  return fd;
}

struct file *
fd2file (int fd)
{
  struct file *tmp;

  if (fd >= NR_OPEN || !(tmp = ctask->filp[fd]))
    {
      printk ("fd2file: fatal, invalid fd (%d) given, pid=%d\n", fd,
	      ctask->pid);
      return NULL;
    }

  return tmp;
}

void
free_fd (int fd)
{
  if (fd < 0 || fd > NR_OPEN)
    {
      printk ("free_fd: invalid fd\n");
      return;
    }

  if (!ctask->filp[fd] || ctask->filp[fd]->f_used != 1)
    {
      printk ("free_fd: used!=1\n");
      return;
    }

  ctask->filp[fd]->f_used = 0;
  ctask->filp[fd] = NULL;
}

struct socket *
get_socket (int fd)
{
  struct file *file;
  struct socket *skt;

  file = ctask->filp[fd];
  if (file == NULL)
    {
      printk ("get_socket: invalid fd\n");
      return NULL;
    }

  if (file->f_socket == NULL)
    printk ("get_socket: NULL skt here\n");

  return file->f_socket;
}

int
sys_open (char *name, int flags, int mode)
{
  int i, fd, maj, min, ret;
  struct file *file;

  fd = 0;
  while (fd < NR_OPEN && ctask->filp[fd])
    fd++;

  if (fd > NR_OPEN)
    return (-EMFILE);		/* process owns too many opened files */

  if ((file = get_free_filp ()) == NULL)
    {
      printf ("sys_open: cannot get free filp!\n");
      return -EMFILE;
    }

  ctask->filp[fd] = file;
  if (ret = vnode_open (&(file->f_vnode), name, flags))	/* mode should come... */
    {
      file->f_used = 0;
      ctask->filp[fd] = NULL;
      return (ret);
    }

#ifdef FSDEBUG
  printf ("sys_open: vnode=0x%p\n", file->f_vnode);
#endif

  if (file->f_vnode)
    {
      file->f_op = file->f_vnode->f_op;
      file->f_dev = file->f_vnode->v_dev;
    }				/* else it is bzeroed by default -- close, init... */

  file->f_used = 1;
  file->f_pos = 0;

  if (file->f_op && file->f_op->open)
    {
      ret = file->f_op->open (file);
      if (ret)
	{
	  file->f_used = 0;
	  ctask->filp[fd] = NULL;
	  return (ret);
	}
    }
/* open is ok if custom open doesn't exist, not ok for read, write, etc. */
  return (fd);
}

int
sys_dup (int fd_orig)
{
  int fd;
  struct file *file, *src;

  if (fd_orig >= NR_OPEN || !ctask->filp[fd_orig]
      || !ctask->filp[fd_orig]->f_used)
    return -EBADF;

  src = ctask->filp[fd_orig];

  if (!(file = get_free_filp ()))
    return -EMFILE;

  single_code ();

  fd = 0;
  while (fd < NR_OPEN && ctask->filp[fd])
    fd++;

  if (fd == NR_OPEN)
    {
      multi_code ();
      file->f_used = 0;
      return -EMFILE;
    }

  ctask->filp[fd] = file;
  multi_code ();

  vnode_gain (src->f_vnode);

  memcpy (file, src, sizeof (struct file));


  return fd;
}

int
sys_creat (char *name, int mode)
{
  return (sys_open (name, O_CREAT | O_WRONLY | O_TRUNC, mode));
}

int
sys_close (unsigned int fd)
{
  int ret;
  struct file *file;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]) || !file->f_used)
    return -EINVAL;

  ret = 0;
  if (file->f_op && file->f_op->close)
    ret = file->f_op->close (file);

  if (!ret)
    {
      if (file->f_vnode)	/* offline mode patch */
	vnode_release (file->f_vnode);
#ifdef FSDEBUG
      else
	printf ("sys_close: offline mode (NULL vnode) occured\n");
#endif
      bzero ((void *) file, sizeof (struct file));
      ctask->filp[fd] = NULL;
      return (0);
    }

#ifdef FSDEBUG
  printf ("sys_close: cannot close\n");
#endif

  return ret;
}

/* name should be absolute path because of ctask->pwd_str */
int
sys_chdir (char *name)
{
  int ret, i;
  struct vnode *vnode;

  if (!name[0])
    return 0;

/* only touch if it really exists */
  if (ret = vnode_open (&vnode, name, O_RDONLY))
    return (ret);

  if (!S_ISDIR (vnode->v_mode))
    {
      vnode_release (vnode);
      return -ENOTDIR;
    }

  if (name[0] != '/')
    {
      i = strlen (ctask->pwd_str);

      if (i && ctask->pwd_str[i - 1] != '/')
	strcat (ctask->pwd_str, "/");

      strcat (ctask->pwd_str, name);
    }
  else
    {
      strcpy (ctask->pwd_str, name);

      i = strlen (ctask->pwd_str);
      if (i > 1 && ctask->pwd_str[i - 1] == '/')
	ctask->pwd_str[i - 1] = 0;
    }

  if (ctask->pwd)
    {
#ifdef FSDEBUG
      printf ("chdir: releasing ctask->pwd, ctask->pwd->v_used=%d\n",
	      ctask->pwd->v_used);
#endif
      vnode_release (ctask->pwd);
    }

  ctask->pwd = vnode;

  return 0;
}

char *
sys_getcwd (char *buff, int len)
{
  if (len < strlen (ctask->pwd_str))
    return buff;
  strcpy (buff, ctask->pwd_str);
  return buff;
}

int
sys_fstat (int fd, struct stat *st)
{
  struct file *file;
  struct vnode *vnode;

  if (fd >= NR_OPEN || !(file = ctask->filp[fd]) || !file->f_used)
    return -EINVAL;

  vnode = file->f_vnode;

  if (!vnode)
    {
      printf ("sys_fstat: fd with NULL vnode, giving up\n");
      return -ENOENT;
    }

/* (device, vnode num) is unique identifier.. */
  st->st_dev = vnode->v_dev;
  st->st_ino = vnode->v_num;
  st->st_mode = vnode->v_mode;
  st->st_size = vnode->v_size;
  st->st_uid = 0;
  st->st_gid = 0;
  st->st_atime = vnode->v_atime;
  st->st_ctime = vnode->v_ctime;
  st->st_mtime = vnode->v_mtime;

  return 0;
}

int
sys_stat (char *name, struct stat *st)
{
  int ret;
  struct vnode *vnode;

  if (ret = vnode_open (&vnode, name, 0))
    {
#ifdef VNODEDEBUG
      printk ("sys_stat: vnode_open fails, ret=%d\n", ret);
#endif
      return ret;
    }

  st->st_dev = vnode->v_dev;
  st->st_mode = vnode->v_mode;
  st->st_size = vnode->v_size;
  st->st_uid = 0;
  st->st_gid = 0;
  st->st_atime = vnode->v_atime;
  st->st_ctime = vnode->v_ctime;
  st->st_mtime = vnode->v_mtime;

  vnode_release (vnode);

  return 0;
}

int
sys_isatty (int fd)
{
  struct file *file;

  printk ("sys_isatty: fd=%d\n", fd);
  if (fd < 0 || fd >= NR_OPEN)
    return 0;

  file = ctask->filp[fd];
  if (MAJOR (file->f_dev) == TTY_MAJOR)
    {
      printk ("  yes, it is\n");
      return 1;
    }

  printk ("  no, it isn't\n");
  return 0;
}
