/*
   fs/vnode.c
 */

#include <phantom/fs.h>
#include <errno.h>
#include <phantom/sched.h>
#include <phantom/dfs.h>
#include <phantom/major.h>

struct vnode vnode_table[NR_VNODE] =
{
  {0,},};

struct vnode *
vnode_getfree ()
{
  struct vnode *vnode;

  for (;;)
    {
      for (vnode = vnode_table; vnode < vnode_table + NR_VNODE; vnode++)
	if (!vnode->v_used)
	  return (vnode);

      printk ("vnode.c: cannot get free vnode!\n");
      sys_sleep (10);		/* :) */
    }
}

int
vread (struct vnode *vnode)
{
  struct super_block *super;

  super = vnode->v_sb;
  if (super->s_op && super->s_op->read_vnode)
    return (super->s_op->read_vnode (vnode));

  printf ("vread: no s_op or read_vnode\n");
  return (-1);
}

int
vwrite (struct vnode *vnode)
{
  struct super_block *super;

  super = vnode->v_sb;
  if (super && super->s_op->write_vnode)
    return (super->s_op->write_vnode (vnode));

  printf ("vwrite: no s_op or write_vnode\n");
  return (-1);
}

struct vnode *
vnode_get (struct super_block *sb, int num)
{
  struct vnode *vnode;

#ifdef VNODEDEBUG
  printf ("vget: sb=0x%p  num=%d\n", sb, num);
#endif

  for (vnode = vnode_table; vnode < vnode_table + NR_VNODE; vnode++)
    if (vnode->v_sb == sb && vnode->v_num == num)
      break;

  if (vnode >= vnode_table + NR_VNODE)	/* not found, get free */
    {
#ifdef VNODEDEBUG
      printf ("vget: does not exist yet\n");
#endif
      vnode = vnode_getfree ();
      vnode->v_sb = sb;
      vnode->v_dev = sb->s_dev;
      vnode->v_num = num;
      vnode->v_used++;
      vnode->v_mount = NULL;
      vnode->v_cover = NULL;
      vnode->v_mode = 0;
      vread (vnode);

      return vnode;
    }

/* found, ok */
  if (vnode->v_used <= 0)
    printf ("vget: badly freed vnode, vnode 0x%p  v_used %d\n",
	    vnode, vnode->v_used);

  vnode->v_used++;
  return (vnode);
}

int
vnode_gain (struct vnode *vnode)
{
  if (!vnode)
    {
      printk ("vnode_gain: alert! NULL vnode given\n");
      return -EINVAL;
    }

  if (!vnode->v_used)
    printk ("vnode_gain: vnode->v_used == 0\n");

  vnode->v_used++;

  return 0;
}

int
vnode_release (struct vnode *vnode)
{
  if (!vnode)
    {
      printf ("vnode_release: alert! NULL vnode given\n");
      return -EINVAL;
    }

  if (vnode->v_used)
    {
      vnode->v_used--;
      if (!vnode->v_used)
	{
/* vwrite(vnode); */
	  bzero ((void *) vnode, sizeof (struct vnode));
	}
      return (0);
    }
  printf ("vnode_release: Cannot free free vnode (0x%p)!\n", vnode);
  return (-1);
}

int
lookup (struct vnode **dst, struct vnode *src, char *name, int namelen)
{
#ifdef VNODEDEBUG
  printk ("lookup: src=0x%x  name=%s  namelen=%d\n", src, name, namelen);
#endif

  if (src->v_op && src->v_op->lookup)
    {
      if (src->v_cover && namelen == 2 && name[0] == '.' && name[1] == '.')
	{
	  src = src->v_cover;
	  src->v_used++;

	  if (src->v_op && src->v_op->lookup)
	    return src->v_op->lookup (dst, src, name, 2);

#ifdef VNODEDEBUG
	  printk ("  no lookup defined!\n");
#endif

	  goto lookup_no_lookup;
	}

      if (namelen == 1 && name[0] == '.')
	{
#ifdef VNODEDEBUG
	  printf ("lookup: 'local' dir open\n");
#endif
	  if (src->v_mount)
	    {
#ifdef VNODEDEBUG
	      printk ("lookup: mount takeover\n");
#endif
	      *dst = src->v_mount;
	      src->v_mount->v_used++;
	      return 0;
	    }

	  *dst = src;
	  src->v_used++;
	  return 0;
	}

      if (src->v_op->lookup (dst, src, name, namelen))
	{
#ifdef VNODEDEBUG
	  printk ("lookup: did not find %s\n", name);
#endif
	  return (-ENOENT);
	}
#ifdef VNODEDEBUG
      printk ("lookup: found %s\n", name);
#endif

      if ((*dst)->v_mount)	/* mount takeover */
	{
#ifdef VNODEDEBUG
	  printk ("lookup: vnode takeover\n");
#endif
	  (*dst)->v_mount->v_used++;
	  (*dst)->v_used--;
	  *dst = (*dst)->v_mount;
	}

      return 0;
    }

lookup_no_lookup:

#ifdef VNODEDEBUG
  printk ("lookup: invalid vnode or filesystem without lookup?\n");
  printk ("  src->v_op=0x%x\n", src->v_op);
#endif

  return -ENXIO;		/* ;) */
}

/* 
   get the 'last' name in path and return vnode of it's dir
   only vnode_get should use this...
 */
int
vnode_getdir (char *pathname, char **name, int *namelen,
	      struct vnode *src, struct vnode **dir)
{
  struct vnode *dst;
  char *lastname;
  int len, ret;

  len = 0;			/* root dir indicator */
  *dir = NULL;
  if (!src)
    src = ctask->pwd;

  if (pathname[0] == '/')	/* absolute path */
    {
      src = ctask->root;
      pathname++;
    }

  if (!src)
    {
      printk ("vnode_getdir: panic, src = NULL\n");
      return -EINVAL;
    }

/* dfs alien block -- root dir listing */
/* what to do with '..' later ? */
  if (src == ctask->root
      && (!strncmp (pathname, DFS_NAME, DFS_NAME_LEN)
	  && (!pathname[DFS_NAME_LEN] || pathname[DFS_NAME_LEN] == '/')))
    {
      pathname += DFS_NAME_LEN;
      if (*pathname)		/* == '/' */
	pathname++;

      lastname = pathname;
      while (*pathname)
	{
	  if (*pathname == '/')	/* sub-dirs not allowed in devfs */
	    return -ENOENT;
	  pathname++;
	  len++;
	}
      *name = lastname;
      *namelen = len;
      *dir = vnode_get (dfs_super, DFS_ROOT_VNODE_NUMBER);
      if (!*dir)
	{
	  printf ("vnode_getdir: dfs vnode problem, dfs-root not found\n");
	  return -ENOENT;
	}
      /*      (*dir)->v_used--;               *//* small 'patch' */
#ifdef VNODEDEBUG
      printf ("vnode_getdir: dfs occurred, ok\n");
#endif
      return 0;
    }

  while (*pathname)
    {
      src->v_used++;
      len = 0;
      lastname = pathname;

      while (*pathname && *pathname != '/')
	{
	  pathname++;
	  len++;
	}

      if (*pathname)
	{
	  pathname++;		/* skip the slash */

	  if (ret = lookup (&dst, src, lastname, len))
	    return ret;

	  vnode_release (src);
	  src = dst;
	}
    }

  *name = lastname;
  *namelen = len;
  *dir = src;

#ifdef VNODEDEBUG
  printf ("vnode_getdir: lastname=%s  len=%d\n", lastname, len);
#endif
  return (0);
}

int
vnode_open (struct vnode **dst, char *pathname, int flags)
{
  char *lastname;
  struct vnode *dir;
  int len, ret;

  if (ret = vnode_getdir (pathname, &lastname, &len, NULL, &dir))
    return ret;

  if (!dir)
    {
      *dst = NULL;
      printk ("vnode_open: offline mode, root dir opened with no vnode...\n");
      return 0;
    }

/* directory open case - recognized via pathname - pathname ends with slash */
/* FIXME: this should be removed.. */
  if (!len)
    {
#ifdef VNODEDEBUG
      printf ("vnode_open: directory open\n");
#endif

      if (flags & O_CREAT)
	return (-EACCES);

      *dst = dir;
      return ret;
    }

/* oridnary file open case - check directory permissions, etc. */
  if (flags & O_CREAT)
    {
      if (!lookup (dst, dir, lastname, len))
	{
	  if (flags & O_EXCL)
	    return -EEXIST;
	  if (flags & O_TRUNC)
	    {
	      if (dir->v_op && dir->v_op->truncate)
		return dir->v_op->truncate (*dst, 0);
	      return -EACCES;
	    }

	  return 0;
	}

      if (dir->v_op && dir->v_op->create)
	ret = dir->v_op->create (dst, dir, lastname, len, 0 /* mode */ );
    }
  else
    ret = lookup (dst, dir, lastname, len);

  vnode_release (dir);		/* we don't need dir anymore */

  return ret;
}

int
sys_unlink (char *name)
{
  int ret;
  struct vnode *dir;
  char *lastname;
  int len;

  if (ret = vnode_getdir (name, &lastname, &len, NULL, &dir))
    return (ret);
  if (dir->v_op && dir->v_op->erase)
    ret = dir->v_op->erase (dir, lastname, len);
  else
    {
      vnode_release (dir);
      return (-EROFS);
    }

  vnode_release (dir);
  return (0);
}

int
sys_mkdir (char *name, int mode)
{
  int ret, len;
  char *lastname;
  struct vnode *vnode;

  if (vnode_getdir (name, &lastname, &len, NULL, &vnode))
    return (-ENOENT);
  ret = -EPERM;
  if (vnode->v_op && vnode->v_op->mkdir)
    ret = vnode->v_op->mkdir (vnode, lastname, len, mode);
  vnode_release (vnode);
  return (ret);
}

int
sys_rmdir (char *name)
{
  int ret, len;
  char *lastname;
  struct vnode *vnode;

  if (vnode_getdir (name, &lastname, &len, NULL, &vnode))
    return (-ENOENT);
  ret = -EPERM;
  if (vnode->v_op && vnode->v_op->rmdir)
    ret = vnode->v_op->rmdir (vnode, lastname, len);
  vnode_release (vnode);
  return (ret);
}

void
vlist ()
{
  int i;
  struct vnode *vnode;
  i = 0;
  printf ("\n%-3s %-10s %-5s %-6s %-4s %s %-10s %-10s\n",
	  "nr", "addr", "dev", "mode", "vnum", "used", "mount", "cover");
  for (vnode = vnode_table; vnode < vnode_table + NR_VNODE; vnode++)
    {
/*      if (vnode->v_used) */
      printf ("%-3d 0x%-8x %02X:%02X 0x%-4x %-4d %-4d 0x%-8x 0x%-8x\n", i,
	      vnode, MAJOR (vnode->v_dev), MINOR (vnode->v_dev),
	      vnode->v_mode, vnode->v_num, vnode->v_used, vnode->v_mount,
	      vnode->v_cover);
      i++;
    }
  printf ("---\nctask->root = 0x%p  ctask->pwd = 0x%p\n", ctask->root,
	  ctask->pwd);
}
