#ifndef FS_H
#define FS_H

#include <dirent.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <phantom/socket.h>
#include <phantom/fs_fat.h>
#include <rg/window.h>

#define PATH_MAX	256

#define NR_FILE		64
#define NR_OPEN		32
#define NR_SUPER	16
#define NR_VNODE	16  /*64*/	/* debuging purposes */

#define SEEK_SET	0
#define SEEK_CUR	1
#define SEEK_END	2

#define O_ACCMODE	  0003
#define O_RDONLY	    00
#define O_WRONLY	    01
#define O_RDWR		    02
#define O_CREAT		  0100	/* not fcntl */
#define O_EXCL		  0200	/* not fcntl */
#define O_NOCTTY	  0400	/* not fcntl */
#define O_TRUNC		 01000	/* not fcntl */
#define O_APPEND	 02000
#define O_NONBLOCK	 04000
#define O_NDELAY	O_NONBLOCK
#define O_SYNC		010000
#define FASYNC		020000
#define O_DIRECT	 040000	/* direct disk access hint - currently ignored */
#define O_LARGEFILE	0100000
#define O_DIRECTORY	0200000	/* must be a directory */
#define O_NOFOLLOW	0400000	/* don't follow links */

struct vnode
{
  int v_dev;			/* parent device */
  int v_used;			/* times used */
  int v_mode;			/* mode (file, dir, etc.) */
  long v_num;			/* number of vnode */
  long v_size;			/* total size */
  struct vnode *v_cover;	/* backlink, mount covered vnode, '..' case */
  struct vnode *v_mount;	/* mount point ? ==> mount root vnode */
  struct vnode_operations *v_op;	/* vnode operations */
  struct file_operations *f_op;	/* file operations :) */
  struct super_block *v_sb;	/* parent's superblock */
  time_t v_atime;		/* last access time */
  time_t v_ctime;		/* creation time */
  time_t v_mtime;		/* last modify time */
  union
  {
    struct fat_vnode_info fat;
  }
  u;
};

struct file
{
  char f_used;
  dev_t f_dev;
  off_t f_pos;
  struct file_operations *f_op;
  struct vnode *f_vnode;
  struct socket *f_socket;
  WINDOW *f_window;
};

struct file_operations
{
  int (*lseek) (struct file *, off_t, int);
  int (*read) (struct file *, char *, size_t);
  int (*write) (struct file *, char *, size_t);
  int (*readdir) (struct file *, struct dirent *, int count);
  int (*ioctl) (struct file *, unsigned int, unsigned int);
  int (*open) (struct file *);
  int (*close) (struct file *);
};

struct vnode_operations
{
  int (*create) (struct vnode **, struct vnode *, char *, int, int);
  int (*erase) (struct vnode *, char *, int);
  int (*lookup) (struct vnode **, struct vnode *, char *, int);
  int (*rename) (struct vnode *, char *);
  int (*truncate) (struct vnode *, int);
  int (*mkdir) (struct vnode *, char *, int, int);
  int (*rmdir) (struct vnode *, char *, int);
};

struct super_operations
{
  int (*read_vnode) (struct vnode *);
  int (*write_vnode) (struct vnode *);
  int (*release_super) (struct super_block *);
};

struct file_system_type
{
  int (*read_super) (struct super_block *);
  char *name;
};

struct super_block
{
  int s_dev;
  int s_blocksize;
  int s_flags;
  int s_used;
  char s_rdonly;
  struct super_operations *s_op;
  struct vnode *s_mount;
  union
  {
    struct fat_super_info fat;
  }
  u;
};

extern struct file file_table[NR_FILE];
extern struct vnode vnode_table[NR_VNODE];
extern struct super_block super_table[NR_SUPER];

extern struct vnode *vnode_get (struct super_block *, int);

extern int *blk_blocks[];

#define vget()		vnode_get()
#define vrel()		vnode_release()

struct socket *get_socket (int fd);
int get_fd ();
void free_fd (int fd);
struct file *fd2file(int fd);

#endif
