/*
   kernel/timer.c
 */

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/regdump.h>
#include <asm/timer.h>
#include <phantom/time.h>
#include <sys/types.h>
#include <conio.h>
#include <phantom/sched.h>

#define TIMER_IRQ	0
#define PIT_FREQ	0x1234DC
#define LATCH		PIT_FREQ/HZ

#define TIMER_STOPPED	0x00
#define TIMER_RUNNING	0x01

#define CMOS_R(a) ({outb_p(0x80|(a),0x70); \
		  inb_p(0x71);})

#define BCD2BIN(a) (a)=((a)&15)+((a)>>4)*10

struct timer_st
{
  ulg flags;
  ulg remaining;
  ulg period;
  void (*handler) ();
};

struct timer_st timer[NR_TIMER];
time_t global_time, start_time;
struct tm gtm;

void
timer_irq (int irq_nr, struct regdump *regs)
{
  int i;

  global_time += 10;

  for (i = 0; i < NR_TIMER; i++)
    {
      if (timer[i].flags == TIMER_RUNNING)
	{
	  if (!timer[i].remaining)
	    {
	      timer[i].remaining = timer[i].period;
	      timer[i].handler ();
	    }
	  else
	    timer[i].remaining--;
	}
    }

  if (ctask->beats)
    ctask->beats--;
}

time_t
time (time_t * t)
{
  if (t)
    *t = global_time / 100;
  return (global_time / 100);
}

time_t
mktime (struct tm * tm)
{
  if (0 >= (int) (tm->tm_mon -= 2))
    {				/* 1..12 -> 11,12,1..10 */
      tm->tm_mon += 12;		/* Puts Feb last since it has leap day */
      tm->tm_year -= 1;
    }
  return (((
	    (unsigned long) (tm->tm_year / 4 - tm->tm_year / 100 +
			     tm->tm_year / 400 + 367 * tm->tm_mon / 12 +
			     tm->tm_mday) + tm->tm_year * 365 - 719499) * 24 + tm->tm_hour	/* now have hours */
	  ) * 60 + tm->tm_min	/* now have minutes */
    ) * 60 + tm->tm_sec;	/* finally seconds */
}

/*
   This one installs user timer and executes user's proc periodicaly
   every interval of ms length.
 */
int
add_timer (int timer_nr, void (*handler) (), int period)
{
  u32 fl;

  if (timer_nr >= NR_TIMER)
    {
      printk ("add_timer: timer_nr>=NR_TIMER\n");
      return -1;
    }

  get_flags (fl);
  cli ();

  timer[timer_nr].flags = TIMER_RUNNING;
  timer[timer_nr].handler = handler;
  timer[timer_nr].period = (period + 9) / 10;
  timer[timer_nr].remaining = timer[timer_nr].period;

  set_flags (fl);

  return (-1);
}

int
free_timer (int timer_nr)
{
  u32 fl;

  if (timer_nr >= NR_TIMER)
    {
      printk ("free_timer: timer_nr>=NR_TIMER\n");
      return -1;
    }

  get_flags (fl);
  cli ();
  timer[timer_nr].flags = TIMER_STOPPED;
  set_flags (fl);
  return 0;
}

int
adjust_timer (int timer_nr, int period)
{
  u32 fl;

  if (timer_nr >= NR_TIMER)
    return -1;

  get_flags (fl);
  cli ();
  timer[timer_nr].period = period;
  set_flags (fl);

  return 0;
}

/* WRITEME: overflowing could be checked... */
int
sys_sleep (int seconds)
{
  time_t final;

  final = global_time + seconds * 1000;
  while (final > global_time)
    schedule();

  return 0;
}

void
timer_init ()
{
  setattr (ATTR (GREEN, BLACK));
  printf ("Time Machine\n");
  setattr (DEFAULT_ATTR);

  gtm.tm_sec = CMOS_R (0x00);
  gtm.tm_min = CMOS_R (0x02);
  gtm.tm_hour = CMOS_R (0x04);
  gtm.tm_mday = CMOS_R (0x07);
  gtm.tm_mon = CMOS_R (0x08);
  gtm.tm_year = CMOS_R (0x09);

  BCD2BIN (gtm.tm_sec);
  BCD2BIN (gtm.tm_min);
  BCD2BIN (gtm.tm_hour);
  BCD2BIN (gtm.tm_mday);
  BCD2BIN (gtm.tm_mon);
  BCD2BIN (gtm.tm_year);
  gtm.tm_mon--;

  printf ("  CMOS time: year=%02d mon=%d mday=%d %02d:%02d:%02d\n",
	  gtm.tm_year, gtm.tm_mon, gtm.tm_mday, gtm.tm_hour, gtm.tm_min,
	  gtm.tm_sec);

  start_time = mktime (&gtm) * 100;	/* centiseconds */
  global_time = 0;

/*  printf ("starting timer, cmos time [%d:%d:%d]\n",
   (start_time / (100 * 3600)) % 24, (start_time / (100 * 60)) % 60, (start_time / 100) % 60);
 */
  outb_p (0x36, 0x43);
  outb_p (LATCH & 0xFF, 0x40);
  outb_p (LATCH >> 8, 0x40);

  irq_request (TIMER_IRQ, IRQ_COMPLEX, timer_irq, "timer");
}

time_t sys_time(time_t *t)
{
  time_t res = (start_time + global_time) / 1000;
  
  if(t!=NULL)
    *t = res;

  return res;
}

int
timer_done ()
{
  outb_p (0x34, 0x43);
  outb_p (0, 0x40);
  outb_p (0, 0x40);

  irq_free (TIMER_IRQ);

  printk ("TIMER: phantom's lifelength [%d:%02d:%02d.%02d]\n",
	  global_time / (1000 * 3600), (global_time / (1000 * 60)) % 60,
	  (global_time / 1000) % 60, global_time);
}
