/*
   mm/kernel.c
   Kernel memory management routines.
 */

#include <asm/system.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <phantom/mm.h>

#define FIRST_OBJ(addr)	((struct obj_descriptor *)((ulg)(addr)+sizeof(struct page_descriptor)))
#define GET_OBJ_HEAD(addr) ((struct obj_descriptor *)(((ulg)(addr))-sizeof(struct obj_descriptor)))

struct obj_descriptor
{
  ulg flags;
  union
  {
    struct obj_descriptor *next;
    u32 length;
  }
  u;
};

struct page_descriptor
{
  struct page_descriptor *next;
  struct obj_descriptor *free_obj;
  int obj_size_nr;
  int free_objs;
};

struct obj_size_descriptor
{
  struct page_descriptor *free_page;
  struct page_descriptor *free_dma_page;
  int size;
  int objs;
  int pages_wanted;
  int bytes_allocated;
  int allocs;
};

#define MAX_OBJ_SIZE_NR		17
static struct obj_size_descriptor obj_size[MAX_OBJ_SIZE_NR] = {
  {NULL, NULL, 32, 127, 1, 0, 0},
  {NULL, NULL, 68, 60, 1, 0, 0},
  {NULL, NULL, 136, 30, 1, 0, 0},
  {NULL, NULL, 252, 16, 1, 0, 0},
  {NULL, NULL, 508, 8, 1, 0, 0},
  {NULL, NULL, 1020, 4, 1, 0, 0},
  {NULL, NULL, 2040, 2, 1, 0, 0},
  {NULL, NULL, 4080, 1, 1, 0, 0},
  {NULL, NULL, 8176, 1, 2, 0, 0},
  {NULL, NULL, 12272, 1, 3, 0, 0},
  {NULL, NULL, 16368, 1, 4, 0, 0},
  {NULL, NULL, 20464, 1, 5, 0, 0},
  {NULL, NULL, 32752, 1, 8, 0, 0},
  {NULL, NULL, 36864, 1, 9, 0, 0},
  {NULL, NULL, 65520, 1, 16, 0, 0},
  {NULL, NULL, 98288, 1, 24, 0, 0},
  {NULL, NULL, 131056, 1, 32, 0, 0}
};

int
get_obj_size_nr (int size)
{
  int i;

  size += sizeof (struct obj_descriptor);
  for (i = 0; i < MAX_OBJ_SIZE_NR; i++)
    if (size <= obj_size[i].size)
      return i;

  printk ("get_obj_size: fatal error, size=%d\n", size);
  return -1;
}

void *
kmalloc (int size, int k_flags)
{
  struct obj_descriptor *obj, *obj2;
  struct obj_size_descriptor *obj_sz;
  struct page_descriptor *m_page;
  int i, obj_size_nr, fl_dma;
  ulg flags;

  if (size > MAX_KMALLOC_SIZE)
    {
      printk ("kmalloc: MAX_KMALLOC_SIZE exceeded\n");
      return NULL;
    }

  obj_size_nr = get_obj_size_nr (size);

#ifdef KMMDEBUG
  printf ("obj_size_nr=%d\n", obj_size_nr);
#endif

  if (obj_size_nr < 0)
    {
      printk ("kmalloc: code/memory corrupt\n");
      return NULL;
    }

  obj_sz = &obj_size[obj_size_nr];
  fl_dma = k_flags & PG_DMA;

#ifdef KMMDEBUG
  if (fl_dma)
    printf ("dma alloc\n");
#endif

  get_flags (flags);
  cli ();
  m_page = fl_dma ? obj_sz->free_dma_page : obj_sz->free_page;

  if (m_page)
    {
#ifdef KMMDEBUG
      printk ("kmalloc: size=%d  found page with %d free objs\n" \
	      "free_obj=0x%x  obj_size_nr=%d  m_page=0x%p\n",
	      size, m_page->free_objs, m_page->free_obj,
	      m_page->obj_size_nr, m_page);
#endif

      obj = m_page->free_obj;
      if (obj->flags != OBJ_FREE)
	{
	  printk ("kmalloc: code/memory corrupt ][\n" \
		  "  %d bytes to alloc, obj->flags=0x%x  obj_size_nr=%d\n" \
		  "  m_page=0x%p  free_obj=0x%x\n",
		  size, obj->flags, obj_size_nr, m_page, m_page->free_obj);
	  set_flags (flags);
	  return NULL;
	}

#ifdef KMMDEBUG
      printk ("obj->u.next=0x%p\n", obj->u.next);
#endif

      obj_sz->bytes_allocated += size;
      obj_sz->allocs++;

      m_page->free_obj = obj->u.next;
      m_page->free_objs--;

      if (!m_page->free_objs)
	{
	  if (fl_dma)		/* problem */
	    obj_sz->free_dma_page = NULL;
	  else
	    obj_sz->free_page = m_page->next;
	}

      obj->flags = OBJ_USED;
      obj->u.length = size;

      set_flags (flags);
      bzero ((void *) (obj + 1), size);
      return obj + 1;
    }

/* second case, no free obj pool available, so we have to alloc and map
   some free page(s) */

  i = obj_sz->pages_wanted;

#ifdef KMMDEBUG
  printk (" creating new obj pool, getting %d pages, k_flags=0x%02x\n",
	  i, k_flags);
#endif

  m_page = fl_dma ? (struct page_descriptor *) dma_pages_alloc (i) :
    (struct page_descriptor *) pages_alloc (i);

  if (!m_page)
    {
      printk ("kmalloc: out of memory\n");
      set_flags (flags);
      return NULL;
    }

#ifdef KMMDEBUG
  printk ("  ok, got free page(s)\n");
#endif

  bzero (m_page, i << PAGE_BITS);

  obj = FIRST_OBJ (m_page);
  obj->flags = fl_dma ? OBJ_USED | OBJ_DMA : OBJ_USED;
  obj->u.length = size;

  m_page->obj_size_nr = obj_size_nr;

  if (obj_sz->objs > 1)
    {
      obj = (struct obj_descriptor *) ((ulg) obj + obj_sz->size);
      m_page->free_obj = obj;
      m_page->free_objs = obj_sz->objs - 1;
      for (i = 1; i < obj_sz->objs - 1; i++)
	{
	  obj->flags = OBJ_FREE;
	  obj->u.next = (struct obj_descriptor *) ((ulg) obj + obj_sz->size);
	  obj = obj->u.next;
	}
/* the last one */
      obj->flags = OBJ_FREE;	/* this is really needed */
      obj->u.next = NULL;

      if (fl_dma)
	obj_sz->free_dma_page = m_page;
      else
	obj_sz->free_page = m_page;
    }

  obj_sz->bytes_allocated += size;
  obj_sz->allocs++;

  set_flags (flags);

#ifdef KMMDEBUG
  printk ("FIRST_OBJ (0x%p) + 1 = 0x%p\n", m_page, FIRST_OBJ (m_page) + 1);
#endif

  return FIRST_OBJ (m_page) + 1;
}

void
kfree (void *addr)
{
  struct obj_descriptor *obj;
  struct page_descriptor *m_page, *m_page2;
  struct obj_size_descriptor *obj_sz;
  int i, obj_size_nr;
  ulg flags;

  m_page = (struct page_descriptor *) ((u32) addr & ADDR_MASK);
  obj = GET_OBJ_HEAD (addr);

  if (m_page->obj_size_nr >= MAX_OBJ_SIZE_NR || !(obj->flags & OBJ_USED) ||
      (ulg) m_page->next & ~ADDR_MASK)
    {
      printk ("kfree: 0x%p was not kmalloced\n", addr);
      return;
    }

  obj_size_nr = get_obj_size_nr (obj->u.length);
  if (obj_size_nr != m_page->obj_size_nr)
    {
      printk ("kfree: wrong obj size!\n");
      return;
    }

  obj_sz = &obj_size[m_page->obj_size_nr];

  get_flags (flags);
  cli ();

  obj_sz->bytes_allocated -= obj->u.length;
  obj_sz->allocs--;

  m_page->free_objs++;

/* 'big' objs and/or whole page free? */
  if (m_page->free_objs == obj_sz->objs)
    {
/* 
   Is it a 'little' obj? There may be m_pages before... 
   Free m_page (m_page with free) linking does not make
   sense for 'big' objs.
 */
      if (obj_sz->objs > 1)
	{
	  m_page2 = obj_sz->free_page;

          if(m_page2 == NULL)
	    {
	    printk("kfree: free-list corrupt, obj_sz->free_page == NULL\n"
		   "obj_sz->size=%d\n", obj_sz->size);
	    return;
	    }

	  if (m_page2 != m_page)
	    {
	      while (m_page2 && m_page2->next != m_page)
		m_page2 = m_page2->next;
	      if (!m_page2)
		{
		  printk ("kfree: free-list of obj_size_nr %d corrupted (obj_sz->size=%d)\n" \
			  " addr=0x%x page=0x%x was not found, freeing performed\n",
			  m_page->obj_size_nr, obj_sz->size,
			  addr, m_page);
		}
	      m_page2->next = m_page->next;
	    }
	  else
	    obj_sz->free_page = m_page->next;
	}

      free_pages (TOPAGE (m_page), obj_sz->pages_wanted);
      set_flags (flags);

      return;
    }

/* 'little' objs only */
  obj->flags = OBJ_FREE;
  obj->u.next = m_page->free_obj;
  m_page->free_obj = obj;

/* first free object? link the page onto freelist */
  if(m_page->free_objs==1)
    {
    m_page->next = obj_sz->free_page;
    obj_sz->free_page = m_page;
    }

  set_flags (flags);
#ifdef KMMDEBUG
  printf ("kfree: 0x%p freed\n", addr);
#endif
}

void
mm_info ()
{
  int i;

  printf ("\n%-2s %-6s %-4s %-4s %-4s %-4s %-8s %-8s\n",
	  "nr", "size", "objs", "pwant", "bytes", "allocs", "fpg", "fdp");

  for (i = 0; i < MAX_OBJ_SIZE_NR; i++)
    {
      printf ("%-2d %-6d %-4d %-4d %-4d %-4d 0x%08X 0x%08X\n",
	      i, obj_size[i].size, obj_size[i].objs, obj_size[i].pages_wanted,
	      obj_size[i].bytes_allocated, obj_size[i].allocs,
	      obj_size[i].free_page, obj_size[i].free_dma_page);
    }
}
