/*
   mm/user.h
   User level mm functions: user_get_pages, user_free_pages.
 */

#include <phantom/mm.h>
#include <sys/types.h>
#include <errno.h>
#include <phantom/sched.h>
#include <phantom/lock.h>

void *
sys_user_get_pages (int count)
{
  u32 *pt;
  int table, entry, start_page, line, count_backup = count;

#ifdef BRKDEBUG
  printk ("sys_user_get_pages: count=%d\n", count);
#endif

  if (count <= 0)
    {
      printk ("map_pages: error, count=%d  start_page=0x%x\n",
	      count, start_page);
      return NULL;
    }

/* choose start point */
  table = PAGE2TABLE (USER_LOWEST_PAGE);
  entry = 0;
  line = 0;

  for (;;)
    {
      pt = PT (table);
      if (!pt)
	{
	  if (!line)
	    start_page = TE2PAGE (table, entry);

	  line += 1024;
	  if (line >= count)
	    goto found_page;

	  table++;
	  continue;
	}

      while (line < count && entry < 1024)
	{
	  if (pt[entry])
	    {
	      line = 0;
	    }
	  else
	    {
	      if (!line)
		start_page = TE2PAGE (table, entry);

	      line++;
	      if (line >= count)
		goto found_page;
	    }

	  entry++;
	}

      entry = 0;
      table++;
    }

  printk ("sys_user_get_pages: nearly impossible to get here... (count=%d)\n",
	  count);

  return NULL;

found_page:
#ifdef USERMMDEBUG
  printk ("sys_user_get_pages: ok, start_page=0x%x, count=%d\n",
	  start_page, count);
#endif

/* ok, all is free, check passed */
  table = PAGE2TABLE (start_page);
  entry = PAGE2ENTRY (start_page);

  for (;;)
    {
      if (!PDE (table))
	{
/* missing table, oughta create it */
#ifdef USERMMDEBUG
	  printk ("sys_user_get_pages: creating missing table 0x%x\n", table);
#endif
	  pt = (u32 *) page_alloc ();
	  bzero (pt, PAGE_SIZE);
	  PDE (table) = (u32) pt | PGS_USER | PGS_WRITEABLE | PGS_PRESENT;
	  invalidate ();
	}
      else
	pt = PT (table);

      while (entry < 1024)
	{
	  pt[entry] = page_alloc () | PGS_USER | PGS_WRITEABLE | PGS_PRESENT;
	  entry++;
	  count--;
	  if (!count)
	    {
	      invalidate ();
#ifdef USERMMDEBUG
	      printk ("sys_user_get_pages: ok, all done...\n");
#endif
	      ctask->nr_userpages += count_backup;
	      return (void *) TOADDR (start_page);
	    }
	}
      entry = 0;
      table++;
    }
}

int
sys_user_free_pages (void *addr, int count)
{
  int i;
  u32 *pt;
  int table, entry, page;
  int low, high;

  if ((u32) addr & FLAGS_MASK)
    {
      printk ("sys_user_free_pages: invalid addr given 0x%x\n", addr);
      return -1;
    }

  if (count > ctask->nr_userpages)
    {
      printk ("sys_user_free_pages: count > nr_userpages, egghead\n");
      return -1;
    }

  page = TOPAGE (addr);

#ifdef USERMMDEBUG
  printk ("sys_user_free_pages: page=0x%x  count=%d\n", page, count);
#endif

  single_code ();

  if (page < USER_AREA)		/* kernel area */
    {
      multi_code ();

      printk
	("sys_user_free_pages: kernel area, no way (page=0x%x, count=%d)\n",
	 page, count);

      return -1;
    }

/* user area */
#ifdef USERMMDEBUG
  printk ("sys_user_free_pages: ua, page=0x%x  count=%d\n", page, count);
#endif

  table = PAGE2TABLE (page);
  entry = PAGE2ENTRY (page);

  for (;;)
    {
      pt = PT (table);
      if (!pt)
	printk ("sys_user_free_pages: pid=%d, table 0x%x not defined\n",
		ctask->pid, table);

      low = entry;
      while (count && entry < 1024)
	{
	  ctask->nr_userpages--;
	  set_map (TOPAGE (pt[entry]), 1, PG_FREE);
	  pt[entry] = 0;	/* unlink */
	  entry++;
	  count--;
	}

      high = entry;

/* whole free table check */
      for (entry = 0; entry < low; entry++)
	if (pt[entry] & ADDR_MASK)
	  break;

      if (entry == low)		/* scan high area, too */
	{
	  for (entry = high; entry < 1024; entry++)
	    if (pt[entry] & ADDR_MASK)
	      break;

	  if (entry == 1024)
	    {
	      printk
		("sys_user_free_pages: table 0x%x found free, chopping...\n",
		 table);
/* use special counter for these... (not nr_userpages) */
	      set_map (TOPAGE (PDE (table)), 1, PG_FREE);
	      PDE (table) = 0;
	    }
	}

      if (!count)
	{
	  multi_code ();
	  return;
	}

      table++;
      entry = 0;
    }
}
