/*
   net/drv/ne2000.c
   Phantom NE2000 packet driver.
 */

#include <asm/irq.h>
#include <asm/io.h>
#include <asm/regdump.h>
#include <asm/system.h>
#include <phantom/protocol.h>
#include <phantom/netdev.h>
#include <phantom/8390.h>
#include <phantom/time.h>

#define NE_BASE		(dev->base_addr)
#define NE_IRQ		(dev->irq)
#define NE_STATUS	(dev->u.ne)

#define NE_CMD	 	0x00
#define NE_DATAPORT	0x10	/* NatSemi-defined port window offset. */
#define NE_IO_EXTENT	0x20

#define NE1SM_START_PG	0x20	/* First page of TX buffer */
#define NE1SM_STOP_PG 	0x40	/* Last page +1 of RX ring */
#define NESM_START_PG	0x40	/* First page of TX buffer */
#define NESM_STOP_PG	0x80	/* Last page +1 of RX ring */


#define NE_DATA		0x10	/* NE2000 port window */
#define NE_RESET	0x1F	/* issue a read for reset */

#define SM_TSTART_PG	0x40	/* First page of TX buffer */
#define SM_RSTART_PG	0x46	/* Starting page of RX ring */
#define SM_RSTOP_PG	0x80	/* Last page +1 of RX ring */

#define EI_PINGPONG

struct device my_ne = { 0, };
struct ne_info ne_i = { 0, };

void
reset_8390 (struct device *dev)
{
  unsigned char c;

  c = inb_p (NE_BASE + NE_RESET);
  sys_sleep (250);
  outb_p (c, NE_BASE + NE_RESET);
  sys_sleep (250);
}

void
NS8390_init (struct device *dev, int startp)
{
  int i;
  int e8390_base = NE_BASE;
  int endcfg = NE_STATUS->word16 ? (0x48 | ENDCFG_WTS) : 0x48;
  struct ne_info *ei_local = dev->u.ne;
  u32 flg;

  /* Follow National Semi's recommendations for initing the DP83902. */
  outb_p (E8390_NODMA + E8390_PAGE0 + E8390_STOP, e8390_base);	/* 0x21 */
  outb_p (endcfg, e8390_base + EN0_DCFG);	/* 0x48 or 0x49 */
  /* Clear the remote byte count registers. */
  outb_p (0x00, e8390_base + EN0_RCNTLO);
  outb_p (0x00, e8390_base + EN0_RCNTHI);
  /* Set to monitor and loopback mode -- this is vital!. */
  outb_p (E8390_RXOFF, e8390_base + EN0_RXCR);	/* 0x20 */
  outb_p (E8390_TXOFF, e8390_base + EN0_TXCR);	/* 0x02 */
  /* Set the transmit page and receive ring. */
  outb_p (ei_local->tx_start_page, e8390_base + EN0_TPSR);
  ei_local->tx1 = ei_local->tx2 = 0;
  outb_p (ei_local->rx_start_page, e8390_base + EN0_STARTPG);
  outb_p (ei_local->stop_page - 1, e8390_base + EN0_BOUNDARY);	/* 3c503 says 0x3f,NS0x26 */
  ei_local->current_page = ei_local->rx_start_page;	/* assert boundary+1 */
  outb_p (ei_local->stop_page, e8390_base + EN0_STOPPG);
  /* Clear the pending interrupts and mask. */
  outb_p (0xFF, e8390_base + EN0_ISR);
  outb_p (0x00, e8390_base + EN0_IMR);

  /* Copy the station address into the DS8390 registers,
     and set the multicast hash bitmap to receive all multicasts. */

  get_flags (flg);
  cli ();

  outb_p (E8390_NODMA + E8390_PAGE1 + E8390_STOP, e8390_base);	/* 0x61 */
  for (i = 0; i < 6; i++)
    {
      outb_p (dev->hw_addr[i], e8390_base + EN1_PHYS + i);
    }
  /* Initialize the multicast list to accept-all.  If we enable multicast
     the higher levels can do the filtering. */
  for (i = 0; i < 8; i++)
    outb_p (0xff, e8390_base + EN1_MULT + i);

  outb_p (ei_local->rx_start_page, e8390_base + EN1_CURPAG);
  outb_p (E8390_NODMA + E8390_PAGE0 + E8390_STOP, e8390_base);

  set_flags (flg);

  ei_local->tx1 = ei_local->tx2 = 0;
  ei_local->txing = 0;
  if (startp)
    {
      outb_p (0xff, e8390_base + EN0_ISR);
      outb_p (ENISR_ALL, e8390_base + EN0_IMR);
      outb_p (E8390_NODMA + E8390_PAGE0 + E8390_START, e8390_base);
      outb_p (E8390_TXCONFIG, e8390_base + EN0_TXCR);	/* xmit on. */
      /* 3c503 TechMan says rxconfig only after the NIC is started. */
      outb_p (E8390_RXCONFIG, e8390_base + EN0_RXCR);	/* rx on,  */
    }
}

void
block_input (struct device *dev, int len, char *buff, int ring_offset)
{
  if (NE_STATUS->dmaing)
    {
      printk ("block_input: dmaing set\n");
      return;
    }

  NE_STATUS->dmaing |= 1;

  outb_p (E8390_NODMA + E8390_PAGE0 + E8390_START, NE_BASE + NE_CMD);
  outb_p (len & 0xFF, NE_BASE + EN0_RCNTLO);
  outb_p (len >> 8, NE_BASE + EN0_RCNTHI);

  outb_p (ring_offset & 0xff, NE_BASE + EN0_RSARLO);
  outb_p (ring_offset >> 8, NE_BASE + EN0_RSARHI);

  outb_p (E8390_RREAD + E8390_START, NE_BASE + NE_CMD);

  if (dev->u.ne->word16)
    {
      insw (NE_BASE + NE_DATAPORT, buff, len >> 1);
      if (len & 0x01)
	{
	  buff[len - 1] = inb (NE_BASE + NE_DATAPORT);
	}
    }
  else
    {
      insb (NE_BASE + NE_DATAPORT, buff, len);
    }

  outb_p (ENISR_RDC, NE_BASE + EN0_ISR);	/* Ack intr. */

  NE_STATUS->dmaing &= ~1;
}

void
block_output (struct device *dev, int len, char *buff, int start_page)
{
  int dma_start;

  if (NE_STATUS->dmaing)
    {
      printk ("block_output: dmaing=0x%x\n", NE_STATUS->dmaing);
      return;
    }

  NE_STATUS->dmaing |= 1;

  if (NE_STATUS->word16 && (len & 0x01))
    len++;

  outb_p (E8390_PAGE0 + E8390_START + E8390_NODMA, NE_BASE + NE_CMD);

  outb_p (ENISR_RDC, NE_BASE + EN0_ISR);

  outb_p (len & 0xFF, NE_BASE + EN0_RCNTLO);
  outb_p (len >> 8, NE_BASE + EN0_RCNTHI);

  outb_p (0x00, NE_BASE + EN0_RSARLO);
  outb_p (start_page, NE_BASE + EN0_RSARHI);

  outb_p (E8390_RWRITE + E8390_START, NE_BASE + NE_CMD);

  if (NE_STATUS->word16)
    {
      outsw (NE_BASE + NE_DATAPORT, buff, len >> 1);
    }
  else
    {
      outsb (NE_BASE + NE_DATAPORT, buff, len);
    }

  dma_start = global_time;
  while ((inb_p (NE_BASE + EN0_ISR) & ENISR_RDC) == 0)
    if (global_time - dma_start > 200)	/* 200 ms */
      {
	printk ("%s: timeout waiting for Tx RDC.\n", dev->name);
	reset_8390 (dev);
	NS8390_init (dev, 1);
	break;
      }

  outb_p (ENISR_RDC, NE_BASE + EN0_ISR);	/* Ack intr. */
  NE_STATUS->dmaing &= ~0x01;
}

void
ei_receive (struct device *dev)
{
  int e8390_base = NE_BASE;
  struct ne_info *ei_local = dev->u.ne;
  int rxing_page, this_frame, next_frame, current_offset;
  int rx_pkt_count = 0;
  struct e8390_pkt_hdr rx_frame;
  int num_rx_pages = ei_local->stop_page - ei_local->rx_start_page;

  while (++rx_pkt_count < 10)
    {
      int pkt_len;

      /* Get the rx page (incoming packet pointer). */
      outb_p (E8390_NODMA + E8390_PAGE1, e8390_base + E8390_CMD);
      rxing_page = inb_p (e8390_base + EN1_CURPAG);
      outb_p (E8390_NODMA + E8390_PAGE0, e8390_base + E8390_CMD);

      /* Remove one frame from the ring.  Boundary is alway a page behind. */
      this_frame = inb_p (e8390_base + EN0_BOUNDARY) + 1;
      if (this_frame >= ei_local->stop_page)
	this_frame = ei_local->rx_start_page;

      /* Someday we'll omit the previous, iff we never get this message.
         (There is at least one clone claimed to have a problem.)  */
      if (this_frame != ei_local->current_page)
	printk ("%s: mismatched read page pointers %2x vs %2x.\n",
		dev->name, this_frame, ei_local->current_page);

      if (this_frame == rxing_page)	/* Read all the frames? */
	break;			/* Done for now */

      current_offset = this_frame << 8;
      block_input (dev, sizeof (rx_frame), (char *) &rx_frame,
		   current_offset);

      pkt_len = rx_frame.count - sizeof (rx_frame);

      next_frame = this_frame + 1 + ((pkt_len + 4) >> 8);

      /* Check for bogosity warned by 3c503 book: the status byte is never
         written.  This happened a lot during testing! This code should be
         cleaned up someday. */
      if (rx_frame.next != next_frame
	  && rx_frame.next != next_frame + 1
	  && rx_frame.next != next_frame - num_rx_pages
	  && rx_frame.next != next_frame + 1 - num_rx_pages)
	{
	  ei_local->current_page = rxing_page;
	  outb (ei_local->current_page - 1, e8390_base + EN0_BOUNDARY);
	  ei_local->stat.rx_errors++;
	  continue;
	}

      if (pkt_len < 60 || pkt_len > 1518)
	{
	  printk ("%s: bogus packet size: %d, status=%#2x nxpg=%#2x.\n",
		  dev->name, rx_frame.count, rx_frame.status, rx_frame.next);
	  ei_local->stat.rx_errors++;
	}
      else if ((rx_frame.status & 0x0F) == ENRSR_RXOK)
	{
	  struct pkt_head *pkt;
	  int pkt_size =
	    sizeof (struct pkt_head) + pkt_len - sizeof (struct eth);

	  pkt = pkt_alloc (dev, pkt_size);
	  if (pkt == NULL)
	    {
	      printk ("%s: Couldn't allocate a pkt_buff of size %d.\n",
		      dev->name, pkt_size);
	      ei_local->stat.rx_dropped++;
	      break;
	    }
	  else
	    {
	      pkt->data_len = pkt_len;
/* pkt->data is done by pkt_alloc */

	      block_input (dev, pkt_len, (char *) (pkt + 1),
			   current_offset + sizeof (rx_frame));

	      pkt_rx (pkt);

	      ei_local->stat.rx_packets++;
	    }
	}
      else
	{
	  int errs = rx_frame.status;
	  printk ("%s: bogus packet: status=%#2x nxpg=%#2x size=%d\n",
		  dev->name, rx_frame.status, rx_frame.next, rx_frame.count);
	  if (errs & ENRSR_FO)
	    ei_local->stat.rx_fifo_errors++;
	}
      next_frame = rx_frame.next;

      /* This _should_ never happen: it's here for avoiding bad clones. */
      if (next_frame >= ei_local->stop_page)
	{
	  printk ("%s: next frame inconsistency, %#2x..", dev->name,
		  next_frame);
	  next_frame = ei_local->rx_start_page;
	}
      ei_local->current_page = next_frame;
      outb (next_frame - 1, e8390_base + EN0_BOUNDARY);
    }
  /* If any worth-while packets have been received, dev_rint()
     has done a mark_bh(INET_BH) for us and will work on them
     when we get to the bottom-half routine. */

  /* Bug alert!  Reset ENISR_OVER to avoid spurious overruns! */
  outb_p (ENISR_RX + ENISR_RX_ERR + ENISR_OVER, e8390_base + EN0_ISR);
  return;
}

static void
ei_rx_overrun (struct device *dev)
{
  int e8390_base = NE_BASE;
  int reset_start_time = global_time;
  struct ne_info *ei_local = dev->u.ne;

  /* We should already be stopped and in page0.  Remove after testing. */
  outb_p (E8390_NODMA + E8390_PAGE0 + E8390_STOP, e8390_base + E8390_CMD);

  ei_local->stat.rx_over_errors++;

  /* The old Biro driver does dummy = inb_p( RBCR[01] ); at this point.
     It might mean something -- magic to speed up a reset?  A 8390 bug? */

  /* Wait for the reset to complete.  This should happen almost instantly,
     but could take up to 1.5msec in certain rare instances.  There is no
     easy way of timing something in that range, so we use 'jiffies' as
     a sanity check. */
  while ((inb_p (e8390_base + EN0_ISR) & ENISR_RESET) == 0)
    if (global_time - reset_start_time > 1)
      {
	printk ("%s: reset did not complete at ei_rx_overrun.\n", dev->name);
	NS8390_init (dev, 1);
	return;
      }

  /* Remove packets right away. */
  ei_receive (dev);

  outb_p (0xff, e8390_base + EN0_ISR);
  /* Generic 8390 insns to start up again, same as in open_8390(). */
  outb_p (E8390_NODMA + E8390_PAGE0 + E8390_START, e8390_base + E8390_CMD);
  outb_p (E8390_TXCONFIG, e8390_base + EN0_TXCR);	/* xmit on. */
}


/* Trigger a transmit start, assuming the length is valid. */
static void
NS8390_trigger_send (struct device *dev, unsigned int length, int start_page)
{
  int e8390_base = NE_BASE;

  NE_STATUS->txing = 1;
  outb_p (E8390_NODMA + E8390_PAGE0, e8390_base);

  if (inb_p (e8390_base) & E8390_TRANS)
    {
      printk ("%s: trigger_send() called with the transmitter busy.\n",
	      dev->name);
      return;
    }
  outb_p (length & 0xff, e8390_base + EN0_TCNTLO);
  outb_p (length >> 8, e8390_base + EN0_TCNTHI);
  outb_p (start_page, e8390_base + EN0_TPSR);
  outb_p (E8390_NODMA + E8390_TRANS + E8390_START, e8390_base);
  return;
}

void
ei_tx_intr (struct device *dev)
{
  int e8390_base = NE_BASE;
  int status = inb (e8390_base + EN0_TSR);
  struct ne_info *ei_local = dev->u.ne;

  outb_p (ENISR_TX, e8390_base + EN0_ISR);	/* Ack intr. */

#ifdef EI_PINGPONG

  /*
   * There are two Tx buffers, see which one finished, and trigger
   * the send of another one if it exists.
   */
  ei_local->txqueue--;
  if (ei_local->tx1 < 0)
    {
      if (ei_local->lasttx != 1 && ei_local->lasttx != -1)
	printk ("%s: bogus last_tx_buffer %d, tx1=%d.\n",
		dev->name, ei_local->lasttx, ei_local->tx1);
      ei_local->tx1 = 0;
      ei_local->tbusy = 0;
      if (ei_local->tx2 > 0)
	{
	  ei_local->txing = 1;
	  NS8390_trigger_send (dev, ei_local->tx2,
			       ei_local->tx_start_page + 6);
	  ei_local->trans_start = global_time;
	  ei_local->tx2 = -1, ei_local->lasttx = 2;
	}
      else
	ei_local->lasttx = 20, ei_local->txing = 0;
    }
  else if (ei_local->tx2 < 0)
    {
      if (ei_local->lasttx != 2 && ei_local->lasttx != -2)
	printk ("%s: bogus last_tx_buffer %d, tx2=%d.\n",
		dev->name, ei_local->lasttx, ei_local->tx2);
      ei_local->tx2 = 0;
      ei_local->tbusy = 0;
      if (ei_local->tx1 > 0)
	{
	  ei_local->txing = 1;
	  NS8390_trigger_send (dev, ei_local->tx1, ei_local->tx_start_page);
	  ei_local->trans_start = global_time;
	  ei_local->tx1 = -1;
	  ei_local->lasttx = 1;
	}
      else
	ei_local->lasttx = 10, ei_local->txing = 0;
    }
  else
    printk ("%s: unexpected TX-done interrupt, lasttx=%d.\n",
	    dev->name, ei_local->lasttx);

#else /* EI_PINGPONG */
  /*
   *  Single Tx buffer: mark it free so another packet can be loaded.
   */
  ei_local->txing = 0;
  dev->tbusy = 0;
#endif

  /* Minimize Tx latency: update the statistics after we restart TXing. */
  if (status & ENTSR_COL)
    ei_local->stat.collisions++;
  if (status & ENTSR_PTX)
    ei_local->stat.tx_packets++;
  else
    {
      ei_local->stat.tx_errors++;
      if (status & ENTSR_ABT)
	ei_local->stat.tx_aborted_errors++;
      if (status & ENTSR_CRS)
	ei_local->stat.tx_carrier_errors++;
      if (status & ENTSR_FU)
	ei_local->stat.tx_fifo_errors++;
      if (status & ENTSR_CDH)
	ei_local->stat.tx_heartbeat_errors++;
      if (status & ENTSR_OWC)
	ei_local->stat.tx_window_errors++;
    }

/*    mark_bh (NET_BH);	*/
}

int
ei_start_xmit (struct pkt_head *pkt)
{
  struct device *dev = pkt->dev;
  int e8390_base = NE_BASE;
  struct ne_info *ei_local = dev->u.ne;
  int length, send_length, output_page;

/*
 *  We normally shouldn't be called if dev->tbusy is set, but the
 *  existing code does anyway. If it has been too long since the
 *  last Tx, we assume the board has died and kick it.
 */

  if (ei_local->tbusy)
    {				/* Do timeouts, just like the 8003 driver. */
      int txsr = inb (e8390_base + EN0_TSR), isr;
      int tickssofar = global_time - ei_local->trans_start;
      if (tickssofar < TX_TIMEOUT ||
	  (tickssofar < (TX_TIMEOUT + 5) && !(txsr & ENTSR_PTX)))
	{
	  return 1;
	}
      isr = inb (e8390_base + EN0_ISR);

      /*
       * Note that if the Tx posted a TX_ERR interrupt, then the
       * error will have been handled from the interrupt handler.
       * and not here.
       */

      printk ("%s: Tx timed out, %s TSR=%#2x, ISR=%#2x, t=%d.\n",
	      dev->name, (txsr & ENTSR_ABT) ? "excess collisions." :
	      (isr) ? "lost interrupt?" : "cable problem?", txsr, isr,
	      tickssofar);

      /* Try to restart the card.  Perhaps the user has fixed something. */
      reset_8390 (dev);
      NS8390_init (dev, 1);
      ei_local->trans_start = global_time;
    }

  length = pkt->data_len + sizeof (struct eth);

  /* Mask interrupts from the ethercard. */
  outb_p (0x00, e8390_base + EN0_IMR);
  if (ei_local->interrupt)
    {
      printk ("%s: Tx request while isr active.\n", dev->name);
      outb_p (ENISR_ALL, e8390_base + EN0_IMR);
      return 1;
    }
  ei_local->irqlock = 1;

  send_length = ETH_ZLEN < length ? length : ETH_ZLEN;

#ifdef EI_PINGPONG

  /*
   * We have two Tx slots available for use. Find the first free
   * slot, and then perform some sanity checks. With two Tx bufs,
   * you get very close to transmitting back-to-back packets. With
   * only one Tx buf, the transmitter sits idle while you reload the
   * card, leaving a substantial gap between each transmitted packet.
   */

  if (ei_local->tx1 == 0)
    {
      output_page = ei_local->tx_start_page;
      ei_local->tx1 = send_length;
      if (ei_local->tx2 > 0)
	printk ("%s: idle transmitter tx2=%d, lasttx=%d, txing=%d.\n",
		dev->name, ei_local->tx2, ei_local->lasttx, ei_local->txing);
    }
  else if (ei_local->tx2 == 0)
    {
      output_page = ei_local->tx_start_page + TX_1X_PAGES;
      ei_local->tx2 = send_length;
      if (ei_local->tx1 > 0)
	printk ("%s: idle transmitter, tx1=%d, lasttx=%d, txing=%d.\n",
		dev->name, ei_local->tx1, ei_local->lasttx, ei_local->txing);
    }
  else
    {				/* We should never get here. */
      printk ("%s: No Tx buffers free! irq=%d tx1=%d tx2=%d last=%d\n",
	      dev->name, ei_local->interrupt, ei_local->tx1, ei_local->tx2,
	      ei_local->lasttx);
      ei_local->irqlock = 0;
      ei_local->tbusy = 1;
      outb_p (ENISR_ALL, e8390_base + EN0_IMR);
      return 1;
    }

  /*
   * Okay, now upload the packet and trigger a send if the transmitter
   * isn't already sending. If it is busy, the interrupt handler will
   * trigger the send later, upon receiving a Tx done interrupt.
   */

  block_output (dev, length, (char *) pkt->hw_data, output_page);
  if (!ei_local->txing)
    {
      ei_local->txing = 1;
      NS8390_trigger_send (dev, send_length, output_page);
      ei_local->trans_start = global_time;
      if (output_page == ei_local->tx_start_page)
	{
	  ei_local->tx1 = -1;
	  ei_local->lasttx = -1;
	}
      else
	{
	  ei_local->tx2 = -1;
	  ei_local->lasttx = -2;
	}
    }
  else
    ei_local->txqueue++;

  ei_local->tbusy = (ei_local->tx1 && ei_local->tx2);

#else /* EI_PINGPONG */

  /*
   * Only one Tx buffer in use. You need two Tx bufs to come close to
   * back-to-back transmits. Expect a 20 -> 25% performance hit on
   * reasonable hardware if you only use one Tx buffer.
   */

  block_output (dev, length, pkt->hw_data, ei_local->tx_start_page);
  ei_local->txing = 1;
  NS8390_trigger_send (dev, send_length, ei_local->tx_start_page);
  ei_local->trans_start = global_time;
  ei_local->tbusy = 1;

#endif /* EI_PINGPONG */

  /* Turn 8390 interrupts back on. */
  ei_local->irqlock = 0;
  outb_p (ENISR_ALL, e8390_base + EN0_IMR);
  pkt_free (pkt);

  return 0;			/* ok, packet sent */
}

void
ei_interrupt (int irq, struct regdump *rd)
{
  struct device *dev = /*(struct device *) (irq2dev_map[irq]); */ &my_ne;
  int e8390_base;
  int interrupts, boguscount = 0;
  struct ne_info *ei_local;

  if (!dev)
    {
      printk ("ei_interrupt: irq %d for unknown device.\n", irq);
      return;
    }
  e8390_base = NE_BASE;
  ei_local = dev->u.ne;
  if (ei_local->interrupt || ei_local->irqlock)
    {
      /* The "irqlock" check is only for testing. */
      sti ();
      printk (ei_local->irqlock
	      ?
	      "%s: Interrupted while interrupts are masked! isr=%#2x imr=%#2x.\n"
	      : "%s: Reentering the interrupt handler! isr=%#2x imr=%#2x.\n",
	      dev->name, inb_p (e8390_base + EN0_ISR),
	      inb_p (e8390_base + EN0_IMR));
      return;
    }

  ei_local->interrupt = 1;
  sti ();			/* Allow other interrupts. */

  /* Change to page 0 and read the intr status reg. */
  outb_p (E8390_NODMA + E8390_PAGE0, e8390_base + E8390_CMD);

  /* !!Assumption!! -- we stay in page 0.      Don't break this. */
  while ((interrupts = inb_p (e8390_base + EN0_ISR)) != 0 && ++boguscount < 5)
    {
      if (interrupts & ENISR_RDC)
	{
	  /* Ack meaningless DMA complete. */
	  outb_p (ENISR_RDC, e8390_base + EN0_ISR);
	}
      if (interrupts & ENISR_OVER)
	{
	  ei_rx_overrun (dev);
	}
      else if (interrupts & (ENISR_RX + ENISR_RX_ERR))
	{
	  /* Got a good (?) packet. */
	  ei_receive (dev);
	}
      /* Push the next to-transmit packet through. */
      if (interrupts & ENISR_TX)
	{
	  ei_tx_intr (dev);
	}
      else if (interrupts & ENISR_COUNTERS)
	{
	  struct ne_info *ei_local = dev->u.ne;
	  ei_local->stat.rx_frame_errors += inb_p (e8390_base + EN0_COUNTER0);
	  ei_local->stat.rx_crc_errors += inb_p (e8390_base + EN0_COUNTER1);
	  ei_local->stat.rx_missed_errors +=
	    inb_p (e8390_base + EN0_COUNTER2);
	  outb_p (ENISR_COUNTERS, e8390_base + EN0_ISR);	/* Ack intr. */
	}

      /* Ignore the transmit errs and reset intr for now. */
      if (interrupts & ENISR_TX_ERR)
	{
	  outb_p (ENISR_TX_ERR, e8390_base + EN0_ISR);	/* Ack intr. */
	}
      outb_p (E8390_NODMA + E8390_PAGE0 + E8390_START,
	      e8390_base + E8390_CMD);
    }

  if (interrupts)
    {
      printk ("%s: unknown interrupt %#2x\n", dev->name, interrupts);
      outb_p (E8390_NODMA + E8390_PAGE0 + E8390_START,
	      e8390_base + E8390_CMD);
      outb_p (0xff, e8390_base + EN0_ISR);	/* Ack. all intrs. */
    }

  ei_local->interrupt = 0;
  return;
}

int
ne_exists (struct device *dev)
{
  int reg0, regd, i;
  u8 SA_prom[32];
  int start_page, stop_page;
  int wordlen = 2;

  outb_p (E8390_NODMA + E8390_PAGE1 + E8390_STOP, NE_BASE + E8390_CMD);
  regd = inb_p (NE_BASE + 0x0d);
  outb_p (0xff, NE_BASE + 0x0d);
  outb_p (E8390_NODMA + E8390_PAGE0, NE_BASE + E8390_CMD);
  inb_p (NE_BASE + EN0_COUNTER0);	/* Clear the counter by reading. */
  if (inb_p (NE_BASE + EN0_COUNTER0) != 0)
    {
      outb_p (reg0, NE_BASE);
      outb (regd, NE_BASE + 0x0d);	/* Restore the old values. */
      return 0;
    }

  printk ("NE2000 probe at 0x%X\n", NE_BASE);

  outb_p (E8390_NODMA + E8390_PAGE0 + E8390_STOP, NE_BASE + E8390_CMD);	/* 0x21 */
  outb_p (0x48, NE_BASE + EN0_DCFG);	/* byte-wide access */

  /* Clear the remote byte count registers. */
  outb_p (0x00, NE_BASE + EN0_RCNTLO);
  outb_p (0x00, NE_BASE + EN0_RCNTHI);

  /* Clear the pending interrupts and mask. */
  outb_p (0x00, NE_BASE + EN0_IMR);
  outb_p (0xFF, NE_BASE + EN0_ISR);

  /* Set to monitor and loopback mode. */
  outb_p (E8390_RXOFF, NE_BASE + EN0_RXCR);	/* 0x20  Set to monitor */
  outb_p (E8390_TXOFF, NE_BASE + EN0_TXCR);	/* 0x02  and loopback mode. */

  outb_p (32, NE_BASE + EN0_RCNTLO);
  outb_p (0x00, NE_BASE + EN0_RCNTHI);

/* DMA starting at 0x0000. */
  outb_p (0x00, NE_BASE + EN0_RSARLO);
  outb_p (0x00, NE_BASE + EN0_RSARHI);

  outb_p (E8390_RREAD + E8390_START, NE_BASE + E8390_CMD);

  for (i = 0; i < 32; i += 2)
    {
      SA_prom[i] = inb (NE_BASE + NE_DATAPORT);
      SA_prom[i + 1] = inb (NE_BASE + NE_DATAPORT);
      if (SA_prom[i] != SA_prom[i + 1])
	wordlen = 1;
    }

  if (wordlen == 2)
    {
      /* We must set the 8390 for word mode. */
      outb_p (0x49, NE_BASE + EN0_DCFG);
      /* Un-double the SA_prom values. */
      for (i = 0; i < 16; i++)
	SA_prom[i] = SA_prom[i + i];
    }

  if (SA_prom[14] != 0x57 || SA_prom[15] != 0x57)
    {
      printk ("No NE2000 at 0x%04X.\n", NE_BASE);
      return -1;
    }

  start_page = NESM_START_PG;
  stop_page = NESM_STOP_PG;

  if (dev->irq < 2)
    {
      irq_probe_start ();
      outb_p (0x50, NE_BASE + EN0_IMR);	/* Enable one interrupt. */
      outb_p (0x00, NE_BASE + EN0_RCNTLO);
      outb_p (0x00, NE_BASE + EN0_RCNTHI);
      outb_p (E8390_RREAD + E8390_START, NE_BASE);	/* Trigger it... */
      outb_p (0x00, NE_BASE + EN0_IMR);	/* Mask it again. */
      dev->irq = irq_probe_end ();
    }
  else if (dev->irq == 2)
    dev->irq = 9;

  if (i = irq_request (dev->irq, IRQ_COMPLEX, ei_interrupt, "ne2000"))
    {
      printk ("ne: cannot gain irq 0x%x\n", dev->irq);
      return i;
    }

  printk ("%s: io=0x%X  irq=0x%x  hw=", dev->name, NE_BASE, NE_IRQ);
  for (i = 0; i < ETHER_ADDR_LEN; i++)
    {
      dev->hw_addr[i] = SA_prom[i];
      printk ("%02x", SA_prom[i]);
      if (i < ETHER_ADDR_LEN - 1)
	printk (":");
    }

  printk ("  ip=");
  for (i = 0; i < IPADDR_LEN; i++)
    {
      printk ("%d", dev->ip_addr[i]);
      if (i < IPADDR_LEN - 1)
	printk (".");
    }

  printk ("\n");

  NE_STATUS->tx_start_page = start_page;
  NE_STATUS->rx_start_page = start_page + TX_PAGES;
  NE_STATUS->stop_page = stop_page;
  NE_STATUS->word16 = wordlen == 2 ? 1 : 0;

  NS8390_init (dev, 1);
  return 0;
}

int
ne_probe (struct device *dev)
{
  int *port;
  static int ports[] = { 0x300, 0x240, 0 };

  port = ports;
  while (*port)
    {
      if (inb_p (*port) != 0xFF)
	{
	  NE_BASE = *port;
	  if (!ne_exists (dev))
	    return 0;
	}
      port++;
    }

  return -1;
}

void
ne_init ()
{
  int i;

  my_ne.u.ne = &ne_i;

  my_ne.hw_hdr_len = sizeof (struct eth);
  my_ne.pkt_tx = &ei_start_xmit;

  for (i = 0; i < HWADDR_LEN; i++)
    my_ne.broad_addr[i] = 0xff;

  my_ne.name = "eth0";

  my_ne.ip_addr[0] = 192;
  my_ne.ip_addr[1] = 168;
  my_ne.ip_addr[2] = 0;
  my_ne.ip_addr[3] = 16;

  my_ne.mtu = 576;

  if(ne_probe (&my_ne))
    {
    printf("NE probe failed.\n");
    return;
    }
    
  printf("NE probed: %s  %d.%d.%d.%d\n", my_ne.name, 
  my_ne.ip_addr[0], my_ne.ip_addr[1], my_ne.ip_addr[2], my_ne.ip_addr[3]);
}
