#include <sys/types.h>
#include <string.h>
#include <rg/window.h>
#include <rg/rgio.h>
#include <rg/kernel.h>
#include <rg/event.h>
#include <phantom/video.h>
#include <phantom/fs.h>
#include <errno.h>
#include <phantom/lock.h>
#include <phantom/tty.h>
#include <phantom/sched.h>
#include <phantom/kb.h>
#include <phantom/keyb.h>
#include <termix.h>
#include <rg/core.h>

WINDOW *window_list = NULL;	/* first -- active window */
WINDOW *window_tail = NULL;

#define ISACTIVE(w)		((w) == window_list)

extern struct file_operations window_file_operations;

WINDOW *
sys_rg_newwin (int x, int y, int sizex, int sizey)
{
  struct file *file;
  WINDOW *w;

  w = (WINDOW *) kmalloc (sizeof (WINDOW) + sizex * sizey, 0);
  if (w == NULL)
    return NULL;

  w->fd = get_fd ();
  if (w->fd == -1)
    {
      kfree (w);
      return NULL;
    }

  file = fd2file (w->fd);

  file->f_window = w;
  file->f_op = &window_file_operations;

  w->screen = (char *) w + sizeof (WINDOW);
  bzero (w->screen, sizex * sizey);
  w->posx = x;
  w->posy = y;
  w->sizex = sizex;
  w->sizey = sizey;
  w->obsc = NULL;
  w->cursor_mode = RG_NOCURSOR;
  w->flags = W_HIDDEN;
  w->next = NULL;
  w->prev = NULL;
  w->pid = ctask->pid;
  w->curx = 0;
  w->cury = 0;

#ifdef RGDEBUG
  printk ("rg: window created (0x%x)\n", w);
#endif

  return w;
}

int
show_window (WINDOW * w)
{
#ifdef RGDEBUG
  printk ("rg: showing window 0x%x\n", w);
#endif
  w->flags &= ~W_HIDDEN;
  single_code ();
  list_head_add ((LIST *) w, (LIST **) & window_list,
		 (LIST **) & window_tail);
  multi_code ();
}

int
hide_window (WINDOW * window)
{
  WINDOW *w;

#ifdef RGDEBUG
  printk ("rg: hiding window 0x%x\n", w);
#endif
  single_code ();
  list_del ((LIST *) window, (LIST **) & window_list,
	    (LIST **) & window_tail);
  multi_code ();
  window->flags |= W_HIDDEN;

#ifdef RGDEBUG
  printk ("rg: redrawing it all...\n");
#endif

  _clrscr (rg_tty);

  w = window_tail;
  while (w != NULL)
    {
      refresh_window (w);
      w = (WINDOW *) w->prev;
    }
}

void
move_window (WINDOW * window, int newx, int newy)
{
  WINDOW *w;

  if (window == NULL)
    return;

  window->posx = newx;
  window->posy = newy;

  if (window->flags & W_HIDDEN)
    return;

/* redraw the scene ;) */
  _clrscr (rg_tty);

  w = window_tail;
  while (w != NULL)
    {
      refresh_window (w);
      w = (WINDOW *) w->prev;
    }
}

int
refresh_window (WINDOW * w)
{
  int j, sx, wsx, px, py;
  void *sb, *scr;

#ifdef RGDEBUG
  printk ("rg: refreshing window 0x%x\n", w);
#endif

  if (w->flags & W_HIDDEN)
    return 0;

  if (w->flags & W_OBSCURED)
    {				/* oh, shit... */
      printk ("cannot draw correctly obscured windows yet...\n");
//      return 0;
    }

  px = w->posx;
  py = w->posy;
  sx = w->sizex;
  wsx = rg_tty->size_x;
  sb = (void *) rg_tty->screen;
  scr = (void *) w->screen;

  for (j = 0; j < w->sizey; j++)	/* whole window is inscreen */
    memcpy (sb + (((j + py) * wsx + px) << 1), scr + ((j * sx) << 1),
	    sx << 1);

  if (ISACTIVE (w))
    {
      if (w->cursor_mode != RG_NOCURSOR)
	{
	  _gotoxy (rg_tty, w->posx + w->curx, w->posy + w->cury);
	}
    }

  return 0;
}

int
top_window (WINDOW * w)
{

}

void
paint_cursor (WINDOW * w)
{
  if (w != NULL && w == window_list)
    {
      rg_tty->cursor_mode = w->cursor_mode;
      _shapecursor (rg_tty);
    }
}

int
cursor_mode (WINDOW * w, int mode)
{
  switch (mode)
    {
    case RG_NOCURSOR:
    case RG_SMALLCURSOR:
    case RG_BIGCURSOR:
      w->cursor_mode = mode;
      paint_cursor (w);
      return 0;
    }

  return -1;
}

int
window_ioctl (struct file *file, unsigned int cmd, unsigned int arg)
{
  switch (cmd)
    {
    case RGIOGEVENT:
      return rg_getevent (file->f_window, (EVENT *) arg);
    case RGIOWREFRESH:
      return refresh_window (file->f_window);
    case RGIOWSHOW:
      return show_window (file->f_window);
    case RGIOWHIDE:
      return hide_window (file->f_window);
    case RGIOSETROOT:
      return 0;
    case RGIOCMODE:
      return cursor_mode (file->f_window, arg);
    }

  return -EINVAL;
}

int
window_close (struct file *file)
{
  WINDOW *w;

  if (file != NULL && (w = file->f_window) != NULL)
    {
      if (!(w->flags & W_HIDDEN))
	hide_window (w);

      kfree (w);
    }

  return 0;
}

struct file_operations window_file_operations = {
  NULL,				/* lseek */
  NULL,				/* read */
  NULL,				/* write */
  NULL,				/* readdir */
  window_ioctl,			/* ioctl */
  NULL,				/* open */
  window_close			/* close */
};

int
switch_window (int d)
{
  WINDOW *tmp;

  if (!d || window_list == NULL || window_list == window_tail)
    return -1;

  if (d > 0)			/* switch next window */
    {
      single_code ();
      window_tail->next = (LIST *) window_list;
      window_list->prev = (LIST *) window_tail;
      window_tail = window_list;
      window_list = (WINDOW *) window_list->next;
      window_tail->next = NULL;
      multi_code ();
    }
  else
    {
/* switch previous window */
      single_code ();
      window_list->prev = (LIST *) window_tail;
      tmp = (WINDOW *) window_tail->prev;
      tmp->next = NULL;
      window_tail->prev = NULL;
      window_tail->next = (LIST *) window_list;
      window_list->prev = (LIST *) window_tail;
      window_list = window_tail;
      window_tail = tmp;
      multi_code ();
    }

  refresh_window (window_list);
  paint_cursor (window_list);

  return 0;
}

void
rg_switch ()			/* switch to next window */
{
  switch_window (+1);
}
