/*
   11081998tu2033, 17111999we1931, 02032002sa2158
   
   tools/remake.c

   Creates main boot image getting bootsect, setup and kernel together.
   It gets setup and kernel sizes, writes it into bootsect and setup
   (later use of MegaLo) and writes final floppy bootimage copying
   modified bootsect and setup adding bytes for size modulo 512
   to be 0. After this adds kernel and fills up-to-sector space 
   with 0 (just for fun).
   Note: If only two parameters are passed to remake it creates hdd
   boot image for use with MegaLo.

   Setup may be 6 block maximal.
 */

#include <stdio.h>
#include <fcntl.h>
#include <getopt.h>
#include <unistd.h>
#include <sys/stat.h>
#include "../include/phantom/major.h"

/*
  etherboot block for phantom boot via ethernet network
  3.2.2002
*/

struct eb_header
{
  unsigned int magic;
  unsigned int flags_length;
  unsigned int location;
  unsigned int start_point;
};

struct eb_section
{
  unsigned int flags_tags_length;
  unsigned int load_addr;
  unsigned int image_len;
  unsigned int mem_len;
};

struct eb_header eb_header = {
  0x1B031336,			/* magic number */
  0x4,				/* length of header is 16 bytes, no vendor info */
  0x90000E00,			/* location for initial 512 bytes in ds:bx format */
  0x90000200			/* execute address in cs:ip format */
};

#define EB_LASTSECTION		0x04000000

/* real-time setup block */
struct eb_section eb_setup = { 4, 0x90200, 0, 0 };
/* kernel image */
struct eb_section eb_kernel = { 4, 0x10000, 0, 0 };
/* ramdisk */
struct eb_section eb_ramdisk = { EB_LASTSECTION | 4, 0, 0, 0 };

#define BLOCK_SIZE	0x200
#define TOBLOCK(size)	((size)+BLOCK_SIZE-1)/BLOCK_SIZE

void
die (char *s)
{
  printf ("remake dies: %s\n", s);
  exit (0xFF);
}

int
filesize (const char *name)
{
  struct stat st;

  if (stat (name, &st))
    {
      printf ("cannot stat file %s\n", name);
      exit (-1);
    }

  return st.st_size;
}

int
main (int argc, char **argv)
{
  int fd_in, fd_out, i, root_dev = -1, option_index = 0;
  unsigned int setup_blocks;
  unsigned int kernel_blocks;
  unsigned int ramdisk_blocks;
  unsigned int ramdisk_block;
  unsigned int total_blocks;
  unsigned int ramdisk_base = 0;
  char buff[512];
  int written = 0;

  char kernel_file[32] = "";
  char setup_file[32] = "";
  char bootsect_file[32] = "";
  char ramdisk_file[32] = "";
  char image_file[32] = "";

  int fl_etherboot = 0, fl_floppy = 0;

  struct option options[] = {
    {"help", 0, NULL, 'h'},
    {"root", 1, NULL, 'r'},
    {"boot", 1, NULL, 'b'},
    {"kernel", 1, NULL, 'k'},
    {"setup", 1, NULL, 's'},
    {"rdimage", 1, NULL, 'R'},
    {"image", 1, NULL, 'i'},
    {"etherboot", 0, NULL, 'e'},
    {"floppy", 0, NULL, 'f'},
    {NULL, 0, NULL, 0}
  };

  while ((i = getopt_long (argc, argv, "", options, &option_index)) != -1)
    {
      switch (i)
	{
	case '?':
	case 'h':
	  printf
	    ("Usage: remake --setup `setup-file' --kernel `kernel-file' --root `root-dev'\n"
	     "              --image `image-file' [--rdimage `ramdisk image']\n"
	     "              [--boot `bootblock-file'] --etherboot\n");
	  return 0;
	case 'r':
	  root_dev = atoi (optarg);
	  break;
	case 'b':
	  strncpy (bootsect_file, optarg, 32);
	  break;
	case 'k':
	  strncpy (kernel_file, optarg, 32);
	  break;
	case 's':
	  strncpy (setup_file, optarg, 32);
	  break;
	case 'R':
	  strncpy (ramdisk_file, optarg, 32);
	  break;
	case 'i':
	  strncpy (image_file, optarg, 32);
	  break;
	case 'e':
	  fl_etherboot = 1;
	  break;
	case 'f':
	  fl_floppy = 1;
	  break;
	}
    }

  if (fl_floppy && fl_etherboot)
    die ("Cannot use --floppy and --etherboot together.\n");

  if (root_dev == -1)
    die ("Root device not specified.");

  if ((fd_out = creat (image_file, 0644)) == -1)
    die ("Write error.");

  setup_blocks = TOBLOCK (filesize (setup_file));
  kernel_blocks = TOBLOCK (filesize (kernel_file));
  if (ramdisk_file[0])
    {
      ramdisk_block = setup_blocks + kernel_blocks;
      ramdisk_blocks = TOBLOCK (filesize (ramdisk_file));
    }
  else
    ramdisk_blocks = 0;

  total_blocks = setup_blocks + kernel_blocks + ramdisk_blocks;

  if (fl_floppy)
    {
      if ((fd_in = open (bootsect_file, O_RDONLY)) == -1)
	die ("Failed to open bootsector image.\n");

      if (lseek (fd_in, 0, SEEK_END) != 512)
	{
	  close (fd_out);
	  close (fd_in);
	  printf ("Bootsect file is not 512 bytes.\n");
	  return (-1);
	}

      lseek (fd_in, 0, SEEK_SET);
      bzero (buff, 512);
      read (fd_in, buff, 512);
      close (fd_in);

      *(unsigned short *) (buff + 504) = setup_blocks;
      *(unsigned short *) (buff + 506) = kernel_blocks;
      *(unsigned short *) (buff + 508) = root_dev;

      write (fd_out, buff, 512);

      ramdisk_block++;
      total_blocks++;
    }
  else if (fl_etherboot)
    {
      write (fd_out, &eb_header, sizeof (eb_header));
      written += sizeof (eb_header);

      eb_setup.image_len = setup_blocks * BLOCK_SIZE;
      eb_setup.mem_len = eb_setup.image_len;

      write (fd_out, &eb_setup, sizeof (eb_setup));
      written += sizeof (eb_setup);

      eb_kernel.image_len = kernel_blocks * BLOCK_SIZE;
      eb_kernel.mem_len = eb_kernel.image_len;

      if (!ramdisk_blocks)
	eb_kernel.flags_tags_length |= EB_LASTSECTION;
      else
	{
/* Try to load it here, it will not overlap with bss. 
   This should be modified soon. */
    ramdisk_base = 0x100000;

	  eb_ramdisk.load_addr = /* eb_kernel.load_addr
	    + kernel_blocks * BLOCK_SIZE; */ ramdisk_base;
	  eb_ramdisk.image_len = ramdisk_blocks * 512;
	  eb_ramdisk.mem_len = eb_ramdisk.image_len;
	}

      write (fd_out, &eb_kernel, sizeof (eb_kernel));
      written += sizeof (eb_kernel);

      if (ramdisk_blocks)
	{
	  write (fd_out, &eb_ramdisk, sizeof (eb_ramdisk));
	  written += sizeof (eb_ramdisk);
	}

      bzero (buff, 512);
      write (fd_out, buff, 510 - written);
      {
	char sig[2] = { 0x55, 0xAA };
	write (fd_out, sig, 2);
      }
    }

  fd_in = open (setup_file, O_RDONLY);

  read (fd_in, buff, 512);

  *(unsigned short *) (buff + 0x08) = setup_blocks;
  *(unsigned short *) (buff + 0x0A) = kernel_blocks;
  *(unsigned int *) (buff + 0x1C) = root_dev;
  *(unsigned int *) (buff + 0x20) = ramdisk_blocks ? ramdisk_block : 0;
  *(unsigned int *) (buff + 0x24) = ramdisk_blocks;
  *(unsigned int *) (buff + 0x28) = ramdisk_base;

  for (i = 0; i < setup_blocks; i++)
    {
      write (fd_out, buff, 512);
      bzero (buff, 512);
      read (fd_in, buff, 512);
    }

  close (fd_in);

  fd_in = open (kernel_file, O_RDONLY);

  for (i = 0; i < kernel_blocks; i++)
    {
      bzero (buff, 512);
      read (fd_in, buff, 512);
      write (fd_out, buff, 512);
    }

  close (fd_in);

  if (ramdisk_file[0])
    {
      fd_in = open (ramdisk_file, O_RDONLY);

      for (i = 0; i < ramdisk_blocks; i++)
	{
	  read (fd_in, buff, 512);
	  write (fd_out, buff, 512);
	}

      close (fd_in);
    }

  close (fd_out);

  printf ("phantom remake 1.3  target: %s image\n",
	  fl_etherboot ? "etherboot" : (fl_floppy ? "floppy" : "plain"));
  printf ("setup blocks: %d  kernel blocks: %d\n",
	  setup_blocks, kernel_blocks);
  if (ramdisk_blocks)
    printf ("ramdisk blocks: %d  ramdisk start block: %d\n",
	    ramdisk_blocks, ramdisk_block);
  else
    printf ("ramdisk not used\n");
  printf ("root dev: ");
  if (!root_dev)
    printf ("offline mode\n");
  else
    printf ("%02x:%02x\n", MAJOR (root_dev), MINOR (root_dev));
  printf ("image blocks: %d\n", total_blocks);

  return (0);
}
